#!/usr/local/bin/perl

package Clink;
use strict;
use Carp;
use AceParse qw( aceQuote );
use Sequence;
use Data::Dumper;

# All clones live in the _subSeq hash in a Clink

# The format of the _subSeq hash is as follows:

#   {
#     Clone_Name => [ start, end, isClone, direction ],
#     Clone_Name => [ start, end, isClone, direction ],
#     Clone_Name => [ start, end, isClone, direction ]
#   };

# Where start and end are integers representing the location of the
# sequence in the Clink, isClone is a boolean, set to 1 if the
# sequence is genomic (and is therefore needed to deduce the DNA
# sequence), and direction is a boolean which indicates wether the
# sequence is currently stored in the correct orientation in the
# database.


# Have choice of two modes of overlap.
# Sequence obtained from bases indicated with "X":

# DOWNSTAIRS:
#
# Clone_A XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
#             Clone_B ------------------XXXXXXXXXXXXXXXXXXXXXXXX

# UPSTAIRS:
#
#             Clone_B XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# Clone_A XXXXXXXXXXXX------------------

BEGIN {
    my $dir = 1; # Default direction is DOWNSTAIRS
    
    sub DOWNSTAIRS () { $dir;         }
    sub   UPSTAIRS () { $dir ? 0 : 1; }
    
    sub import {
        my( $pkg, $mode ) = @_;
        
        if    ( $mode eq 'UPSTAIRS' )   { $dir = 0; }
        elsif ( $mode eq 'DOWNSTAIRS' ) { $dir = 1; }
        else { croak "Unknown overlap mode '$mode'" if defined $mode }
    }
}

sub new {
    my $pkg = shift;
    my $name = shift; # Optional - only present if read from acedb

    my $new = {
	_name   => {},
	_subSeq => {}
    };
    bless $new, $pkg;
    $new->name($name) if $name;
    return $new;
}

# Duplicate a clink object
sub duplicate {
    my( $clink ) = @_;
    
    my $pkg = ref($clink);
    my $new = $pkg->new();
    $new->name( $clink->name() );
    foreach my $sub ($clink->seqNameList()) {
        $new->subSeq($sub, @{$clink->subSeq($sub)});
    }
    return $new;
}

# Returns ref to the array corresponding to the subsequence name
# given, or overwrites the parameters with new list supplied
sub subSeq {
    my $clink = shift;
    my $name = shift || confess "subSeq() requires a name argument";
    
    if (@_) {
	# Check that we have the correct number of arguments
	unless (scalar @_ == 4) {
 	    croak "subSeq requires four values, ",
	    scalar @_, ' supplied: [', join(', ', map "'$_'", @_), ']';
	}
	# @_ should contain start, end, isClone, direction
	$clink->{'_subSeq'}{$name} = [ @_ ];

    } else {	
	# Return a ref to the array
	return $clink->{'_subSeq'}{$name};
    }
}

# Methods to access elements in subSeq arrays by name
BEGIN {

    my( %fields );
    {
        my $i = 0;
        %fields = map {$_, $i++} qw( start end isClone direction );
    }
    
    while (my( $field, $i ) = each %fields) {
        no strict 'refs';
                
        *$field = sub {
            my( $clink, $name, $val ) = @_;
            
            eval {
                if ($val) {
	            $clink->subSeq( $name )->[$i] = $val;
                } else {
	            $val = $clink->subSeq( $name )->[$i];
                }
            };
            confess "No data for: $name->$field()", Dumper($clink) if $@;
            return $val;
        }
    }
}

sub name {
    my $clink = shift;
    
    if (@_) {
        foreach my $name (@_) {
	    $clink->{'_name'}{$name} = 1;
        }
    } else {
	return sort keys %{$clink->{'_name'}};
    }
}

# The name of the LINK, which is the name of the 
# first clone in the clink
sub current_name {
    my( $clink, $prefix ) = @_;
    
    $prefix ||= 'LINK_';
    return $prefix. ($clink->cloneNameList)[0];
}

# Calculated: read-only
# Called with no arguments it returns the length of the 
# entire Clink
sub length {
    my $clink = shift;
    if (@_) {
	my $name = shift;
	return $clink->end($name) - $clink->start($name) + 1;
    } else {
	# Get length of the whole clink
	my $max;
	foreach my $clone ($clink->cloneNameList) {
	    my $end = $clink->end($clone);
	    if ($end > $max) {
		$max = $end;
	    }
	}
	return $max;
    }
}

# Returns a list of clones in the clink, ordered by their
# left, then right ends
sub cloneNameList {
    my $clink = shift;
    my @ret;

    # Take ref to hash to make sort more readable
    my $sub = $clink->{'_subSeq'};

    return sort { $sub->{$a}[0] <=> $sub->{$b}[0] or
	          $sub->{$a}[1] <=> $sub->{$b}[1] }
           grep $clink->isClone($_), keys %$sub;
}

# Returns a list of all sequences in the clink, ordered by their
# left, then right ends
sub seqNameList {
    my $clink = shift;
    my @ret;

    # Take ref to hash to make sort more readable
    my $sub = $clink->{'_subSeq'};

    return sort { $sub->{$a}[0] <=> $sub->{$b}[0] or
	          $sub->{$a}[1] <=> $sub->{$b}[1] } keys %$sub;
}

sub endClones {
    my $clink = shift;
    my @list = $clink->cloneNameList();
    return @list[0, $#list];
}

sub clinkFromLink {
    my( $pkg, $link ) = @_;

    my $subs = $link->at('Structure.Subsequence');
    my $clink = $pkg->new( $link->name() );

    foreach my $row ($subs->col()) {
	my( $gen, $start, $end, $isClone );
	$gen = $row->fetch();

	# Record the start and end
	($start, $end) = map "$_", $row->row(1);
	confess "Missing position for '$gen' in '$link': '$start', '$end'"
	    unless ($start and $end);

	# Find out if the sequence is a clone
	if ($gen->at('Properties.Genomic_canonical')) {
	    $isClone = 1;
	} else {
	    $isClone = 0;
	}

	# Add the info to the clink. "1" at end for all sequences
	# indicates that they are in the same (and correct) orientation
	$clink->subSeq($gen->name, $start, $end, $isClone, 1);
    }
    return $clink;
}

# Generates acefile format for reading a link into acedb
sub toAce {
    my( $clink, $prefix ) = @_;
    my(@return, @subs);
    
    # Link is always named after its leftmost sequence,
    # but can supply a different prefix
    my $link = $clink->current_name($prefix);
    push(@return, "\n", aceQuote('-D', 'Sequence', $link));
    
    # Delete the old link objects
    foreach my $old ($clink->name()) {
        next if $old eq $link; # Only delete old link once
        push(@return, "\n", aceQuote('-D', 'Sequence', $old));
    }
    
    # Format the new link
    push(@return, "\n", aceQuote('Sequence', $link), "Link\n");
    foreach my $name ($clink->seqNameList()) {
        my $clone = $clink->subSeq($name);
        push(@return, aceQuote('Subsequence', $name, @$clone[0,1]));
        push(@subs, "\n", aceQuote('Sequence', $name),
                          aceQuote('Source', $link));
    }
    return( @return, @subs );
}

# Join two links which share a common clone
sub fuse {
    my( $x, $y ) = @_;
    my( $join );
        
    # Look for a common sequence in the two clinks
    foreach my $name ($x->cloneNameList()) {
        if ($y->subSeq($name)) {
            $join = $name;
            last;
        }
    }
    # Can't join two links if they don't share a sequence
    warn "No common sequence found" and return unless $join;
    
    my $x_dir = $x->direction($join);
    my $y_dir = $y->direction($join);
    if    ($x_dir == 0 and $y_dir == 1) {
        $y->reverse();
    }
    elsif ($x_dir == 1 and $y_dir == 0) {
        $x->reverse();
    }
    
    # Find the leftmost clink and make it $x
    if ($x->start($join) < $y->start($join)) {
        my $z = $x;
        $x = $y;
        $y = $z;
    }
    
    # Make a new link by cloning $x
    my $new = $x->duplicate();
    
    # Copy across any old names from $y into $new,
    # to be deleted by the new object
    $new->name( $y->name() );
    
    # Calculate the offset of the coordinates in $y
    # relative to $x
    my $offset = $x->start($join) - $y->start($join);
    
    # Add the subsequences in $y to $new,
    # with adjusted coordinates
    foreach my $name ($y->seqNameList()) {
        my $sub   = $y->subSeq($name);
        my $start = $y->start($name) + $offset;
        my $end   = $y->end($name)   + $offset;
        $new->subSeq($name, $start, $end, @$sub[2,3]);        
    }
    return $new;
}


=pod

sub split {
    my( $clink, $name ) = @_;
    
    my $sub = $clink->subSeq( $name )
        or confess "No such subSeq '$name'";
    
    
    return( $x, $y, @dead );
}

=cut

# Returns a Sequence object with the dna of the entire link
sub sequence {
    if (DOWNSTAIRS) {
        downstairs_sequence( @_ );
    } else {
        confess "UPSTAIRS 'sequence' method not yet implemented";
    }
}

sub downstairs_sequence {
    my( $clink, $seq_sub ) = @_;
    
    # &$dna_sub should return a Sequence object when
    # given the name of the subsequence
    croak "Argument to dna_sub() must be ref to subroutine"
        unless ref($seq_sub) eq 'CODE';
    
    my $link_name = $clink->current_name();
    
    my( $DNA );
    my $last = 0;
    foreach my $me ($clink->cloneNameList) {
    
        my $end = $clink->end( $me );
        
        # Avoid sequences which are contained entirely within
        # a previous sequence.
        next if $end <= $last;
        
        my $len = $end - $last;
        my $off = $clink->length( $me ) - $len;

        # Get Sequence object
        my( $seq );
        {
            my( $test );
            $seq = &$seq_sub( $me );
            eval{ $test = $seq->isa('Sequence') };
            confess("seq_sub('$me') didn't return a sequence object") unless $test;
        }
        
        # Reverse complement it if needed.
        unless ($clink->direction($me)) {
            $seq->revcomp_iub();
        }
        
        $DNA .= substr($seq->seq(), $off, $len);
        
        # $last is the end of the previous sequence in the link
        $last = $end;
    }
    
    # Return a new sequence object
    return Sequence->new_from_strings($clink->current_name(), $DNA);
}

sub seq_coords {
    if (DOWNSTAIRS) {
        downstairs_coords( @_ );
    } else {
        confess "UPSTAIRS 'seq_coords' method not yet implemented";
    }
}

sub downstairs_coords {
    my( $clink, $seq_sub ) = @_;
    
    my $link_name = $clink->current_name();
    my $last = 0;
    my( @coords );
    
    foreach my $me ($clink->cloneNameList) {
    
        my $link_end = $clink->end( $me );
        
        # Avoid sequences which are contained entirely within
        # a previous sequence.
        next if $link_end <= $last;
                
        my $link_start = $last + 1;
        my $seq_end = $clink->length($me);
        my $seq_start = $seq_end - ($link_end - $last) + 1;
        
        push( @coords, [$link_name, $link_start, $link_end, $me, $seq_start, $seq_end] );
                        
        # $last is the end of the previous sequence in the link
        $last = $link_end;
    }
    
    return @coords;
}


# Reverses the coordinates of a clink in place
sub reverse {
    my( $clink ) = @_;
    
    # Get the total length of the clink
    my $length = $clink->length();
    
    foreach my $name ($clink->seqNameList()) {
    
        # Flip coordinates
        my $start = $length - $clink->end($name)   + 1;
        my $end   = $length - $clink->start($name) + 1;
        $clink->start($name, $start);
        $clink->end($name, $end);
        
    }
    $clink->toggle_directions();
}

# Toggle direction flags if majority are reverse matches
sub correctDirection {
    my( $clink ) = @_;
    
    if ($clink->direction_vote() < 0) {
        $clink->reverse()
    }
}

sub direction_vote {
    my( $clink ) = @_;
    
    my $vote = 0;
    
    # Only clones get to vote!
    foreach my $name ($clink->cloneNameList) {
        $clink->direction($name) ? $vote++ : $vote--;
    }
    return $vote;
}

sub toggle_directions {
    my( $clink ) = @_;
    
    foreach my $name ($clink->seqNameList) {
        if ($clink->direction($name)) {
            $clink->direction($name, 0);
        } else {
            $clink->direction($name, 1);
        }
    }
}


sub glue {
    my( $clink,     # The clink object
        $join,      # Name of sequence already in clink
        $new,       # Name of new sequence to add to link
        $dir,       # TRUE for correct direction FALSE for reverse
        $offset,    # Postition of start of $new relative to start of $join
        $length,    # Length of $new
        ) = @_;
        
    # Make $offset of $new relative to whole clink
    my $join_coord = $clink->start( $join ) - 1;
    $offset = $join_coord + $offset;
            
    # Offset is less than zero if the clink is being extended
    # leftwards by $new
    if ($offset < 0) {
        $offset *= -1;
        foreach my $seq ($clink->seqNameList) {
            my $x = $clink->start( $seq );
            $clink->start( $seq, $offset + $x );
            my $y = $clink->end( $seq );
            $clink->end( $seq, $offset + $y );
        }
        $offset = 0;
    }
        
    my $x = 1       + $offset;
    my $y = $length + $offset;
    
    $clink->subSeq( $new, $x, $y, 1, $dir );
}


################################################################################
################################################################################
################################################################################
################################################################################
################################################################################

package Clink::Contig;
use strict;
use Carp;
use CrossMatch;
use Data::Dumper;

# Create a new contig object
sub new {
    my $pkg = shift;

    return bless {
	name => undef,
	source => undef,
	clones => [],
	tester => \&CrossMatch::endHits,
	LUT => {}
    }, $pkg;
}
# Accessor methods for simple parameters
sub mapName {
    my $contig = shift;
    @_ ? $contig->{'name'} = shift : $contig->{'name'};
}
sub source {
    my $contig = shift;
    @_ ? $contig->{'source'} = shift : $contig->{'source'};
}
sub cloneList {
    my $contig = shift;
    return @{$contig->{'clones'}};
}
sub clinkList {
    my $contig = shift;
    my( %cl, @clinks );
    foreach my $clone ($contig->cloneList()) {
        if (my $clink = $clone->clink()) {
            unless ($cl{$clink}) {
                $cl{$clink} = 1;
                push( @clinks, $clink );
            }
        }
    }
    return @clinks;
}
sub tester {
    my $contig = shift;
    @_ ? $contig->{'tester'} = shift : $contig->{'tester'};
}

# Pull a named clone out of a contig,
# or add a new one onto the end
sub clone {
    my $contig = shift;
    my $name = shift;

    if (@_) {
	my ($position, $length) = @_;

	my $pkg = ref($contig);
	my $clone = {name       => $name,
		     position   => $position,
		     length     => $length,
		     clink      => undef,
		     matches    => {},
		     properties => {} };
	# Bless clone into same package as contig
	bless $clone, $pkg;

	# Add to end of clones array
	push( @{$contig->{'clones'}}, $clone );
	# Add entry in look up table
	$contig->{'LUT'}{$name} = $#{$contig->{'clones'}};
    } else {
	# Look up clone in LUT, and return
	my $i = $contig->{'LUT'}{$name};
        confess "Clone '$name' not in contig" unless defined $i;
	return $contig->{'clones'}[ $i ];
    }
}

# Return the position of a clone object in a contig
sub cloneIndex {
    my $contig = shift;
    my $clone = shift;

    return $contig->{'LUT'}{ $clone->name() };
}

sub super_link {
    my( $contig, $spacer_length ) = @_;
    
    # Default space between link elements is 800bp
    $spacer_length ||= 800;
    
    my $super = Clink->new();
    
    my( %seen );
    my $last = 1;
    foreach my $clone ($contig->cloneList) {
        my $name = $clone->name();
        my $length = $clone->length();
                
        # If the clone's in a clink, then the link is added
        my( $clink );
        if ($clink = $clone->clink) {
            $name = $clink->current_name();
            if ($seen{ $name }) {
                # Don't add the link if it's already in $super
                next;
            } else {
                $seen{ $name } = 1;
                $length = $clink->length();
            }
        }
        # ...otherwise the clone itself is added
        else {
            $name = $clone->name();
            $length = $clone->length()
                or confess "Can't determine length of $clone";
        }
        my $start_pos = $last;
        my $end_pos   = $last + $length - 1;
        
        $super->subSeq($name, $start_pos, $end_pos, 1, 1);
        
        # Next subSeq will be added after a spacer
        $last += $spacer_length + $length;
    }
    
    return $super;
}

# Accessor methods for the simple fields posessed
# by each element of the clones hash
BEGIN {
    foreach my $field (qw( name length position clink )) {
        no strict 'refs';
        
        *$field = sub {
            my( $clone, $val ) = @_;

            if ($val) {
	        $clone->{$field} = $val;
            } else {
	        return $clone->{$field};
            }
        }
    }
}

# The match method stores or retrieves CrossMatch objects
sub match {
    my( $clone, $name, $match ) = @_;

    confess "match() - No name provided" unless $name;

    if ($match) {
        # Storage
	$clone->{'matches'}{$name} = $match;
    } else {
        # Retrieval
	return $clone->{'matches'}{$name};
    }
}

# Method which provides access to a hash for storage of miscellaneous
# information
sub property {
    my $clone = shift;

    my $key = shift || confess "property() - No key supplied";

    if (@_) {
	$clone->{'property'}{$key} = shift;
    } else {
	return $clone->{'property'}{$key};
    }
}

# Has effect of sorting by name if position fields are empty or the
# same.
sub sortByPosition {
    my $contig = shift;

    # Sort contig by position or name
    @{$contig->{'clones'}} = sort {   $a->{'position'} <=> $b->{'position'}
				   or $a->{'name'}     cmp $b->{'name'}    }
                                @{$contig->{'clones'}};
    
    $contig->makeIndex();
}

# Takes a contig and a clink, and makes sure that
# all the clones named in the clink point to that clink
sub registerClink {
    my( $contig, $clink ) = @_;
        
    foreach my $name ($clink->cloneNameList()) {
        my $clone = $contig->clone($name)
            or confess "Clone '$name' missing from contig";
        $clone->clink( $clink );
    }
}

# Deltes all the matches from a contig
sub deleteMatches {
    my( $contig ) = @_;
        
    foreach my $clone ($contig->cloneList) {
        $clone->{'matches'} = {};
    }
}

# Re-makes the look up table for a contig
sub makeIndex {
    my $contig = shift;
    
    $contig->{'LUT'} = {};
    
    foreach my $i (0..$#{$contig->{'clones'}}) {
	my $name = $contig->{'clones'}[$i]{'name'};
	$contig->{'LUT'}{ $name } = $i;
    }
}

# Reports wether a clone object has an un-linked end,
# and hence is a candidate for fusing to another clone
sub isEnd {
    my $clone = shift;

    if (my $clink = $clone->clink) {
	# Is the clone on the end of a clink
	foreach my $end ($clink->endClones) {
	    return 1 if $clone->name() eq $end;
	}
	return 0;
    } else {
	# Isn't in a clink, so it must be an end
	return 1;
    }
}

# Method to match the free ends in a contig
sub matchEnds {
    my( $contig, $crossMatcher, $offset ) = @_;
    $offset ||= 2;

    unless ($offset =~ /^\d+$/) {
        confess("Not a positive integer '$offset'");
    }

    my @clones = $contig->cloneList();

    MATCH: foreach my $range (1..$offset) {

        foreach my $clone (@clones) {
        
	    # Not interested in this clone if it isn't an end
            next unless $clone->isEnd;
            
            my $name = $clone->name();
	    my $index = $contig->cloneIndex($clone);

	    # Match against the two clones $range away on either side
	    foreach my $p ($index - $range, $index + $range) {
                                
                next if ($p < 0 or $p > $#clones);
                
	        my $other = $clones[$p];
	        my $otherName = $other->name();
                                
	        # Skip this one if we've already done a match
	        next if $clone->match( $otherName );

	        $contig->doMatch( $crossMatcher, $name, $otherName );
	    }
        }
    }
}

sub doMatch {
    my( $contig, $crossMatcher, $Aname, $Bname ) = @_;
    
    my $A = $contig->clone($Aname);
    my $B = $contig->clone($Bname);
    
    # Do a cross_match
    my $match = $crossMatcher->crossMatch( $Aname, $Bname )
        or confess "No match returned for '$Aname' vs '$Bname'";

    # Add matches to clone under name of match partner
    $A->match( $Bname, $match );
    $B->match( $Aname, $match );

    # Attempt to join them
    if ($contig->clone($Aname)->clink()) {
	$contig->tryClones( $A, $B );
    } else {
        $contig->tryClones( $B, $A );
    }
}

sub tryClones {
    if (Clink::DOWNSTAIRS) {
        downstairs_tryClones( @_ );
    } else {
        confess "UPSTAIRS 'tryClones' method not implemented";
    }
}

# Deals with all possible overlap scenarios, and calls
# $clink->glue with the appropriate parameters
sub downstairs_tryClones {
    my( $contig, $A, $B ) = @_;
    my $Aname = $A->name();
    my $Bname = $B->name();

    # Get the match from the contig object and filter it
    my $m = $A->match( $Bname )
        or confess "Can't fetch match for '$Bname' from clone '$Aname'\n";
    $m->filter( $contig->tester );
    
    #print "$Aname vs $Bname\n";
    
    # Can't join clones in a clink if don't have a unique match
    unless ($m->count == 1) {
        warn "Ambiguous hits '$Aname' vs '$Bname'",
            Dumper( $m ) if $m->count > 1;
        return;
    }

    # Record the length of the sequences as reported by cross_match
    my $Alength = $m->seqlength( $Aname );
    my $Blength = $m->seqlength( $Bname );

    # Get the clink for $A, or make one
    my( $clink );
    unless ($clink = $A->clink()) {
        $clink = Clink->new();
        $clink->subSeq($Aname,
                       1,
                       $Alength,
                       1, 1 );
    }
    
    # Record offset and direction of $A in clink
    my $dir = $clink->direction( $Aname );

    # Initialise decision variables
    my( $As, $Bs, $Ar, $Br );
    $As = 1 if $m->start(  $Aname ) == 1;
    $Bs = 1 if $m->start(  $Bname ) == 1;
    $Ar = 1 if $m->remain( $Aname ) == 0;
    $Br = 1 if $m->remain( $Bname ) == 0;

    # Use the decision variables to deal with all the
    # possible overlap scenarios
    my( $offset );
    if ($m->strand eq 'W') {

	# Both sequences are in the same direction
	if    ($As) {
            if ($dir) {
                # B--A print "A ";
                $offset = $m->end($Aname) - $Blength;
            } else {
                # A--B print "B ";
                $offset = $Alength - ($Blength - $m->start($Bname)) - 1;
            }
	}
        elsif ($Bs) {
            if ($dir) {
                # A--B print "C ";
                $offset = $Alength - $m->end($Bname);
            } else {
                # B--A print "D ";
                $offset = ($Alength - $m->start($Aname)) - $Blength + 1;
            }
	}
        else {
	    warn "Bad match '$Aname' vs '$Bname' : ",
                join(', ', map "'$_'", ($As, $Bs, $Ar, $Br)),
                "\n", Dumper( $m );
            return;
	}
    } else {
	# Sequences are in opposite directions

	if    ($As and $Ar) {
            # A is contained within B
            if ($dir) {
                #print "E ";
                $offset = -1 * $m->remain($Bname);
            } else {
                #print "F ";
                $offset = -1 * ($m->start($Bname) - 1);
            }
	}
        elsif ($Bs and $Br) {
            # B is contained within A
            if ($dir) {
                #print "G ";
                $offset = ($m->start($Aname) - 1);
            } else {
                #print "H ";
                $offset = $m->remain($Aname);
            }
	}
        elsif ($As and $Bs) {
            if ($dir) {
                # B--A print "I ";
                $offset = $m->end($Aname) - $Blength;
            } else {
                # A--B print "J ";
                $offset = $Alength - $m->end($Bname);
            }
	}
        elsif ($Ar and $Br) {
            if ($dir) {
                # A--B print "K ";
                $offset = $Alength - ($Blength - $m->start($Bname)) - 1;
            } else {
                # B--A print "L ";
                $offset = ($Alength - $m->start($Aname)) - $Blength - 1;
            }
	}
        else {
	    warn "Bad match '$Aname' vs '$Bname' : ",
                join(', ', map "'$_'", ($As, $Bs, $Ar, $Br)),
                "\n", Dumper( $m );
            return;
	}
        
        # $B is in opposite direction to $A
        $dir = $dir ? 0 : 1;
    }

    # Add the clone into the clink structure
    $clink->glue($Aname, $Bname, $dir, $offset, $Blength);

    #print "\n";

    # If $B is also in a clink, then fuse the two clinks
    if (my $Bclink = $B->clink) {
        $clink = $clink->fuse($Bclink)
            or confess "Can't fuse links";
    }

    # Point all the clones appearing in the clink to it
    $contig->registerClink($clink);
}

# Deals with all possible overlap scenarios, and calls
# $clink->glue with the appropriate parameters
sub old_tryClones {

    my $contig = shift;
    my( $A, $B ) = @_;
    my $Aname = $A->name();
    my $Bname = $B->name();

    # Get the match from the contig object and filter it
    my $m = $A->match( $Bname )
        or confess "Can't fetch match for '$Bname' from clone '$Aname'\n";
    $m->filter( $contig->tester );
    
    # Can't join clones in a clink if don't have a unique match
    unless ($m->count == 1) {
        warn "Ambiguous hits '$Aname' vs '$Bname'" if $m->count > 1;
        return;
    }

    # Record the length of the sequences as reported by cross_match
    my $Alength = $m->seqlength( $Aname );
    my $Blength = $m->seqlength( $Bname );

    # Get the clink for $A, or make one
    my( $clink );
    unless ($clink = $A->clink()) {
        $clink = Clink->new();
        $clink->subSeq($Aname,
                       1,
                       $Alength,
                       1, 1 );
    }
    
    # Record offset and direction of $A in clink
    my $dir = $clink->direction( $Aname );

    # Initialise decision variables
    my( $As, $Bs, $Ar, $Br );
    $As = 1 if $m->start(  $Aname ) == 1;
    $Bs = 1 if $m->start(  $Bname ) == 1;
    $Ar = 1 if $m->remain( $Aname ) == 0;
    $Br = 1 if $m->remain( $Bname ) == 0;

    # Use the decision variables to deal with all the
    # possible overlap scenarios
    if ($m->strand eq 'W') {

	# Both sequences are in the same direction
	if    ($As) {
            if ($dir) {
                # B--A  print "A\n";
                $clink->glue($Aname, $Bname, $dir, -1 * ($m->start($Bname) - 1), $Blength);
            } else {
                # A--B  print "B\n";
                $clink->glue($Aname, $Bname, $dir, $m->remain($Aname), $Blength);
            }
	}
        elsif ($Bs) {
            if ($dir) {
                # A--B  print "C\n";
                $clink->glue($Aname, $Bname, $dir, ($m->start($Aname) - 1), $Blength);
            } else {
                # B--A  print "D\n";
                $clink->glue($Aname, $Bname, $dir, -1 * $m->remain($Bname), $Blength);
            }
	}
        else {
	    confess "Bad match '$Aname' vs '$Bname' : ",
                join ', ', map "'$_'", ($As, $Bs, $Ar, $Br),
                "\n", Dumper( $m );
	}
    } else {

	# Sequences are in opposite directions
	if    ($As and $Ar) {
            # A is contained within B
            if ($dir) {
                # print "E\n";
                $clink->glue($Aname, $Bname, 0, -1 *      $m->remain($Bname), $Blength);
            } else {
                # print "F\n";
                $clink->glue($Aname, $Bname, 1, -1 * ($m->start($Bname) - 1), $Blength);
            }
	}
        elsif ($Bs and $Br) {
            # B is contained within A
            if ($dir) {
                # print "G\n";
                $clink->glue($Aname, $Bname, 0, ($m->start($Aname) - 1), $Blength);
            } else {
                # print "H\n";
                $clink->glue($Aname, $Bname, 1,      $m->remain($Aname), $Blength);
            }
	}
        elsif ($As and $Bs) {
            if ($dir) {
                # B--A  print "I\n";
                $clink->glue($Aname, $Bname, 0, -1 * $m->remain($Bname), $Blength);
            } else {
                # A--B  print "J\n";
                $clink->glue($Aname, $Bname, 1,      $m->remain($Aname), $Blength);
            }
	}
        elsif ($Ar and $Br) {
            if ($dir) {
                # A--B  print "K\n";
                $clink->glue($Aname, $Bname, 0,      ($m->start($Aname) - 1), $Blength);
            } else {
                # B--A  print "L\n";
                $clink->glue($Aname, $Bname, 1, -1 * ($m->start($Bname) - 1), $Blength);
            }
	}
        else {
	    confess "Bad match '$Aname' vs '$Bname' : ",
                join ', ', map "'$_'", ($As, $Bs, $Ar, $Br),
                "\n", Dumper( $m );
	}
    }


    # If $B is also in a clink, then fuse the two clinks
    if (my $Bclink = $B->clink) {
        $clink = $clink->fuse($Bclink)
            or confess "Can't fuse links";
    }

    # Point all the clones appearing in the clink to it
    $contig->registerClink($clink);
}

1;

__END__

=head1 NAME Clink - manipulate acedb link objects

=head1 SYNOPSIS


=head1 DESCRIPTION

B<Clink.pm> facilitates the creation and management of acedb link
objects (B<C>lone B<link>s).  Link objects have I<Properties.Link>,
and contain I<Subsequence> objects which are genomic clones.

=head1 AUTHOR

B<James Gilbert> email jgrg@sanger.ac.uk

