# Curve_plot.pm

# Copyright (c) 1999 by Raphael Leplae <lp1@sanger.ac.uk>.
# All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.

=head1 NAME

Curve_plot.pm - Generates 2D curves in a GIF file + handles imagemaps.

=head1 SYNOPSIS

    use Curve_plot;
    
    # Create new graph, specifying width and height
    $graph = new Curve_plot (800, 700);
    
    # Define titles
    $graph->title("My plot");
    $graph->x_title ('My X axis title');
    $graph->y_title ('My Y axis title');
    
    # Add data sets for curve drawing
    $graph->add_dataset ("Curve1", \@xdata1, \@ydata1);
    $graph->add_dataset ("Curve2", \@xdata2, \@ydata2);
    $graph->add_dataset ("Curve3", \@xdata3, \@ydata3);
    
    # Define X axis
    $graph->x_axis (\@x_values);
    
    # Set my x_format function for nice label display (see below)
    $graph->set ('x_label_format' => \&x_format);
    
    # I want to automatically redefine X axis for
    # evenly spaced X values and assign automatically a colour to each curve
    $graph->set ('x_evenlyspace' => 1, 'auto_colour' => 1);
    
    # Define threshold for spacing X axis entries
    $graph->set ('x_threshold' => 3);
    
    # Define a default font for the output
    $graph->set ('font' => 'small');
    
    # Define Y axis and pass my &to_log function
    # to transform each Y values (axis and data sets)
    # => Y becomes logarithmic with &to_log (see below)
    $graph->y_axis (\@y_values, \&to_log);
    
    # Set my y_format function for well formatted label display
    $graph->set ('y_label_format' => \&y_format);
    
    # Create the plot
    $graph->plot ();
    
    # Save it in a gif file
    $graph->save_gif ("plot.gif");
    
    # Write HTML code for the image + image map named 'map' !
    print '<IMG SRC="plot.gif" ALT="curves" BORDER="0" USEMAP="#map">', "\n";
    
    # Add image map for three curves
    print $graph->start_imagemap ('map');
    
    # In this case, I want a map for each curve
    print $graph->imagemap_area (0, "Target1");
    print $graph->imagemap_area (1, "Target2");
    print $graph->imagemap_area (2, "Target3");
    
    # That's all
    print $graph->end_imagemap ();
    
    # Add image map targets here
    print '<A NAME="Target1">', "\n";
    
    # Colour title with the same colour of the curve
    print '<FONT COLOR="', $graph->get_col_hex(0), '">Curve1</FONT><P>', "\n";
    print "Curve 1 shows strong correlation between the number of cancers developed in Area 1 between 1980 and 1990 and the number of pay phones installed in the same area during the same period.<P>\n";
    
    # And so on for the other curves
    
    # Formatting the labels
    sub x_format {
        my $val = shift;
        return sprintf "%3.2f", $val;
    }
        
    sub y_format {
        my $val = shift;
        return sprintf "%0.2e%%", $val;
    }
        
    # Transforming Y values on the fly
    sub to_log {
        my $val = shift;
	
	# Return lowest value of the Y scale (0 would not work!)
        $val = $y_values[0] if ($val == 0);
	
        return (log($val)/log(10));
    }

=head1 DESCRIPTION

Curve_plot was originally designed to plot database search evaluation curves but can be used for general curve plotting. To create a plot, the basic information required are the data to be plotted. Once the settings are performed, the plot can be stored in a GIF file. Only values are accepted, no characters, dates or other "non-values" can be plotted on the graph (at the moment).

This module allows to transform the X and/or Y values while plotting the graph by providing your transformation functions. For example, you can convert your Y data to a logarithmic scale while plotting the curves. If the values passed to Curve_plot lead to long labels, you can also provide a function for reformatting these values before drawing them on the axis. An imagemap utility is available if you want to provide more functionalities on the web page with the plot.

It is also possible to evenly space the X and Y axis according to the data passed to the module. In that case, a new set of data is generated from the available information. If an X axis is provided with each data set, the evenly spaced X axis will contain the lowest value and highest value from the X axis information. If no X data is available at all, the method will build a standard X axis based on the number of entries present in the datasets. The evenly space method is rather accurate and you can play with the spacing through a threshold value (from numerous close ticks to few spaced ticks).

The module GD.pm by Lincoln Stein is required for Curve_plot.

=head1 METHODS

=cut

package Curve_plot;

use strict;
use GD;
use Carp qw(cluck confess);
#use Data::Dumper;

use constant LINE_STYLE =>  0;
use constant DOT_STYLE =>   1;

use constant PLAIN_LINE =>  0;
use constant DOTTED_LINE => 1;
use constant DASHED_LINE => 2;

use constant SQUARE_DOT =>  0;
use constant CROSS_DOT =>   1;
use constant PLUS_DOT =>    2;
use constant CIRC_DOT =>    3;
use constant POINT_DOT =>   4;

# Default font
my $FONT = gdMediumBoldFont ;
my $FONTWIDTH  = 7;
my $FONTHEIGHT = 13;

require Config;

# Define default global variables & values
use vars qw($VERSION $OS $needs_binmode $base_colours);

$VERSION = '0.5';

# Define basic colours
@{$base_colours->[0]} = (255, 0, 0);
@{$base_colours->[1]} = (0, 175, 0);
@{$base_colours->[2]} = (0, 0, 255);
@{$base_colours->[3]} = (255, 175, 0);
@{$base_colours->[4]} = (255, 0, 255);
@{$base_colours->[5]} = (0, 175, 255);
@{$base_colours->[6]} = (255, 125, 125);
@{$base_colours->[7]} = (125, 0, 125);
@{$base_colours->[8]} = (125, 125, 0);

# Get OS type
$OS = $Config::Config{'osname'};

# Assign OS name
if ($OS=~/Win/i) {
    $OS = 'WINDOWS'; }
elsif ($OS=~/vms/i) {
    $OS = 'VMS'; }
elsif ($OS=~/^MacOS$/i) {
    $OS = 'MACINTOSH'; }
elsif ($OS=~/os2/i) {
    $OS = 'OS2'; }
else {
    $OS = 'UNIX'; }

# Define binmode for GIF file
$needs_binmode = $OS=~/^(WINDOWS|VMS|OS2)/;

# ============================================================
# new
# Passed:   width and height of the graph
# Returns:  new object
# On error: 
# Description: 
# Creates new Curve_plot object and set all default values.
# See init() for default values set.
# ============================================================

=head2 new()

    $graph = new Curve_plot ($width, $height);

Curve_plot::new creates a new object with the width and height passed to the method. Default size is 600x400. The optional third argument can be a GD object on which the plot has to be drawn.
On failure, an error message will be printed and the script will stop.

=cut

sub new {
    my $class = shift;
    my $w = shift;
    my $h = shift;
    my $gd = shift;
    my $transparent;
    
    my $self = {};

    bless $self, $class;

    if (defined $w){
	$self->{'width'} = $w; }
    else {
	$self->{'width'} = 600; }
    
    if (defined $h) {
	$self->{'height'} = $h; }
    else {
	$self->{'height'} = 400; }

    # Create new GD image if needed
    if (defined $gd) {
    	$self->{'gd'} = $gd; }
    else {
	$self->{'gd'} = new GD::Image($self->{'width'}, $self->{'height'});
	confess ("Unable to create new GD image.\n") if (! defined $self->{'gd'});
	}

    $self->_init();
    
    if (!defined($gd) or defined($transparent)) {
    	# Set transparency
	$self->{'gd'}->transparent($self->{'bg_colour'});
	$self->{'gd'}->interlaced('true');

	# Make a filled rec on the whole gif with background colour
	$self->{'gd'}->filledRectangle(0, 0, 
    	    	    		   $self->{'width'} - 1, 
    	    	    		   $self->{'height'} - 1,
				   $self->{'bg_colour'});
    	}
	
    return $self;
}

# ============================================================
# Function: _transformation_matrix
# Passed: boolean indicate of 'vertical mode' == defined
#         GIF (x,y) origin also (needed for reflection != (0,0)) 
# Returns:  closure sub to transform (x,y) coordinates
# On error: none
# Description: 
#  Generates a transformation matrix function by closure
#  to reflect the graph on a diagonal axis (for vertical 
#  plots). Usage: allocate when vertical mode, use to
#  transform all coordinates passed to GD functions.
#  The transformation function takes and returns a 2-tuple (x,y) list
# ============================================================
sub _transformation_matrix {
    my $v = shift ;
    my $h = shift ; $h=0 if !defined($h);
    my $k = shift ; $k=0 if !defined($k);
    if(defined($v) and $v) {
	return sub { my ($x,$y)=@_ ; return ($y-$k+$h,$x-$h+$k) ; } ;
    } else {
	return sub { my ($x,$y)=@_ ; return ($x,$y) ; } ; # identity matrix!
    }
}

# ============================================================
# Function: _init
# Passed:   
# Returns:  1
# On error: confess
# Description: 
# Set all default values required to generated the gif plot.
# Initialise the GD object and set background and foreground
# colours.
# ============================================================

sub _init {
    my $self = shift;
    
    # Assing B & W
    $self->{'fg_colour'} = $self->{'gd'}->colorAllocate(0,0,0);
    $self->{'bg_colour'} = $self->{'gd'}->colorAllocate(255,255,255);
    
    # Default graph location
    $self->{'x'} = 0;
    $self->{'y'} = 0;
    
    # Set default values
    $self->{'x_off'} = 5;
    $self->{'y_off'} = 5;
    
    # Default font
    $self->{'font'}  = $FONT ;
    $self->{'font_width'}  = $FONTWIDTH ;
    $self->{'font_height'} = $FONTHEIGHT ;
    
    # Set default transformation: identity!
    $self->{'vertical'}  = undef ; # normal horizontal graph
    $self->{'transform'} = _transformation_matrix ;

    # Plot title
    $self->{'title'}->{'str'} = "";
    $self->{'title'}->{'x'} = 0;
    $self->{'title'}->{'y'} = 0;
    $self->{'title'}->{'w'} = 0;
    $self->{'title'}->{'h'} = 0;
    $self->{'title'}->{'col'} = undef;
    
    # Plot axis
    $self->{'axis'}->{'x'}->{'title'}->{'str'} = "";
    $self->{'axis'}->{'x'}->{'title'}->{'x'} = 0;
    $self->{'axis'}->{'x'}->{'title'}->{'y'} = 0;
    $self->{'axis'}->{'x'}->{'title'}->{'w'} = 0;
    $self->{'axis'}->{'x'}->{'title'}->{'h'} = 0;
    $self->{'axis'}->{'x'}->{'title'}->{'col'} = undef;
    $self->{'axis'}->{'x'}->{'grid'} = 0;
    $self->{'axis'}->{'x'}->{'inc'} = 0;
    $self->{'axis'}->{'x'}->{'min'} = 0;
    $self->{'axis'}->{'x'}->{'max'} = 0;
    $self->{'axis'}->{'x'}->{'threshold'} = 1;
    $self->{'axis'}->{'x'}->{'lab_size'} = 0;
    $self->{'axis'}->{'x'}->{'evenspace'} = 0;
    $self->{'axis'}->{'x'}->{'plot_labels'} = 1;
    $self->{'axis'}->{'x'}->{'plot_ticks'} = 1;
    $self->{'axis'}->{'x'}->{'plot_base'} = 1;
    $self->{'axis'}->{'x'}->{'labels'} = undef;
    $self->{'axis'}->{'x'}->{'scaling'} = \&_simple_trans;
    $self->{'axis'}->{'x'}->{'format'} = \&_simple_format;
    
    $self->{'axis'}->{'y'}->{'title'}->{'str'} = "";
    $self->{'axis'}->{'y'}->{'title'}->{'x'} = 0;
    $self->{'axis'}->{'y'}->{'title'}->{'y'} = 0;
    $self->{'axis'}->{'y'}->{'title'}->{'w'} = 0;
    $self->{'axis'}->{'y'}->{'title'}->{'h'} = 0;
    $self->{'axis'}->{'y'}->{'title'}->{'col'} = undef;
    $self->{'axis'}->{'y'}->{'inc'} = 0;
    $self->{'axis'}->{'y'}->{'min'} = 0;
    $self->{'axis'}->{'y'}->{'max'} = 0;
    $self->{'axis'}->{'y'}->{'threshold'} = 1;
    $self->{'axis'}->{'y'}->{'evenspace'} = 0;    
    $self->{'axis'}->{'y'}->{'lab_size'} = 0;
    $self->{'axis'}->{'y'}->{'plot_base'} = 1;
    $self->{'axis'}->{'y'}->{'grid'} = 0;
    $self->{'axis'}->{'y'}->{'plot_labels'} = 1;
    $self->{'axis'}->{'y'}->{'plot_ticks'} = 1;
    $self->{'axis'}->{'y'}->{'labels'} = undef;
    $self->{'axis'}->{'y'}->{'scaling'} = \&_simple_trans;
    $self->{'axis'}->{'y'}->{'format'} = \&_simple_format;
    
    $self->{'axis'}->{'x2'}->{'title'}->{'str'} = "";
    $self->{'axis'}->{'x2'}->{'title'}->{'x'} = 0;
    $self->{'axis'}->{'x2'}->{'title'}->{'y'} = 0;
    $self->{'axis'}->{'x2'}->{'title'}->{'w'} = 0;
    $self->{'axis'}->{'x2'}->{'title'}->{'h'} = 0;
    $self->{'axis'}->{'x2'}->{'title'}->{'col'} = undef;
    $self->{'axis'}->{'x2'}->{'plot_labels'} = 0;
    $self->{'axis'}->{'x2'}->{'plot_ticks'} = 0;
    $self->{'axis'}->{'x2'}->{'plot_base'} = 1;
    $self->{'axis'}->{'x2'}->{'inverse'} = 0;
    $self->{'axis'}->{'x2'}->{'inc'} = 0;
    $self->{'axis'}->{'x2'}->{'min'} = 0;
    $self->{'axis'}->{'x2'}->{'max'} = 0;
    $self->{'axis'}->{'x2'}->{'threshold'} = undef;
    $self->{'axis'}->{'x2'}->{'evenspace'} = 0;    
    $self->{'axis'}->{'x2'}->{'lab_size'} = 0;
    $self->{'axis'}->{'x2'}->{'grid'} = 0;
    $self->{'axis'}->{'x2'}->{'labels'} = undef;
    $self->{'axis'}->{'x2'}->{'scaling'} = \&_simple_trans;
    $self->{'axis'}->{'x2'}->{'format'} = \&_simple_format;

    $self->{'axis'}->{'y2'}->{'title'}->{'str'} = "";
    $self->{'axis'}->{'y2'}->{'title'}->{'x'} = 0;
    $self->{'axis'}->{'y2'}->{'title'}->{'y'} = 0;
    $self->{'axis'}->{'y2'}->{'title'}->{'w'} = 0;
    $self->{'axis'}->{'y2'}->{'title'}->{'h'} = 0;
    $self->{'axis'}->{'y2'}->{'title'}->{'col'} = undef;
    $self->{'axis'}->{'y2'}->{'inc'} = 0;
    $self->{'axis'}->{'y2'}->{'min'} = 0;
    $self->{'axis'}->{'y2'}->{'max'} = 0;
    $self->{'axis'}->{'y2'}->{'threshold'} = undef;
    $self->{'axis'}->{'y2'}->{'evenspace'} = 0;    
    $self->{'axis'}->{'y2'}->{'lab_size'} = 0;
    $self->{'axis'}->{'y2'}->{'plot_labels'} = 1;
    $self->{'axis'}->{'y2'}->{'plot_ticks'} = 1;
    $self->{'axis'}->{'y2'}->{'plot_base'} = 1;
    $self->{'axis'}->{'y2'}->{'grid'} = 0;
    $self->{'axis'}->{'y2'}->{'labels'} = undef;
    $self->{'axis'}->{'y2'}->{'scaling'} = \&_simple_trans;
    $self->{'axis'}->{'y2'}->{'format'} = \&_simple_format;

    $self->{'axis'}->{'tick_size'} = 5;
    
    # Plot area data
    $self->{'plot'}->{'x'} = 0;
    $self->{'plot'}->{'y'} = 0;
    $self->{'plot'}->{'h'} = 0;
    $self->{'plot'}->{'w'} = 0;
    $self->{'plot'}->{'space'} = 2;

    # Plot legends
    $self->{'legend'}->{'w'} = 0;
    $self->{'legend'}->{'h'} = 0;
    $self->{'legend'}->{'tl'} = 24;
    $self->{'legend'}->{'title'} = 'Legend:';
    @{$self->{'legend'}->{'entries'}} = ();
    $self->{'legend2'}->{'w'} = 0;
    $self->{'legend2'}->{'h'} = 0;
    $self->{'legend2'}->{'tl'} = 24;
    $self->{'legend2'}->{'title'} = 'Legend2:';
    @{$self->{'legend2'}->{'entries'}} = ();

    # Data to plot
    $self->{'gdata'}->{'nb_sets'} = 0;
    $self->{'gdata'}->{'colours'} = [];
    $self->{'gdata'}->{'names'}   = [];

    $self->{'gdata2'}->{'nb_sets'} = 0;
    $self->{'gdata2'}->{'colours'} = [];
    $self->{'gdata2'}->{'names'}   = [];
    
    $self->{'x_top'} = 5;
    $self->{'y_top'} = 5;
    $self->{'evenlyspace_method'} = \&_axis_evenlyspace;
    $self->{'end_imgmap'} = 1;
    $self->{'line_width'} = 1;
    $self->{'auto_colour'} = 0;
    $self->{'auto_color'} = 0;
    $self->{'border'} = 0;
    $self->{'transparent'} = $self->{'bg_colour'};
    
    return 1;
}

# ============================================================
# Function: set
# Passed:   hash table with argument-value entries
# Returns:  
# On error: 
# Description: 
# Simply parse the content of the hash table and assign the 
# name/value entries.
# Could be dangerous but it is the most efficient way to 
# set quickly parameters.
# ============================================================

=head2 set()

    $graph->set ('x_label_format' => \&x_format, 'auto_colour' => 1);

Curve_plot::set is used to assign specific parameters for the plot.
This method does not returns a value.
Possible values to set are:

=over 4

=item vertical:

Flag to set vertical drawing mode. By default, graphs are plotted
with the x axis horizontal and y axis vertical, reading left to right.
Setting this flag to a defined value causes the graph to be plotted
vertically. Namely, the horizontal graph is mirrored on a plot diagonal
axis, such that the x axis is vertical and the plot reads from top 
to bottom instead of left to right. Horizontal text becomes vertical,
vertical text, horizontal, but plotted in essentially the same place
in the graph as before. undef'ing this argument resets the graph
back to horizontal plotting.

If the graph is to be mirrored horizontally about a GIF origin 
which is not (0,0), then this origin should be set prior to,
or concurrently with the specification of the 'vertical' flag.

Default: undef (horizontal). Example: $graph->set ('vertical' => '1');

=item x_label_format:

Reference to a function used to format the labels on the X axis.
Default: None. Example: $graph->set ('x_label_format' => \&my_xformat);

with my_xformat defined for example as

sub my_xformat { my $val = shift; return sprintf "%3.2f" $val; }

=item y_label_format: 

Reference to a function used to format the labels on the Y axis.
Default: None. Example: $graph->set ('y_label_format' => \&my_yformat);

with my_yformat defined for example as

sub my_yformat { my $val = shift; return sprintf "%3.2f" $val; }

=item y2_label_format: 

Same as y_label_format but applied for the second Y axis.

=item x_scaling:

Reference to a function used to transform the labels AND values on the X axis.
Default: None. Example: $graph->set ('x_scaling' => \&my_xscaling);

with my_xscaling defined for example as

sub my_xscaling { my $val = shift; return ($val**2) * 100; }

=item x2_scaling:

Same as x_scaling but applied for the second X axis.

=item y_scaling:

Reference to a function used to transform the labels AND values on the Y axis.
Default: None. Example: $graph->set ('y_scaling' => \&my_yscaling);

with my_yscaling defined for example as

sub my_yscaling { my $val = shift; return ($val**2) * 100; }

=item x_off:

Number of pixels to detach the graph from the left side from the GIF area.
Default: 5. Example: $graph->set ('x_off' => '10');

=item y_off:

Number of pixels to detach the graph from the bottom side from the GIF area.
Default: 5. Example: $graph->set ('y_off' => '10');

=item x_top:

Number of pixels to detach the graph from the right side from the GIF area.
Default: 5. Example: $graph->set ('x_top' => '15');

=item y_top:

Number of pixels to detach the graph from the top side from the GIF area.
Default: 5. Example: $graph->set ('y_top' => '15');

=item font:

Sets the default GD font for text in the graph to one of 'tiny' (5 x 8), 
'small' (6 x 12), 'medium' (7 x 13) or 'giant' (9 x 15)
Default: 'medium'. Example: $graph->set ('font' => 'giant');

=item x_evenlyspace:

Boolean value (0/1) for evenly spacing of the X axis.
Default: 0. Example: $graph->set ('x_evenlyspace' => '1');

=item x_threshold:

Threshold value between 1 and 9 used to determine the density of X labels when evenly spacing.
Default: 1. Example: $graph->set ('x_threshold' => '3');

=item y_evenlyspace:

Boolean value (0/1) for evenly spacing of the Y axis.
Default: 0. Example: $graph->set ('y_evenlyspace' => '1');

=item y_threshold:

Threshold value between 1 and 9 used to determine the density of Y labels when evenly spacing.
Default: 1. Example: $graph->set ('y_threshold' => '3');

=item x2_evenlyspace:

Boolean value (0/1) for evenly spacing of the second X axis.
Default: 0. Example: $graph->set ('x2_evenlyspace' => '1');

=item x2_threshold:

Threshold value between 1 and 9 used to determine the density of second X labels when evenly spacing.
Default: 1. Example: $graph->set ('x2_threshold' => '3');

=item y2_evenlyspace:

Boolean value (0/1) for evenly spacing of the second Y axis.
Default: 0. Example: $graph->set ('y2_evenlyspace' => '1');

=item y2_threshold:

Threshold value between 0 and 1 used to determine the density of second Y labels when evenly spacing.
Default: 1. Example: $graph->set ('y2_threshold' => '3');

=item evenlyspace_method:

Reference to a function used for evenly space the axis in the case the internal method may fail.
Default: Internal. Example: $graph->set ('evenlyspace_method' => \&my_evenlyspace);

=item x_grid:

Can be a boolean value (0/1) or a string with RGB values separated by a comma for drawing the grid part on the X axis.
Default: 0. Example: $graph->set ('x_grid' => 1); or $graph->set ('x_grid' => '0,255,0');

=item x2_grid:

Same as 'x_grid' except that it is applied for the second X axis.

=item y_grid:

Can be a boolean value (0/1) or a string with RGB values separated by a comma for drawing the grid part on the Y axis.
Default: 0. Example: $graph->set ('y_grid' => 1); or $graph->set ('y_grid' => '0,255,0');

=item y2_grid:

Same as 'y_grid' except that it is applied for the second Y axis.

=item plot_xlabels

Set to 1 by default. If set to 0, the labels on the X axis will not be drawn.

=item plot_ylabels

Set to 1 by default. If set to 0, the labels on the Y axis will not be drawn.

=item plot_x2labels

Set to 1 by default. If set to 0, the labels on the second X axis will not be drawn.

=item plot_y2labels

Set to 1 by default. If set to 0, the labels on the second Y axis will not be drawn.

=item plot_xticks

Set to 1 by default. If set to 0, the ticks on the X axis will not be drawn.

=item plot_yticks

Set to 1 by default. If set to 0, the ticks on the Y axis will not be drawn.

=item plot_x2ticks

Set to 1 by default. If set to 0, the ticks on the second X axis will not be drawn.

=item plot_y2ticks

Set to 1 by default. If set to 0, the ticks on the second Y axis will not be drawn.


=item line_width:

Default size in pixels of the line width drawn on the plot, i.e. all curves with no 
specific line width defined (see 'dataset_define' method below). Note: it is not 
recommended to mix line types (dashed or dotted) with this option, the result is not very neat.
Default: 1. Example: $graph->set ('line_width' => 3);

=item auto_colour or auto_color:

Boolean value (0/1) used to tell the module to assign automatically a colour for 
each curve during their plotting. The colouring scheme starts at red and ends at blue.
The colour of a specific curve can be overridden using the dataset_define() method.
Default: 0. Example: $graph->set ('auto_color' => 1);

=item border:

Width size in pixel of the GIF border. 0 means no border and is the default value. Maximal value is 20.

=item legend:

Set the legend location. Can be 'top', 'bottom', 'left' and 'right', or a pairwise 
combination of two compatible specifications, e.g. 'top,left', where the first specification
is taken as the primary location, to which the second specification is added as an 
alignment adjustment.

Examples: $graph->set ('legend' => 'top');
          $graph->set ('legend' => 'top,left');

=item legend2:

Same at 'legend' except that it applies for the second set of data.
Example: $graph->set ('legend2' => 'bottom');

=item legend_title:

Define the legend title, default is 'Legend:'.
Example: $graph->set ('legend_title' => 'Data set1');

=item legend2_title:

Define the legend title for the second set of data, default is 'Legend2:'.
Example: $graph->set ('legend2_title' => 'Data set2');

=back

=cut

sub set {
    my $self = shift;
    my %args = @_;
    my $key;
    
    # Need to parse the parameters
    foreach $key (keys %args) {
    	# Vertical drawing mode? maybe 'undef'
	if ($key eq "vertical") {
	    $self->{'vertical'}  = $args{$key};
	    my $x = $args{'x'}; $x = $self->{'x'} if !defined($x);
	    my $y = $args{'y'}; $y = $self->{'y'} if !defined($y);
	    $self->{'transform'} = _transformation_matrix($args{$key},$x,$y); }
    	# Grid flags
	elsif ($key eq "x_grid") {
	    $self->{'axis'}->{'x'}->{'grid'} = $args{$key}; }
	elsif ($key eq "y_grid") {
	    $self->{'axis'}->{'y'}->{'grid'} = $args{$key}; }
	elsif ($key eq "x2_grid") {
	    $self->{'axis'}->{'x2'}->{'grid'} = $args{$key}; }
	elsif ($key eq "y2_grid") {
	    $self->{'axis'}->{'y2'}->{'grid'} = $args{$key}; }
	# Label flags
	elsif ($key eq "plot_xlabels") {
	    $self->{'axis'}->{'x'}->{'plot_labels'} = $args{$key}; }
	elsif ($key eq "plot_ylabels") {
	    $self->{'axis'}->{'y'}->{'plot_labels'} = $args{$key}; }
	elsif ($key eq "plot_x2labels") {
	    $self->{'axis'}->{'x2'}->{'plot_labels'} = $args{$key}; }
	elsif ($key eq "plot_y2labels") {
	    $self->{'axis'}->{'y2'}->{'plot_labels'} = $args{$key}; }
	# Tick flags
	elsif ($key eq "plot_xticks") {
	    $self->{'axis'}->{'x'}->{'plot_ticks'} = $args{$key}; }
	elsif ($key eq "plot_yticks") {
	    $self->{'axis'}->{'y'}->{'plot_ticks'} = $args{$key}; }
	elsif ($key eq "plot_x2ticks") {
	    $self->{'axis'}->{'x2'}->{'plot_ticks'} = $args{$key}; }
	elsif ($key eq "plot_y2ticks") {
	    $self->{'axis'}->{'y2'}->{'plot_ticks'} = $args{$key}; }
	# Axis base line flag
	elsif ($key eq "plot_xbase") {
	    $self->{'axis'}->{'x'}->{'plot_base'} = $args{$key};
	    $self->{'axis'}->{'x'}->{'plot_ticks'} = $args{$key};
	    }
	elsif ($key eq "plot_ybase") {
	    $self->{'axis'}->{'y'}->{'plot_base'} = $args{$key};
	    $self->{'axis'}->{'y'}->{'plot_ticks'} = $args{$key};
	    }
	elsif ($key eq "plot_x2base") {
	    $self->{'axis'}->{'x2'}->{'plot_base'} = $args{$key};
	    $self->{'axis'}->{'x2'}->{'plot_ticks'} = $args{$key};
	    }
	elsif ($key eq "plot_y2base") {
	    $self->{'axis'}->{'y2'}->{'plot_base'} = $args{$key};
	    $self->{'axis'}->{'y2'}->{'plot_ticks'} = $args{$key};
	    }
    	# Label formats
    	elsif ($key eq "x_label_format") {
    	    $self->{'axis'}->{'x'}->{'format'} = $args{$key}; }
    	elsif ($key eq "y_label_format") {
    	    $self->{'axis'}->{'y'}->{'format'} = $args{$key}; }
    	elsif ($key eq "x2_label_format") {
    	    $self->{'axis'}->{'x2'}->{'format'} = $args{$key}; }
	elsif ($key eq "y2_label_format") {
    	    $self->{'axis'}->{'y2'}->{'format'} = $args{$key}; }
	# Axis scaling
	elsif ($key eq "x_scaling") {
	    $self->{'axis'}->{'x'}->{'scaling'} = $args{$key}; }
	elsif ($key eq "y_scaling") {
	    $self->{'axis'}->{'y'}->{'scaling'} = $args{$key}; }
	elsif ($key eq "x2_scaling") {
	    $self->{'axis'}->{'x2'}->{'scaling'} = $args{$key}; }
	elsif ($key eq "y2_scaling") {
	    $self->{'axis'}->{'y2'}->{'scaling'} = $args{$key}; }
	# Scale thresholds
	elsif ($key eq "x_threshold") {
	    $self->{'axis'}->{'x'}->{'threshold'} = $args{$key}; }
	elsif ($key eq "y_threshold") {
	    $self->{'axis'}->{'y'}->{'threshold'} = $args{$key}; }
	elsif ($key eq "x2_threshold") {
	    $self->{'axis'}->{'x2'}->{'threshold'} = $args{$key}; }
	elsif ($key eq "y2_threshold") {
	    $self->{'axis'}->{'y2'}->{'threshold'} = $args{$key}; }
	# Evenly spaces
	elsif ($key eq "evenlyspace_method") {
	    $self->{'evenlyspace_method'} = $args{$key}; }
	elsif ($key eq "x_evenlyspace") {
	    $self->{'axis'}->{'x'}->{'evenspace'} = $args{$key}; }
	elsif ($key eq "y_evenlyspace") {
	    $self->{'axis'}->{'y'}->{'evenspace'} = $args{$key}; }
	elsif ($key eq "x2_evenlyspace") {
	    $self->{'axis'}->{'x2'}->{'evenspace'} = $args{$key}; }
	elsif ($key eq "y2_evenlyspace") {
	    $self->{'axis'}->{'y2'}->{'evenspace'} = $args{$key}; }
	# Graph offsets
	elsif ($key eq "x") {
	    $self->{'x'} = $args{$key}; }
	elsif ($key eq "y") {
	    $self->{'y'} = $args{$key}; }
	elsif ($key eq "x_off") {
	    $self->{'x_off'} = $args{$key}; }
	elsif ($key eq "y_off") {
	    $self->{'y_off'} = $args{$key}; }
	elsif ($key eq "x_top") {
	    $self->{'x_top'} = $args{$key}; }
	elsif ($key eq "y_top") {
	    $self->{'y_top'} = $args{$key}; }
	# General plot styles
	elsif ($key eq "line_width") {
	    $self->{'line_width'} = $args{$key}; }
	elsif ($key eq "auto_colour") {
	    $self->{'auto_colour'} = $args{$key}; }
	elsif ($key eq "auto_color") {
	    $self->{'auto_color'} = $args{$key}; }
	elsif ($key eq "border") {
	    $self->{'border'} = $args{$key}; }
	# Legends
	elsif ($key eq "legend") {
	    $self->{'legend_loc'} = $args{$key}; }
	elsif ($key eq "legend2") {
	    $self->{'legend2_loc'} = $args{$key}; }
	elsif ($key eq "legend_title") {
	    $self->{'legend'}->{'title'} = $args{$key}; }
	elsif ($key eq "legend2_title") {
	    $self->{'legend2'}->{'title'} = $args{$key}; }
    	# Transparency
	elsif ($key eq "transparent") {
	    $self->{'transparent'} = $args{$key}; }
	# Plot font
	elsif ($key eq "font") {
	    if($args{'font'} eq 'giant'){
		$self->{'font'} = gdGiantFont ; # 9x15
		$self->{'font_width'}  = 9;
		$self->{'font_height'} = 15;
	    }elsif($args{'font'} eq 'medium'){
		$self->{'font'} = gdMediumBoldFont ; # 7x13
		$self->{'font_width'}  = 7 ;
		$self->{'font_height'} = 13 ;
	    }elsif($args{'font'} eq 'small'){
		$self->{'font'} = gdSmallFont ; # 6x12
		$self->{'font_width'}  = 6 ;
		$self->{'font_height'} = 12 ;
	    }elsif($args{'font'} eq 'tiny'){
		$self->{'font'} = gdTinyFont ; # 5x8
		$self->{'font_width'}  = 5 ;
		$self->{'font_height'} = 8 ;
	    }
	}
    }
}

# ============================================================
# Function: title
# Passed:   Title of the graph, optionally its colour.
# Returns:  1
# On error: 0
# Description: 
# Get the graph title and assign it to the plot. 
# ============================================================

=head2 title()

    $graph->title ("Interest rates", (234,34,67));

Curve_plot::title is used to assign a title to the plot. The second argument can be passed as an array containing the R,G,B values for the title colour.
This method is optional.
It returns 1 on success and 0 on failure.

=cut

sub title {
    my $self = shift;
    my $title = shift;
    my @col = @_;
    
    return 0 unless (defined $title);

    $self->{'title'}->{'str'} = $title;
    $self->{'title'}->{'h'} = $self->{'font_height'} + $self->{'plot'}->{'space'};
    $self->{'title'}->{'w'} = length ($title) * $self->{'font_width'};
    
    # Define title colour
    $self->{'title'}->{'col'} = $self->{'gd'}->colorAllocate(@col) if (defined @col);

    # Set default if problems
    $self->{'title'}->{'col'} = $self->{'fg_colour'} unless defined $self->{'title'}->{'col'};
    
    return 1;
}

# ============================================================
# Function: _draw_title
# Passed:   optional x and y position
# Returns:  1
# On error: 0
# Description: 
# Draw on the gif the plot title. If no x or y passed, calculated
# from the plot dimensions.
# ============================================================

sub _draw_title {
    my $self = shift;
    
    return 0 if ($self->{'title'}->{'str'} eq "");
    
    # Define x and y if needed
    $self->{'title'}->{'x'} = $self->{'x'} + int (($self->{'width'} / 2) - ($self->{'title'}->{'w'} / 2));

    $self->{'title'}->{'y'} = $self->{'y'} + $self->{'top_corr'};

    if(defined($self->{'vertical'})) {
	$self->{'gd'}->stringUp ($self->{'font'}, 
			       &{$self->{'transform'}}($self->{'title'}->{'x'}, $self->{'title'}->{'y'}),
			       $self->{'title'}->{'str'}, $self->{'title'}->{'col'});
    } else {
	$self->{'gd'}->string ($self->{'font'}, 
			       &{$self->{'transform'}}($self->{'title'}->{'x'}, $self->{'title'}->{'y'}),
			       $self->{'title'}->{'str'}, $self->{'title'}->{'col'});
    }
    return 1;
}

sub _set_xtitle {
    my $self = shift;
    my $ref = shift;
    my $title = shift;
    my $col = shift;
    
    return 0 unless (defined $title);
    
    $ref->{'str'} = $title;
    $ref->{'w'} = length ($title) * $self->{'font_width'};
    $ref->{'h'} = $self->{'font_height'} + $self->{'plot'}->{'space'};
    
    # Define title colour
    $ref->{'col'} = $self->{'gd'}->colorAllocate(@{$col}) if (defined @{$col});

    # Set default if problems
    $ref->{'col'} = $self->{'fg_colour'} unless defined $ref->{'col'}; 

    return 1;
}


=head2 x_title()

    $graph->x_title ("Interest", (0,0,255));

Curve_plot::x_title is used to assign a title to the X axis. The title will be written horizontally. The second argument can be passed as an array containing the R,G,B values for the X title colour.
This method is optional.
It returns 1 on success and 0 on failure.

=cut

sub x_title {
    my $self = shift;
    my $label = shift;
    my @col = @_;

    return $self->_set_xtitle ($self->{'axis'}->{'x'}->{'title'}, $label, \@col);
}

=head2 x2_title()

    $graph->x2_title ("Interest", (0,0,255));

Curve_plot::x2_title does the same as 'x_title' except that it is used for the X labels related to the second Y axis.

It returns 1 on success and 0 on failure.

=cut

sub x2_title {
    my $self = shift;
    my $label = shift;
    my @col = @_;

    return $self->_set_xtitle ($self->{'axis'}->{'x2'}->{'title'}, $label, \@col);
}

# ============================================================
# Function: _draw_x_title
# Passed:   optional x and y positions
# Returns:  1
# On error: 0
# Description: 
# Draw on the gif plot the X axis title, 
# calculates x and y if needed.
# ============================================================

sub _draw_x_titles {
    my $self = shift;
        
    if ($self->{'axis'}->{'x'}->{'title'}->{'str'} ne "") {
	
	# Draw main X axis title
	my $ref = $self->{'axis'}->{'x'}->{'title'};

	$ref->{'x'} = $self->{'plot'}->{'x'} + (($self->{'plot'}->{'w'} - $ref->{'w'}) / 2);
	$ref->{'y'} = $self->{'plot'}->{'y'} + $self->{'plot'}->{'space'};
    	
	$ref->{'y'} += $self->{'axis'}->{'x'}->{'lab_size'} if ($self->{'axis'}->{'x'}->{'plot_labels'} == 1);
	$ref->{'y'} += $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'x'}->{'plot_ticks'} == 1);
	
	# Draw title on the plot now	
	if(defined($self->{'vertical'})) {
	    $self->{'gd'}->stringUp($self->{'font'}, 
				  &{$self->{'transform'}}($ref->{'x'}, $ref->{'y'}), 
				  $ref->{'str'},$ref->{'col'});
	} else {
	    $self->{'gd'}->string($self->{'font'}, 
				  &{$self->{'transform'}}($ref->{'x'}, $ref->{'y'}), 
				  $ref->{'str'},$ref->{'col'});
	}
    }
	
    # Draw second X axis title if defined
    if ($self->{'axis'}->{'x2'}->{'title'}->{'str'} ne "") {
	my $ref = $self->{'axis'}->{'x2'}->{'title'};
	
	$ref->{'x'} = $self->{'plot'}->{'x'} + (($self->{'plot'}->{'w'} - $ref->{'w'}) / 2);
	$ref->{'y'} = $self->{'plot'}->{'y'} - $self->{'plot'}->{'h'} - $ref->{'h'} - $self->{'plot'}->{'space'};
    	
	$ref->{'y'} -= $self->{'axis'}->{'x2'}->{'lab_size'} if ($self->{'axis'}->{'x2'}->{'plot_labels'} == 1);
	$ref->{'y'} -= $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'x2'}->{'plot_ticks'} == 1);
	
	# Draw title on the plot now
	if(defined($self->{'vertical'})) {
	    $self->{'gd'}->stringUp($self->{'font'}, 
				  &{$self->{'transform'}}($ref->{'x'}, $ref->{'y'}), 
				  $ref->{'str'},$ref->{'col'});
	} else {
	    $self->{'gd'}->string($self->{'font'}, 
				  &{$self->{'transform'}}($ref->{'x'}, $ref->{'y'}), 
				  $ref->{'str'},$ref->{'col'});
    	}
    }
    
    return 1;
}

sub _set_ytitle {
    my $self = shift;
    my $ref = shift;
    my $title = shift;
    my $col = shift;
    
    return 0 unless (defined $title);
    
    $ref->{'str'} = $title;
    $ref->{'w'} = $self->{'font_height'} + $self->{'plot'}->{'space'};
    $ref->{'h'} = length ($title) * $self->{'font_width'};
    
    # Define title colour
    $ref->{'col'} = $self->{'gd'}->colorAllocate(@$col) if (defined @$col);

    # Set default if problems
    $ref->{'col'} = $self->{'fg_colour'} unless defined $ref->{'col'}; 

    return 1;
}

=head2 y_title()

    $graph->y_title ("Rate in %", (255,0,0));

Curve_plot::y_title is used to assign a title to the Y axis. The title will be written vertically. The second argument can be passed as an array containing the R,G,B values for the X title colour.
This method is optional.
It returns 1 on success and 0 on failure.

=cut

# For editor colouring: "

sub y_title {
    my $self = shift;
    my $label = shift;
    my @col = @_;
    
    return 0 unless (defined $label);
    return $self->_set_ytitle ($self->{'axis'}->{'y'}->{'title'}, $label, \@col);
}

=head2 y2_title()

    $graph->y2_title ("Rate in %", (255,0,0));

Curve_plot::y2_title does the same as 'y_title' except that it is used for the Y labels related to the second Y axis.

=cut
# For editor colouring: "

sub y2_title {
    my $self = shift;
    my $label = shift;
    my @col = @_;
    
    return 0 unless (defined $label);
    return $self->_set_ytitle ($self->{'axis'}->{'y2'}->{'title'}, $label, \@col);
}

# ============================================================
# Function: _draw_y_title
# Passed:   optional x and y positions
# Returns:  1
# On error: 0
# Description: 
# Draw on the gif plot the Y axis title, 
# calculates x and y if needed.
# ============================================================

sub _draw_y_titles {
    my $self = shift;

    # Draw main Y axis title
    if ($self->{'axis'}->{'y'}->{'title'}->{'str'} ne "") {
	my $ref = $self->{'axis'}->{'y'}->{'title'};

	$ref->{'x'} = $self->{'plot'}->{'x'} - $ref->{'w'} - $self->{'plot'}->{'space'};
	$ref->{'x'} -= $self->{'axis'}->{'y'}->{'lab_size'} if ($self->{'axis'}->{'y'}->{'plot_labels'} == 1);
	$ref->{'x'} -= $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'y'}->{'plot_ticks'} == 1);
	
	$ref->{'y'} = $self->{'plot'}->{'y'} - (($self->{'plot'}->{'h'} - $ref->{'h'}) / 2);

	# Draw title on the plot now
	if(defined($self->{'vertical'})) {
	    $self->{'gd'}->string($self->{'font'}, 
				    &{$self->{'transform'}}($ref->{'x'}, 
							    $ref->{'y'}-length($ref->{'str'})*$self->{'font_width'}), 
				    $ref->{'str'},$ref->{'col'});
	} else {
	    $self->{'gd'}->stringUp($self->{'font'}, 
				    &{$self->{'transform'}}($ref->{'x'}, $ref->{'y'}), 
				    $ref->{'str'},$ref->{'col'});
    	}
    }
	
    # Draw second Y axis title if defined
    if ($self->{'axis'}->{'y2'}->{'title'}->{'str'} ne "") {
	my $ref = $self->{'axis'}->{'y2'}->{'title'};
	
	#$x = $self->{'left_corr'} + ($self->{'width'} / 2) - ($ref->{'w'} / 2);

	$ref->{'x'} = $self->{'plot'}->{'x'} + $self->{'plot'}->{'w'} + $self->{'plot'}->{'space'};
	$ref->{'x'} += $self->{'axis'}->{'y2'}->{'lab_size'} if ($self->{'axis'}->{'y2'}->{'plot_labels'} == 1);
	$ref->{'x'} += $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'y2'}->{'plot_ticks'} == 1);

	$ref->{'y'} = $self->{'plot'}->{'y'} - (($self->{'plot'}->{'h'} - $ref->{'h'}) / 2);
    	
	# Draw title on the plot now
	if(defined($self->{'vertical'})) {
	    $self->{'gd'}->string($self->{'font'}, 
				    &{$self->{'transform'}}($ref->{'x'}, 
							    $ref->{'y'}-length($ref->{'str'})*$self->{'font_width'}), 
				    $ref->{'str'},$ref->{'col'});
	} else {
	    $self->{'gd'}->stringUp($self->{'font'}, 
				  &{$self->{'transform'}}($ref->{'x'}, $ref->{'y'}), 
				  $ref->{'str'},$ref->{'col'});
    	}
    }
    return 1;
}

sub _draw_legend {
    my $self = shift;
    my $legend = shift;
    my $loc = shift;
    my $dataref = shift;
        
    # Get pos type, vertical by default
    my $lmode = 0;
    my $w_limit = 0;
    my $h_limit = 0;
    if ($loc =~ /(top|bottom)/) {
    	# Set horizontal mode
    	$lmode = 1;
	# Get width limit
	$w_limit = $self->{'width'} - $self->{'right_corr'} - $self->{'left_corr'} - 5;
	}
    else {
    	$h_limit = $self->{'height'} - $self->{'top_corr'} - $self->{'bottom_corr'} - 5;
	}

    my $x = $self->{'x'};
    my $y = $self->{'y'};
    
    # Richard's modif for corner locations support
    # Define starting point
    if ($loc =~ /^left/) {
	$x += $self->{'left_corr'};
	if($loc =~ /top$/) {
	    $y += $self->{'top_corr'}; # left,top alignment
	} elsif($loc =~ /bottom$/) {
	    $y += $self->{'height'} - $self->{'bottom_corr'} - $legend->{'lh'};# left,bottom
	} else {
	    $y += int (($self->{'height'} - $legend->{'lh'}) / 2); # centred
        }
    } elsif ($loc =~ /^right/) {
    	$x += $self->{'width'} - $self->{'right_corr'} - $legend->{'lw'};
	if($loc =~ /top$/) {
	    $y += $self->{'top_corr'} ; # right,top alignment
	} elsif($loc =~ /bottom$/) {
	    $y += $self->{'height'} - $self->{'bottom_corr'} - $legend->{'lh'} ;# right,bottom
	} else {
	    $y += int (($self->{'height'} - $legend->{'lh'}) / 2); # centred
	}
    } elsif ($loc =~ /^top/) {
	if($loc =~ /left$/) {
	    $x += $self->{'left_corr'}; # top,left hand corner?
	} elsif($loc =~ /right$/) {
	    $x += $self->{'width'} - $self->{'right_corr'} - $legend->{'lw'} ; # top,right
	} else {
	    $x += int (($self->{'width'} - $legend->{'lw'}) / 2); # centred
	}
	$y += $self->{'top_corr'} + $self->{'title'}->{'h'};
    } elsif ($loc =~ /^bottom/) {
	if($loc =~ /left$/) {
	    $x += $self->{'left_corr'}; # bottom,left
	} elsif($loc =~ /right$/) {
	    $x += $self->{'width'} - $self->{'right_corr'} - $legend->{'lw'} ; # bottom,right
	} else {
	    $x += int (($self->{'width'} - $legend->{'lw'}) / 2); # centred
	}
	$y += $self->{'height'} - $self->{'bottom_corr'} - $legend->{'lh'};
    } else {
    	confess ("No legend location specified!\n"); }
	  
    my $lx = $x + 3;
    my $ly = $y + 2;
    
    my $h = 0;
    my $w = 0;
    my $px;
    my $py;
    my $max_w = 0;
    my $cref = 0;
    my $e_pos = 0;
    
    foreach my $entry (@{$legend->{'entries'}}) {
    	unless (defined $entry) {
	    $cref++;
	    next;
	    }

	my $len = length $entry;
	my $lw = $len * $self->{'font_width'};
    	$max_w = $lw if ($lw > $max_w);
	
	# If horizontal legend
	if ($lmode == 1) {
	    # If out of limits
	    if ($w + $lw + 5 + $legend->{'tl'} > $w_limit) {
	    	# Pass to next line
	    	$ly += $self->{'font_height'} + 2;
		$w = 0;
		}
	    $px = $lx + $w;
	    $py = $ly;
	    $w += $lw + 5 + $legend->{'tl'};
	    }
        else {
	    # If out of limits
	    if ($h + $self->{'font_height'} + 2 > $h_limit) {
	    	# Pass to next column
	    	$lx += $max_w + 5 + $legend->{'tl'};
		}
	    $px = $lx;
	    $py = $ly + $h;
	    $h += $self->{'font_height'} + 2;
            }
		
	# Now draw legend
	if ($e_pos == 0) {
	    # "Legend:" tag
	    if(defined($self->{'vertical'})) {
		$self->{'gd'}->stringUp ($self->{'font'}, 
				       &{$self->{'transform'}}($px+length($entry)*$self->{'font_width'}, $py),
				       $entry, $self->{'fg_colour'});
	    } else {
		$self->{'gd'}->string ($self->{'font'}, 
				       &{$self->{'transform'}}($px, $py),
				       $entry, $self->{'fg_colour'});
    	    }
	}
	else {
	    # Curve legends
	    if(defined($self->{'vertical'})) {
		$self->{'gd'}->stringUp ($self->{'font'}, 
				       &{$self->{'transform'}}($px+length($entry)*$self->{'font_width'}, $py),
				       $entry, $dataref->{'colours'}->[$cref]);
	    } else {
		$self->{'gd'}->string ($self->{'font'}, 
				       &{$self->{'transform'}}($px, $py),
				       $entry, $dataref->{'colours'}->[$cref]);
	    }	    
	    # Draw tick corresponding to curve
	    my $xline = $px + $lw + 1;
	    
#	    $self->{'gd'}->line (&{$self->{'transform'}}($xline, $py + 6), 
#	                         &{$self->{'transform'}}($xline + 10, $py + 6), 
#	                         $dataref->{'colours'}->[$cref]);
	    
	    $self->_legend_tick ($dataref->{'sets'}->{$dataref->{'names'}->[$cref]},
	    	    	    	 $xline, $py + 6, $dataref->{'colours'}->[$cref],
				 $legend->{'tl'});
	    $cref++;
	    }
    	$e_pos++;
	}
    # Now draw legend borders
    $lx += $max_w;
    $ly += $self->{'font_height'} + 2;

    $self->{'gd'}->rectangle(&{$self->{'transform'}}($x,$y),
			     &{$self->{'transform'}}($x+$legend->{'lw'}-1, $y+$legend->{'lh'}-2)
			     ,$self->{'fg_colour'});
    
    return 1;
}

sub _legend_tick {
    my $self = shift;
    my $data_set = shift;
    my $x = shift;
    my $y = shift;
    my $col = shift;
    my $tl = shift;
        
    # If defined line style
    if (defined $data_set->{'line_style'}) {
	# Get style
	if ($data_set->{'line_style'} == PLAIN_LINE) {
	    $self->{'gd'}->setStyle($col); }
	elsif ($data_set->{'line_style'} == DOTTED_LINE) {
	    $self->{'gd'}->setStyle($col,$col,$col,gdTransparent,gdTransparent); }
	elsif ($data_set->{'line_style'} == DASHED_LINE) {
	    $self->{'gd'}->setStyle($col,$col,$col,$col,$col,$col,$col,$col,
	    	    	    	    $col,$col,gdTransparent,gdTransparent,gdTransparent); }

	# Define line width, including user defined for this curve
	my $lwidth = $self->{'line_width'};
	$lwidth = $data_set->{'line_width'} if defined $data_set->{'line_width'};
	$lwidth = 1 if ($lwidth < 1 || $lwidth > $self->{'font_height'});
		
	my $ny = $y - $lwidth + 1;
	for (my $pos = 0; $pos < $lwidth; $pos++) {
	    $self->{'gd'}->line (&{$self->{'transform'}}($x, $ny), 
				 &{$self->{'transform'}}($x + 24, $ny), 
				 gdStyled);
	    $ny++;
	    }
    	}
    if (defined $data_set->{'dot_style'}) {
    	$self->{'gd'}->setStyle($col);
	# Shift x to center of tick
	$x += 12;
	
	# If dot block style
	if ($data_set->{'dot_style'} == SQUARE_DOT) {
	    $self->{'gd'}->filledRectangle(&{$self->{'transform'}}($x-2,$y-2),
					   &{$self->{'transform'}}($x+2,$y+2),gdStyled); }
	# If dot X style
	elsif ($data_set->{'dot_style'} == CROSS_DOT) {
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x,$y),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y+1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y+1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y+2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y+2),gdStyled);
	    }
	# Dot O style
	elsif ($data_set->{'dot_style'} == CIRC_DOT) {
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y+1),gdStyled);
	    
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y+1),gdStyled);

	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y-2),gdStyled);
	    
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y+2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y+2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y+2),gdStyled);
    	    }
	# Dot + style
	elsif ($data_set->{'dot_style'} == PLUS_DOT) {
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y)  ,gdStyled);

	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y+1),gdStyled);
	    
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y+2),gdStyled);
	    }
	# Dot . style
	elsif ($data_set->{'dot_style'} == POINT_DOT) {
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x,$y),gdStyled);
	    }
    	}
}

=head2 x_offset()

    $graph->x_offset (20);

Curve_plot::x_offset is used to define in pixel the distance between the left part of the plot and the left border of the GIF file.
This method does not returns a value.

=cut

sub x_offset {
    my $self = shift;
    my $offset = shift;

    if (defined $offset) {
	$self->{'x_off'} = $offset; }
    else { $self->{'x_off'} = 5; }
}

=head2 y_offset()

    $graph->y_offset (20);

Curve_plot::y_offset is used to define in pixel the distance between the bottom part of the plot and the bottom border of the GIF file.
This method does not returns a value.

=cut

sub y_offset {
    my $self = shift;
    my $offset = shift;

    if (defined $offset) {
	$self->{'y_off'} = $offset; }
    else { $self->{'y_off'} = 5; }
}

sub _draw_border {
    my $self = shift;
    
    # Draw plot border
    $self->{'border'} = 10 if ($self->{'border'} > 10);
    $self->{'border'} = 0 if ($self->{'border'} < 0);
    my $w = $self->{'x'} + $self->{'width'};
    my $h = $self->{'y'} + $self->{'height'};
    for (my $pos = 0; $pos < $self->{'border'}; $pos++) {
	$self->{'gd'}->rectangle(&{$self->{'transform'}}($self->{'x'} + $pos, $self->{'y'} + $pos), 
				 &{$self->{'transform'}}($w - ($pos+1), $h - ($pos+1)),
				 $self->{'fg_colour'});
    	}
}

sub _set_axis {
    my $self = shift;
    my $ref = shift;
    my $labels = shift;
    my $scaling = shift;

    # Some checks
    return 0 unless (defined $labels);
    return 0 if (ref ($labels) ne "ARRAY");

    # Keep scaling function
    if (defined $scaling) {
	return 0 if (ref ($scaling) ne "CODE");
	$ref->{'scaling'} = $scaling;
	}

    # If passing labels, update graph offsets
    $ref->{'labels'} = $labels;

    return 1;
}

=head2 x_axis()

    $graph->x_axis (\@x_labels, \&x_scaling);

Curve_plot::x_axis is used to assign the values for the X axis. The optional second argument must be a reference to a function used to scale the X values.
When drawing the graph, the scaling function will be called on each X values, however the original values will be displayed. If not defined, it will be derived from the X axis values passed with the add_dataset method (see below). If no X axis are defined at all, the evenly spaced method will be called to generate a standard X axis based on the number of entries present in the datasets.
It returns 1 on success and 0 on failure.

=cut

sub x_axis {
    my $self = shift;
    my $labels = shift;
    my $scaling = shift;
    
    return $self->_set_axis ($self->{'axis'}->{'x'}, $labels, $scaling);
}

=head2 x2_axis()

    $graph->x2_axis (\@x_labels, \&x_scaling);

Curve_plot::x2_axis does the same as 'x_axis' expect that it is applied for the X axis related to the second Y axis.

It returns 1 on success and 0 on failure.

=cut

sub x2_axis {
    my $self = shift;
    my $labels = shift;
    my $scaling = shift;
    
    return $self->_set_axis ($self->{'axis'}->{'x2'}, $labels, $scaling);
}

=head2 y_axis()

    $graph->y_axis (\@y_labels, \&y_scaling);

Curve_plot::y_axis is used to assign the values for the Y axis. The optional second argument must be a reference to a function used to scale the Y values.
    When drawing the graph, the scaling function will be called on each Y values, however the original values will be displayed. If not defined, the evenly space method will be called to generate an Y axis based on the values present in the defined datasets. The Y labels are right justified to the Y axis.
It returns 1 on success and 0 on failure.

=cut

sub y_axis {
    my $self = shift;
    my $labels = shift;
    my $scaling = shift;
    
    return $self->_set_axis ($self->{'axis'}->{'y'}, $labels, $scaling);
}

=head2 y2_axis()

    $graph->y2_axis (\@y2_labels, \&y2_scaling);

Curve_plot::y2_axis does the same as 'y_axis' except that it applies for the second Y axis on the right side of the combined graph.

It returns 1 on success and 0 on failure.

=cut

sub y2_axis {
    my $self = shift;
    my $labels = shift;
    my $scaling = shift;
    
    return $self->_set_axis ($self->{'axis'}->{'y2'}, $labels, $scaling);
}

sub _draw_x_labels {
    my $self = shift;
    
    my $xfont_corr = ($self->{'font_height'} / 2);
    	
    # If need to plot X labels
    if (defined $self->{'axis'}->{'x'}->{'labels'}) {
    	my $ref = $self->{'axis'}->{'x'};
	# Keep refs of data and scale function
	my $labels = $ref->{'labels'};
	my $scale  = $ref->{'scaling'};
	my $format = $ref->{'format'};
    	
	# If need to plot X grid?
	my $grd_col = undef;
	if (defined $ref->{'grid'}) {
	    if ($ref->{'grid'} =~ /(\d+),(\d+),(\d+)/) {
	    	$grd_col = $self->{'gd'}->colorAllocate($1, $2, $3); }
	    elsif ($ref->{'grid'} == 1) {
	    	$grd_col = $self->{'fg_colour'}; }
	    }

	# Define base on Y axis
	my $y_base =  $self->{'plot'}->{'y'};

	# Define Y position to draw labels as Y plot base line + label size + tick
	my $y = $self->{'plot'}->{'y'} + $ref->{'lab_size'} + 6;

	# Define increment value
	my $inc = $ref->{'inc'};
	my $min = $ref->{'min'};
	my $col;

	if (defined $ref->{'title'}->{'col'}) {
    	    $col = $ref->{'title'}->{'col'}; }
	else {
    	    $col = $self->{'fg_colour'}; }

	my $x_base = $self->{'plot'}->{'x'};

	# Loop on labels
	my $len = @{$labels};

	for (my $pos = 0; $pos < $len; $pos++) {
	    next if (! defined $labels->[$pos]);

	    # Draw label
	    my $val = $scale->($labels->[$pos]);
	    my $x = $x_base + (($val - $min) * $inc);
	    if ($self->{'axis'}->{'x'}->{'plot_labels'} == 1) {
    		my $xstr = $x - $xfont_corr;

		# Format label
		my $label = $format->($labels->[$pos]);

		# Draw label
		if(defined($self->{'vertical'})) {
		    $self->{'gd'}->string($self->{'font'}, 
					  &{$self->{'transform'}}($xstr,$y-length($label)*$self->{'font_width'}), 
					  $label, $col);
		} else {
		    $self->{'gd'}->stringUp($self->{'font'}, 
					    &{$self->{'transform'}}($xstr, $y), 
					    $label, $col);
		}
	    }
		
	    # Draw tick
	    $self->{'gd'}->line (&{$self->{'transform'}}($x, $y_base), 
				 &{$self->{'transform'}}($x, $y_base + 5), 
				 $col) if ($self->{'axis'}->{'x'}->{'plot_ticks'} == 1);

	    # Need X grid?
	    if (defined $grd_col) {
		$self->{'gd'}->line (&{$self->{'transform'}}($x, $y_base), 
				     &{$self->{'transform'}}($x, $y_base-$self->{'plot'}->{'h'}), 
				     $grd_col);
		}
    	    }
    	}
	
    if ($self->{'gdata2'}->{'nb_sets'} > 0 &&
    	defined $self->{'axis'}->{'x2'}->{'labels'}) {
    	
	my $ref = $self->{'axis'}->{'x2'};
	# Keep refs of data and scale function
	my $labels = $ref->{'labels'};
	my $scale = $ref->{'scaling'};
	my $format = $ref->{'format'};

	# If need to plot X grid?
	my $grd_col = undef;
	if (defined $ref->{'grid'}) {
	    if ($ref->{'grid'} =~ /(\d+),(\d+),(\d+)/) {
	    	$grd_col = $self->{'gd'}->colorAllocate($1, $2, $3); }
	    elsif ($ref->{'grid'} == 1) {
	    	$grd_col = $self->{'fg_colour'}; }
	    }

	# Define base on Y axis
	my $y_base = $self->{'plot'}->{'y'} - $self->{'plot'}->{'h'};

	# Define 1st pos on Y axis to draw label
	my $y = $y_base - 6;
    	
	# Define increment value
	my $inc = $ref->{'inc'};
	my $min = $ref->{'min'};
	my $col;

	if (defined $ref->{'title'}->{'col'}) {
    	    $col = $ref->{'title'}->{'col'}; }
	else {
    	    $col = $self->{'fg_colour'}; }

	my $x_base = $self->{'plot'}->{'x'};

	# Loop on labels
	my $len = @{$labels};
	my $ele1;
	my $ele2;
	my $val;
	
	if ($ref->{'inverse'}) {
	    $val = \$ele2;
	    $ele1 = $ref->{'max'};
	    }
	else {
	    $val = \$ele1;
	    $ele2 = $ref->{'min'};
	    }
	    
	for (my $pos = 0; $pos < $len; $pos++) {
	    next if (! defined $labels->[$pos]);

	    # Draw label
	    $$val = $scale->($labels->[$pos]);
#	    my $x = $x_base + (($val - $min) * $inc);
	    my $x = $x_base + (($ele1 - $ele2) * $inc);
	    if ($self->{'axis'}->{'x2'}->{'plot_labels'} == 1) {
    		my $xstr = $x - $xfont_corr;

		# Format label
		my $label = $format->($labels->[$pos]);

		# Draw label
		if(defined($self->{'vertical'})) {
		    $self->{'gd'}->string($self->{'font'}, 
					  &{$self->{'transform'}}($xstr,$y-length($label)*$self->{'font_width'}), 
					  $label, $col);
		} else {
		    $self->{'gd'}->stringUp($self->{'font'}, 
					    &{$self->{'transform'}}($xstr, $y-2), 
					    $label, $col);
    	    	}
	    }
		
	    # Draw tick
	    $self->{'gd'}->line (&{$self->{'transform'}}($x, $y), 
				 &{$self->{'transform'}}($x, $y+5), 
				 $col) if ($self->{'axis'}->{'x2'}->{'plot_ticks'} == 1);

	    # Need X grid?
	    if (defined $grd_col) {
		$self->{'gd'}->line (&{$self->{'transform'}}($x,$y), 
				     &{$self->{'transform'}}($x+$self->{'font_width'},$y+$self->{'plot'}->{'h'}), 
				     $grd_col);
		}
    	    }
    	}

    return 1;
}

sub _draw_y_labels {
    my $self = shift;
    
    if (defined $self->{'axis'}->{'y'}->{'labels'}) {
    	my $ref = $self->{'axis'}->{'y'};
	my $scale = $ref->{'scaling'};
	my $labels = $ref->{'labels'};
    	my $format = $ref->{'format'};

	# If need to plot Y grid?
	my $grd_col = undef;
	if (defined $ref->{'grid'}) {
	    if ($ref->{'grid'} =~ /(\d+),(\d+),(\d+)/) {
	    	$grd_col = $self->{'gd'}->colorAllocate($1, $2, $3); }
	    elsif ($ref->{'grid'} == 1) {
	    	$grd_col = $self->{'fg_colour'}; }
	    }

	# Y start point
	my $y_base = $self->{'plot'}->{'y'};
    	
	# X base coord
	my $x_base = $self->{'plot'}->{'x'};
	
	# Nb of entries in Y axis
	my $len = @{$labels};
	my $inc = $ref->{'inc'};

	# Get min value
	my $min = $ref->{'min'};
	my $col;

	if (defined  $ref->{'title'}->{'col'}) {
	   $col = $ref->{'title'}->{'col'}; }
	else {
    	    $col = $self->{'fg_colour'}; }

	my $font_adj = int ($self->{'font_height'} / 2);
    	
	for (my $pos = 0; $pos < $len; $pos++) {
	    # Skip undef entries
	    next if ! defined $labels->[$pos];

	    # Draw label
	    my $val = $scale->($labels->[$pos]);

	    my $y = $y_base - (($val - $min) * $inc) - $font_adj;

	    my $y_grad = $y_base - (($val - $min) * $inc);
    	    
	    # Draw label
    	    if ($self->{'axis'}->{'y'}->{'plot_labels'} == 1) {
		# Format label
		my $label = $format->($labels->[$pos]);
		if(defined($self->{'vertical'})) {
		    $self->{'gd'}->stringUp ($self->{'font'}, 
					   &{$self->{'transform'}}($x_base, $y), 
					   $label, $col);
		} else {
		    # Set x from x_base according to label size (right justify!)
		    $self->{'gd'}->string ($self->{'font'}, 
					     &{$self->{'transform'}}($x_base-((length($label)+1)*$self->{'font_width'}), $y), 
					     $label, $col);
    	    	}
	    }
		
	    # Draw tick
	    $self->{'gd'}->line (&{$self->{'transform'}}($x_base - 5, $y_grad), 
				 &{$self->{'transform'}}($x_base, $y_grad), 
				 $col) if ($self->{'axis'}->{'y'}->{'plot_ticks'} == 1);

	    # Need Y grid?
	    if (defined $grd_col) {
		$self->{'gd'}->line (&{$self->{'transform'}}($x_base, $y_grad), 
	    	    	    	     &{$self->{'transform'}}($x_base + $self->{'plot'}->{'w'}, $y_grad), 
				     $grd_col);
		}
	    }
    	}
	
    if ($self->{'gdata2'}->{'nb_sets'} > 0 && 
    	defined $self->{'axis'}->{'y2'}->{'labels'}) {
    	my $ref = $self->{'axis'}->{'y2'};

	# Keep refs of data and scale function
	my $labels = $ref->{'labels'};
	my $scale = $ref->{'scaling'};
	my $format = $ref->{'format'};

	# If need to plot X grid?
	my $grd_col = undef;
	if (defined $ref->{'grid'}) {
	    if ($ref->{'grid'} =~ /(\d+),(\d+),(\d+)/) {
	    	$grd_col = $self->{'gd'}->colorAllocate($1, $2, $3); }
	    elsif ($ref->{'grid'} == 1) {
	    	$grd_col = $self->{'fg_colour'}; }
	    }

	# Y start point
	my $y_base = $self->{'plot'}->{'y'};
    	
	# X base coord
	my $x_base = $self->{'plot'}->{'x'} + $self->{'plot'}->{'w'};
	
	# Nb of entries in Y axis
	my $len = @{$labels};
	my $inc = $ref->{'inc'};

	# Get min value
	my $min = $ref->{'min'};
	my $col;
    	
	if (defined  $ref->{'title'}->{'col'}) {
	   $col = $ref->{'title'}->{'col'}; }
	else {
    	    $col = $self->{'fg_colour'}; }

	my $font_adj = int ($self->{'font_height'} / 2);
    	
	for (my $pos = 0; $pos < $len; $pos++) {
	    # Skip undef entries
	    next if ! defined $labels->[$pos];

	    # Draw label
	    my $val = $scale->($labels->[$pos]);

	    my $y = $y_base - (($val - $min) * $inc) - $font_adj;

	    my $y_grad = $y_base - (($val - $min) * $inc);
    	    if ($self->{'axis'}->{'y2'}->{'plot_labels'} == 1) {
		# Format label
		my $label = $format->($labels->[$pos]);
		my $x;
		if(defined($self->{'vertical'})) {
		    $self->{'gd'}->stringUp ($self->{'font'}, 
					     &{$self->{'transform'}}($x_base+((length($label)+1)*$self->{'font_width'}), $y), 
					     $label, $col);
		} else {
		    # Set x from x_base according to label size (right justify!)
		    $self->{'gd'}->string ($self->{'font'}, 
					   &{$self->{'transform'}}($x_base+$self->{'font_width'}, $y), 
					   $label, $col);
    	    	}
	    }
		
	    # Draw tick
	    $self->{'gd'}->line (&{$self->{'transform'}}($x_base, $y_grad), 
				 &{$self->{'transform'}}($x_base + 5, $y_grad), 
				 $col) if ($self->{'axis'}->{'y2'}->{'plot_ticks'} == 1);

	    # Need Y2 grid?
	    if (defined $grd_col) {
		$self->{'gd'}->line (&{$self->{'transform'}}($self->{'plot'}->{'x'}, $y_grad), 
	    	    	    	     &{$self->{'transform'}}($x_base, $y_grad), $grd_col);
		}
	    }
	}

    return 1;
}

sub _set_labels_limits {
    my $self = shift;
    my $ref = shift;
    
    return 0 unless defined $ref->{'labels'};
    
    my $scale = $ref->{'scaling'};
    my $format = $ref->{'format'};
    
    # Ref func refs
    my $labels = $ref->{'labels'};

    my $d_len = @{$labels};
        
    # Pick first valid label
    my $pos = 0;
    for (; $pos < $d_len; $pos++) {
	last if (defined $labels->[$pos]); }
    
    my $val = $scale->($labels->[$pos]);
    
    # Set starting values
    my $x1 = $labels->[$pos];
    my $x2 = $format->($x1);
    my $max_len = length ($x2);
#    my $max_len = length ($format->($labels->[$pos]));
    my $max = $val;
    my $min = $val;
    
    # Screen labels and define limits
    for ($pos = 0; $pos < $d_len; $pos++) {
	next if (! defined $labels->[$pos]);
	# Get max limits
	my $val = $scale->($labels->[$pos]);	    
	my $len = length($format->($labels->[$pos]));
	$max_len = $len if ($len > $max_len);
	$max = $val if ($val > $max);
	$min = $val if ($val < $min);
	}
    
    # Set values
    $ref->{'lab_size'} = ($max_len * $self->{'font_width'}) + $self->{'plot'}->{'space'};
    	
    $ref->{'min'} = $min;
    $ref->{'max'} = $max;
    
    return 1;
}

sub _set_labels_fromX {
    my $self = shift;
    my $ref = shift;
    my $gdata = shift;
        
    my $scale = $ref->{'scaling'};
    my $format = $ref->{'format'};
    
    my $min = undef;
    my $max = undef;
    my $max_len = 1;

    # Force envenly space call
    $ref->{'evenspace'} = 1 unless $ref->{'evenspace'} > 0;
    
    # Define min & max from X values of the passed data sets
    foreach my $entry (keys %{$gdata}) {
	# Has X values?
	next if (! defined $gdata->{$entry}->{'x'});
	# Get ref to X data array
	my $array = $gdata->{$entry}->{'x'};
	# Process content
	my $alen = @{$array};
	for (my $pos = 0; $pos < $alen; $pos++) {
	    next if (! defined $array->[$pos]);
    	    
	    my $val = $scale->($array->[$pos]);
	    my $len = length($format->($array->[$pos]));
    	    
	    # Get limits
	    $max_len = $len if ($len > $max_len);
	    $max = $val if (! defined $max || $val > $max);
	    $min = $val if (! defined $min || $val < $min);
	    }
	}
	
    # If no X labels in data set, define X from the nb of entries in Y sets
    if (! defined $min) {
	# Min is zero
	$min = 0;
	$max = 0;
	$max_len = 1;
	my $count = 0;
	foreach my $entry (keys %{$gdata}) {
	    next if ! defined $gdata->{$entry}->{'y'};
	    # get the nb of entries
	    my $nb_val = @{$gdata->{$entry}->{'y'}};
#	    my $len = length ($format->($nb_val));
	    my $len = length ("$nb_val");
	    	    
	    # Get limits
	    $max_len = $len if ($len > $max_len);
	    $max = $nb_val if ($nb_val > $max);

	    # Define the X labels for this set
	    my $val = 1;
#	    my $inc = $max / $nb_val;
	    for (my $pos = 0; $pos < $nb_val; $pos++) {
		$ref->{'x'}->[$pos] = $val;
		$val++;
		}
	    	    
	    $count++;
	    }
    	return 0 unless $count > 0;
	}
        
    # Set values
    $ref->{'lab_size'} = $max_len * $self->{'font_width'};
	
#    $self->{'x_range'} = $max - $min;
    $ref->{'min'} = $min;
    $ref->{'max'} = $max;
        
    return 1;
}

sub _set_labels_fromY {
    my $self = shift;
    my $ref = shift;
    my $gdata = shift;

    my $scale = $ref->{'scaling'};
    my $format = $ref->{'format'};

    my $min = undef;
    my $max = undef;
    my $max_len = 1;
        
    # Force envenly space call
    $ref->{'evenspace'} = 1;
    
    # Define min & max from the passed Y values
    my $count = 0;
    foreach my $entry (keys %{$gdata}) {
	# Has Y values?
	next if (! defined $gdata->{$entry}->{'y'});
    	
	# Define starting values
	my $array = $gdata->{$entry}->{'y'};
	my $len = @{$array};
	my $val = $scale->($array->[0]);
    	
	$max_len = length ($format->($array->[0]));
	$max = $val if (! defined $max);
	$min = $val if (! defined $min);
    	
	# Loop on Y data
	for (my $pos = 0; $pos < $len; $pos++) {
	    next if (! defined $array->[$pos]);
	    # Get value
	    my $val = $scale->($array->[$pos]);
	    my $len = length ($format->($array->[$pos]));
	    # Get limits
	    $max_len = $len if ($len > $max_len);
	    $max = $val if (! defined $max || $val > $max);
	    $min = $val if (! defined $min || $val < $min);
    	    }
	$count++;
	}

    return 0 unless $count > 0;

    # Get Y data limits
    $ref->{'lab_size'} = $max_len * $self->{'font_width'};
        
    $ref->{'min'} = $min;
    $ref->{'max'} = $max;

    return 1;
}

sub _set_legend_limits {
    my $self = shift;
    my $legend = shift;
    my $loc = shift;
        
    # Defined legend?
    if (! @{$legend->{'entries'}}) {
    	$legend->{'plot'} = 0;
	return 0;
	}
    
    # The first legend is actually the 'Legend:' string
    $legend->{'entries'}->[0] = $legend->{'title'};
    
    # Get pos type, vertical by default
    my $lmode = 0;
    my $w_limit = 0;
    my $h_limit = 0;
    if ($loc =~ /(top|bottom)/) {
    	# Set horizontal mode
    	$lmode = 1;
	# Get width limit
	$w_limit = $self->{'width'} - $self->{'right_corr'} - $self->{'left_corr'} - 6;
	}
    else {
    	$h_limit = $self->{'height'} - $self->{'top_corr'} - $self->{'bottom_corr'} - 6;
	}
	   
    my $h = 0;
    my $w = 0;
    my $max_w = 0;
    my $max_h = $self->{'font_height'};
    my $max_len = 0;
    
    foreach my $entry (@{$legend->{'entries'}}) {
    	next unless defined $entry;
	
	my $len = length $entry;
	my $lw = $len * $self->{'font_width'};
    	$max_len = $lw if ($lw > $max_len);
	
	# If horizontal legend
	if ($lmode == 1) {
	    # If out of limits
	    if ($w + $lw + 5 + $legend->{'tl'} > $w_limit) {
	    	# Pass to next line
	    	$max_h += $self->{'font_height'} + 2;
		$w = 0;
		}
	    $w += $lw + 5 + $legend->{'tl'};
	    $max_w = $w if ($w > $max_w);
	    }
        else {
	    # If out of limits
	    if ($h + $self->{'font_height'} + 2 > $h_limit) {
	    	# Pass to next column
	    	$max_w += $max_len + 5 + $legend->{'tl'};
		$max_len = 0;
		$h = 0;
		}
	    $h += $self->{'font_height'} + 2;
	    $max_h = $h if ($h > $max_h);
            }
	}
    
    # Update max width with last column
    if ($lmode == 1) {
	$max_w += 4;
	}
    else {
    	$max_w += $max_len + 9 + $legend->{'tl'};
    	}
    $max_h += 4;
    
    # Defined legend area now
    $legend->{'lw'} = $max_w + $self->{'plot'}->{'space'};
    $legend->{'lh'} = $max_h + $self->{'plot'}->{'space'};
    
    # Set params for plot definition
    if ($lmode) {
	$legend->{'w'} = 0;
	$legend->{'h'} = $legend->{'lh'};
	}
    else {
    	$legend->{'w'} = $legend->{'lw'};
	$legend->{'h'} = 0;
	}
 
    return 1;
}

# ============================================================
# Function: _set_graph_limits
# Passed:   n/a
# Returns:  1
# On error: 
# Description: 
# Calculates all parameters necessary to plot correctly the curves
# on the gif.
# ============================================================

sub _set_graph_limits {
    my $self = shift;

    my $max_len;
    my $max;
    my $min;
    my $pos;
    my $d_len;
#    my $len;
    my $range;    
    # Get max height of the X labels
    
    # Try with user defined labels    
    if (! $self->_set_labels_limits($self->{'axis'}->{'x'})) {
    	# Try with X values associated to datasets
	if (! $self->_set_labels_fromX($self->{'axis'}->{'x'}, $self->{'gdata'}->{'sets'})) {
	    # Need X labels!
	    cluck ("Error: impossible to define X axis, check your data.\n");
	    return 0;
	    }
    	}
            
    # Set Y axis labels data
    if (! $self->_set_labels_limits($self->{'axis'}->{'y'})) {
    	# Try with Y values of the datasets
	if (! $self->_set_labels_fromY($self->{'axis'}->{'y'}, $self->{'gdata'}->{'sets'})) {
	    # Need Y labels!
	    cluck ("Error: impossible to define Y axis, check your data.\n");
	    return 0;
	    }
    	}
    
    if ($self->{'gdata2'}->{'nb_sets'} > 0) {
	# Second data sets has X labels too?
	if (! $self->_set_labels_limits($self->{'axis'}->{'x2'})) {
    	    # Try with X values associated to datasets
	    if (! $self->_set_labels_fromX($self->{'axis'}->{'x2'}, $self->{'gdata2'}->{'sets'})) {
		# Ok, no labels and ticks for X2 axis.
    		$self->{'axis'}->{'x2'}->{'plot_labels'} = 0;
    		$self->{'axis'}->{'x2'}->{'plot_ticks'} = 0;
		# Use those of the X one
		$self->{'axis'}->{'x2'}->{'min'} = $self->{'axis'}->{'x'}->{'min'};
		$self->{'axis'}->{'x2'}->{'max'} = $self->{'axis'}->{'x'}->{'max'};
		$self->{'axis'}->{'x2'}->{'labels'} = $self->{'axis'}->{'x'}->{'labels'};
		}
    	    }

	# Set Y2 axis labels data
	if (! $self->_set_labels_limits($self->{'axis'}->{'y2'})) {
    	    # Try with Y values of the datasets
	    if (! $self->_set_labels_fromY($self->{'axis'}->{'y2'}, $self->{'gdata2'}->{'sets'})) {
		# Ok, no labels for Y2 axis.
    		$self->{'axis'}->{'y2'}->{'plot_labels'} = 0;
    		$self->{'axis'}->{'y2'}->{'plot_ticks'} = 0;
		# Use data from the Y one
		$self->{'axis'}->{'y2'}->{'min'} = $self->{'axis'}->{'y'}->{'min'};
		$self->{'axis'}->{'y2'}->{'max'} = $self->{'axis'}->{'y'}->{'max'};
		$self->{'axis'}->{'y2'}->{'labels'} = $self->{'axis'}->{'y'}->{'labels'};
		}
    	    }
	}

    # Now define legends limits
    if (defined $self->{'legend_loc'}) {
    	$self->_set_legend_limits ($self->{'legend'}, $self->{'legend_loc'}); }
    if (defined $self->{'legend2_loc'}) {
    	$self->_set_legend_limits ($self->{'legend2'}, $self->{'legend2_loc'}); }
        
    # Define plot area
    
    # Plot height is gif heigth less top and bottom corrections (border, offset, ...), X1 and X2 titles, X1 and X2 label heigths and main graph title
    $self->{'plot'}->{'h'} = $self->{'height'} - $self->{'top_corr'} - $self->{'bottom_corr'} - $self->{'title'}->{'h'} - $self->{'plot'}->{'space'};
    
    $self->{'plot'}->{'h'} -= ($self->{'axis'}->{'x'}->{'title'}->{'h'} + $self->{'plot'}->{'space'});
    $self->{'plot'}->{'h'} -= $self->{'axis'}->{'x'}->{'lab_size'} if ($self->{'axis'}->{'x'}->{'plot_labels'} == 1);
    $self->{'plot'}->{'h'} -= $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'x'}->{'plot_ticks'} == 1);
     
     $self->{'plot'}->{'h'} -= ($self->{'axis'}->{'x2'}->{'title'}->{'h'} + $self->{'plot'}->{'space'});
     $self->{'plot'}->{'h'} -= $self->{'axis'}->{'x2'}->{'lab_size'} if ($self->{'axis'}->{'x2'}->{'plot_labels'} == 1);
     $self->{'plot'}->{'h'} -= $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'x2'}->{'plot_ticks'} == 1);
    
    $self->{'plot'}->{'h'} -= ($self->{'legend'}->{'h'} + $self->{'legend2'}->{'h'});
    
    # Set Y increment value    
    $range = $self->{'axis'}->{'y'}->{'max'} - $self->{'axis'}->{'y'}->{'min'};
    if ($range == 0) {
    	confess ("Y axis range is zero length, curves can not be plotted.\n");
	}

    $self->{'axis'}->{'y'}->{'inc'} = $self->{'plot'}->{'h'} / $range;
    
    # Plot width is gif width less left and right corrections, Y1 and Y2 titles and Y1 and Y2 label width.
    $self->{'plot'}->{'w'} = $self->{'width'} - $self->{'right_corr'} - $self->{'left_corr'};
    
    $self->{'plot'}->{'w'} -= ($self->{'axis'}->{'y'}->{'title'}->{'w'} + $self->{'plot'}->{'space'});
    $self->{'plot'}->{'w'} -= $self->{'axis'}->{'y'}->{'lab_size'} if ($self->{'axis'}->{'y'}->{'plot_labels'} == 1);
    $self->{'plot'}->{'w'} -= $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'y'}->{'plot_ticks'} == 1);
     
    $self->{'plot'}->{'w'} -= ($self->{'axis'}->{'y2'}->{'title'}->{'w'} + $self->{'plot'}->{'space'});
    $self->{'plot'}->{'w'} -= $self->{'axis'}->{'y2'}->{'lab_size'} if ($self->{'axis'}->{'y2'}->{'plot_labels'} == 1);
    $self->{'plot'}->{'w'} -= $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'y2'}->{'plot_ticks'} == 1);
    
    $self->{'plot'}->{'w'} -= ($self->{'legend'}->{'w'} + $self->{'legend2'}->{'w'});
    
    # Set X increment value
    $range = $self->{'axis'}->{'x'}->{'max'} - $self->{'axis'}->{'x'}->{'min'};
    if ($range == 0) {
    	confess ("X axis range is zero length, curves can not be plotted.\n");
	}

    $self->{'axis'}->{'x'}->{'inc'} = $self->{'plot'}->{'w'} / $range;

    if ($self->{'gdata2'}->{'nb_sets'} > 0) {
	# Set Y2increment value
	$range = $self->{'axis'}->{'y2'}->{'max'} - $self->{'axis'}->{'y2'}->{'min'};
	if ($range == 0) {
    	    confess ("Second Y axis range is zero length, curves can not be plotted.\n");
	    }
	    
	$self->{'axis'}->{'y2'}->{'inc'} = $self->{'plot'}->{'h'} / $range;

	# Set X increment value
	$range = $self->{'axis'}->{'x2'}->{'max'} - $self->{'axis'}->{'x2'}->{'min'};
	if ($range == 0) {
    	    confess ("Second X axis range is zero length, curves can not be plotted.\n");
	    }

	$self->{'axis'}->{'x2'}->{'inc'} = $self->{'plot'}->{'w'} / $range;
    	}
    
    # Define graph refs
    $self->{'plot'}->{'x'} = $self->{'x'} + $self->{'left_corr'};
    
    $self->{'plot'}->{'x'} += $self->{'legend'}->{'w'} if (defined $self->{'legend_loc'} && $self->{'legend_loc'} =~/left/);
    $self->{'plot'}->{'x'} += $self->{'legend2'}->{'w'} if (defined $self->{'legend2_loc'} && $self->{'legend2_loc'} =~/left/);
    
    $self->{'plot'}->{'x'} += $self->{'axis'}->{'y'}->{'title'}->{'w'} + $self->{'plot'}->{'space'};
    $self->{'plot'}->{'x'} += $self->{'axis'}->{'y'}->{'lab_size'} if ($self->{'axis'}->{'y'}->{'plot_labels'} == 1);
    $self->{'plot'}->{'x'} += $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'y'}->{'plot_ticks'} == 1);
        
    $self->{'plot'}->{'y'} = $self->{'y'} + $self->{'height'} - $self->{'bottom_corr'};
    
    $self->{'plot'}->{'y'} -= ($self->{'axis'}->{'x'}->{'title'}->{'h'} + $self->{'plot'}->{'space'});
    $self->{'plot'}->{'y'} -= $self->{'axis'}->{'x'}->{'lab_size'} if ($self->{'axis'}->{'x'}->{'plot_labels'} == 1);
    $self->{'plot'}->{'y'} -= $self->{'axis'}->{'tick_size'} if ($self->{'axis'}->{'x'}->{'plot_ticks'} == 1);
    
    $self->{'plot'}->{'y'} -= $self->{'legend'}->{'h'} if (defined $self->{'legend_loc'} && $self->{'legend_loc'} =~/bottom/);
    $self->{'plot'}->{'y'} -= $self->{'legend2'}->{'h'} if (defined $self->{'legend2_loc'} && $self->{'legend2_loc'} =~/bottom/);
    return 1;
}

sub _add_dataset {
    my $self = shift;
    my $dataref = shift;
    my $name = shift;
    my $x_data = shift;
    my $y_data = shift;

    # Some checks
    if (! defined $y_data) {
	cluck ("Error: pass at least an array reference for Y values to when adding a data set.\n");
	return 0;
    	}
    
    if (ref ($y_data) ne "ARRAY") {
    	cluck ("Pass a reference to an array of Y values when adding a dataset.\n");
	return 0;
	}
    
    # Now keep rest of data
    $name = "Data set ". $dataref->{'nb_sets'} + 1 unless (defined $name);
    
    # Keep ref of the X data
    if (defined $x_data) {
	if (ref ($x_data) ne "ARRAY") {
	    cluck ("Pass a reference to an array of X values when adding a dataset.\n");
    	    return 0;
	    }
	    
	$dataref->{'sets'}->{$name}->{'x'} = $x_data;
	}
    
    # Keep ref of the Y data
    $dataref->{'sets'}->{$name}->{'y'} = $y_data;
    
    # Set default style
    $dataref->{'sets'}->{$name}->{'line_style'} = PLAIN_LINE;
    
    # Ignore gaps by default
    $dataref->{'sets'}->{$name}->{'ignore_gaps'} = 1;

    # Bgrnd colour by default
    $dataref->{'colours'}->[$dataref->{'nb_sets'}] = undef;
    
    # Keep track of the curve names
    push (@{$dataref->{'names'}}, $name);
    
    $dataref->{'nb_sets'}++;
    
    return 1;
}

=head2 add_dataset()

    $graph->add_dataset ("Data set1", \@xdata, \@ydata);

Curve_plot::add_dataset is used to add a new set of values to be plotted on the graph. The name of the data set is optional (pass undef). The data for the X axis is also optional (pass undef).
It returns 1 on success and 0 on failure.

=cut

sub add_dataset {
    my $self = shift;
    my $name = shift;
    my $x_data = shift;
    my $y_data = shift;

    return $self->_add_dataset ($self->{'gdata'}, $name, $x_data, $y_data);
}

=head2 add_dataset2()

    $graph->add_dataset2 ("Data set1", \@xdata, \@ydata);

Curve_plot::add_dataset2 does exactly the same as 'add_dataset' except that the data are added for the second Y axis for combined graphs.

It returns 1 on success and 0 on failure.

=cut

sub add_dataset2 {
    my $self = shift;
    my $name = shift;
    my $x_data = shift;
    my $y_data = shift;

    return $self->_add_dataset ($self->{'gdata2'}, $name, $x_data, $y_data);
}

sub _remove_dataset {
    my $self = shift;
    my $dataref = shift;
    my $ref = shift;

    my @new_colours;
    my @new_names;
    
    $ref = $self->_get_data_ref ($dataref, $ref);
    
    return 0 unless defined $ref;
    
    delete $dataref->{'sets'}->{$dataref->{'names'}->[$ref]};
    
    $self->{'gd'}->colorDeallocate ($dataref->{'colours'}->[$ref]);
    
    # Clean arrays
    my $cpos = 0;
    for (my $pos = 0; $pos < $self->{'nb_sets'}; $pos++) {
    	next if ($pos == $ref);
	$new_colours[$cpos] = $dataref->{'colours'}->[$pos];
	$new_names[$cpos] = $dataref->{'names'}->[$pos];
	$cpos++;
    	}
    
    # Update arrays
    @{$dataref->{'names'}}   = @new_names;
    @{$dataref->{'colours'}} = @new_colours;
    
    $dataref->{'nb_sets'}--;
    
    return 1;
}

=head2 remove_dataset()

    $graph->remove_dataset ($curve_ref);

Curve_plot::remove_dataset is used to remove a set of values ( a curve) added with the 'add_dataset' method. Pass either the curve reference number (0 = first) or its name.
It returns 1 on success and 0 on failure.

=cut

sub remove_dataset {
    my $self = shift;
    my $ref = shift;

    return $self->_remove_dataset ($self->{'gdata'}, $ref);
}


=head2 remove_dataset2()

    $graph->remove_dataset ($set_nb);

Curve_plot::remove_dataset2 does the same as 'remove_dataset' except that it applies for the second Y axis.

It returns 1 on success and 0 on failure.

=cut

sub remove_dataset2 {
    my $self = shift;
    my $ref = shift;

    return $self->_remove_dataset ($self->{'gdata2'}, $ref);
}

sub _dataset_define {
    my $self = shift;
    my $dataref = shift;
    my $ref = shift;
    my $args = shift;
    
    $ref = $self->_get_data_ref ($dataref, $ref);
    
    return 0 unless defined $ref;
    
    my $data = $dataref->{'sets'}->{$dataref->{'names'}->[$ref]};

    # Parse defs but check syntax => don't want rubbish
    foreach my $key (keys %{$args}) {
	if ($key eq "color" || $key eq "colour") {
	    $dataref->{'colours'}->[$ref] = $self->{'gd'}->colorAllocate(@{$args->{$key}}); }
	elsif ($key eq "line_width") {
	    $data->{$key} = $args->{$key}; }
	elsif ($key eq "ignore_gaps") {
	    $data->{$key} = $args->{$key}; }
	elsif ($key eq "line_style") {
#	    $data->{'style'} = LINE_STYLE;
	    if ($args->{$key} eq "plain") {
		$data->{$key} = PLAIN_LINE; }
	    elsif ($args->{$key} eq "dotted") {
		$data->{$key} = DOTTED_LINE; }
	    elsif ($args->{$key} eq "dashed") {
		$data->{$key} = DASHED_LINE; }
	    elsif ($args->{$key} eq "none") {
	    	# Set default to square dots
	    	$data->{'dot_style'} = SQUARE_DOT unless defined $data->{'dot_style'};
		$data->{$key} = undef;
		}
	    }
	elsif ($key eq "dot_style") {
#	    $data->{'style'} = DOT_STYLE;
	    if ($args->{$key} eq "#") {
		$data->{$key} = SQUARE_DOT; }
	    elsif ($args->{$key} eq "X") {
		$data->{$key} = CROSS_DOT; }
	    elsif ($args->{$key} eq "+") {
		$data->{$key} = PLUS_DOT; }
	    elsif ($args->{$key} eq "O") {
		$data->{$key} = CIRC_DOT; }
	    elsif ($args->{$key} eq ".") {
		$data->{$key} = POINT_DOT; }
	    elsif ($args->{$key} eq "none") {
		# Set default to plain line
	    	$data->{'line_style'} = PLAIN_LINE unless defined $data->{'line_style'};
		$data->{$key} = undef;
		}
	    }
    	}
    return 1;
}

=head2 dataset_define()

    $graph->dataset_define ($curve, 'colour' => [124,56,213], 'line_style' => 'dotted');

Curve_plot::dataset_define is used to define attributes to the set. The first argument is the reference number of the curve (0 for first one, 1 for second, etc...) or its name.
The second argument is a hash table with the following parameters:

=over 4

=item 'line_width'

Width of the line in pixels.

Default: 1 pixel or the global line width if specified. Example: $graph->dataset_define ($curve, 'line_width' => 3).

=item 'color' or 'colour'

Reference to array with RGB colours for the line.

Default: foreground colour or colour defined by the module if auto_colour used. Example: $graph->dataset_define ($curve, 'colour' => [124,56,213]).

=item 'line_style'

Set this to specify the line style as:
'plain', 'dashed' or 'dotted'. Use 'none' to remove line drawing (this will set the default to 'dot_style' = '#' if not defined). Avoid mixing line width and non-plain line types, the result is not very nice.

Default: 'plain'. Example: $graph->dataset_define ($curve, 'line_style' => 'dashed').

=item 'dot_style'

Set this if you want a special style plot for each point. The style can be '#' for squares, '+' for "plus", 'X' for "check", 'O'  (letter o upper case!) for circles or '.' for single pixel. Use 'none' to remove the dot style (this will set the default to 'line_style' = 'plain' if not defined).

Default: 'none'. Example: $graph->dataset_define ($curve, 'line_type' => 'none', 'dot_style' => '+').

=item 'ignore_gaps'

By default set to 1 and means that any value in the Y array equal to 'undef' is ignored. If set to 0, undefined values in the array will be taken into account. When plotting the curve, any undefined value will break the curve, For dot plots, it does not change anything.

=back

It returns 1 on success and 0 on failure.

=cut

sub dataset_define {
    my $self = shift;
    my $ref = shift;
    my %args = @_;
       
    return $self->_dataset_define ($self->{'gdata'}, $ref, \%args);
}

=head2 dataset_define2()

    $graph->dataset_define2 ($curve, 'line_width' => 3);

Curve_plot::dataset_define2 does the same as 'dataset_define' expect that it applies for the data sets used on the second Y axis.

It returns 1 on success and 0 on failure.

=cut

sub dataset_define2 {
    my $self = shift;
    my $ref = shift;
    my %args = @_;
       
    return $self->_dataset_define ($self->{'gdata2'}, $ref, \%args);
}

sub _get_data_ref {
    my $self = shift;
    my $dataref = shift;
    my $ref = shift;
    
    # Passed a name or a value?
    if ($ref !~ /^\d+$/) {
    	my $pos = 0;
	foreach my $entry (@{$dataref->{'names'}}) {
	    last if ($entry eq $ref);
	    $pos++;
	    }
	if ($pos == $dataref->{'nb_sets'}) {
	    cluck ("Error: invalid dataset reference name $ref\n");
	    return undef;
	    }
    	return $pos;
	}
    else {
    	if (! defined $dataref->{'names'}->[$ref] || 
	    ! defined $dataref->{'sets'}->{$dataref->{'names'}->[$ref]}) {
	    cluck ("Error: invalid dataset array reference number $ref.\n");
	    return undef;
	    }
	return $ref;
	}
}

sub _plot_lines {
    my $self = shift;
    my $data_set = shift;
    my $x_axis = shift;
    my $y_axis = shift;
    my $col = shift;
    my $lwidth = shift;
    my $ignore_gaps = shift;
        
    my $x_data = $data_set->{'x'};
    my $y_data = $data_set->{'y'};
    
    # Get plot info
    my $y_base = $self->{'plot'}->{'y'};
    my $x_base = $self->{'plot'}->{'x'};

    my $y_scale = $y_axis->{'scaling'};
    my $x_scale = $x_axis->{'scaling'};
        
    my $x_min = $x_axis->{'min'};
    my $y_min = $y_axis->{'min'};
    my $y_max = $y_axis->{'max'};
        
    # Define increment ratio
    my $x_inc = $x_axis->{'inc'};
    my $y_inc = $y_axis->{'inc'};
    
    my $len = @{$y_data} - 1;
    
    # Are there data to plot?
    return 0 unless ($len != -1);
    
    my $pos1 = 0;

    # First coord -> get first non undef val
    while (! defined $y_data->[$pos1]){ 
	$pos1++;
	last if ($pos1 > $len);
	}
    
    if ($pos1 > $len) {
    	cluck ("No defined Y values.\n");
	return 0;
	}
    
    # Set x1 and y1
    my $val = $y_scale->($y_data->[$pos1]);

    $val = $y_max if ($val > $y_max);
    $val = $y_min if ($val < $y_min);

    my $xval = $x_scale->($x_data->[$pos1]);
    
    my $x1 = $x_base + (($xval - $x_min) * $x_inc);
    my $y1 = $y_base - (($val - $y_min) * $y_inc);
    
    my $pos2;
    my $x2;
    my $y2;
    
    while ($pos1 < $len) {
    	# Handle undef values
	if (! defined $y_data->[$pos1]) {
	    # If need to break curve with undef val
	    if (! $ignore_gaps) {
	    	# Reach next defined one
	    	while (! defined $y_data->[$pos1]) {
		    $pos1++;
		    last if ($pos1 >= $len);
		    }
		    
		last if ($pos1 >= $len);
		
		# Set new x1,y1
		$val = $y_scale->($y_data->[$pos1]);

		$val = $y_max if ($val > $y_max);
		$val = $y_min if ($val < $y_min);

		$xval = $x_scale->($x_data->[$pos1]);

		$x1 = $x_base + (($xval - $x_min) * $x_inc);
		$y1 = $y_base - (($val - $y_min) * $y_inc);
		}
	    else {
		$pos1++;
		next;
		}
	    }
    	
	last if ($pos1 >= $len);
	
	# Get 2nd poimt and cope with undef values if needed
	$pos2 = $pos1+1;
    	
	while ($pos2 < $len) {
	    if (defined $y_data->[$pos2]) {
	    	last; }
	    elsif (! $ignore_gaps) {
		$pos1 = $pos2;
		last;
		}
	    $pos2++;
	    }
    	
	last if ($pos2 >= $len);
	
    	# If undef Y val restart plot from next defined one
	redo if ($pos1 == $pos2);
	
#	 last if $pos2 == $len;
    	
	# Work with second Y val
 	$val = $y_scale->($y_data->[$pos2]);

	$val = $y_max if ($val > $y_max);
	$val = $y_min if ($val < $y_min);

	$xval = $x_scale->($x_data->[$pos2]);

	$x2 = $x_base + (($xval - $x_min) * $x_inc);
	$y2 = $y_base - (($val - $y_min) * $y_inc);
    	
	# Draw segment now
	$self->{'gd'}->line (&{$self->{'transform'}}($x1, $y1), 
			     &{$self->{'transform'}}($x2, $y2), gdStyled);
    	
	# Add line width
	for (my $lpos = 1; $lpos < $lwidth; $lpos++) {
	    $self->{'gd'}->line (&{$self->{'transform'}}($x1, $y1 - $lpos), 
				 &{$self->{'transform'}}($x2, $y2 - $lpos), gdStyled); }

	for (my $lpos = 1; $lpos < $lwidth; $lpos++) {
	    $self->{'gd'}->line (&{$self->{'transform'}}($x1, $y1 + $lpos), 
				 &{$self->{'transform'}}($x2, $y2 + $lpos), gdStyled); }

	# Second pos becomes first now
	$pos1 = $pos2;
	$x1 = $x2;
	$y1 = $y2;
	}
    return 1;
}

sub _plot_dots {
    my $self = shift;
    my $data_set = shift;
    my $x_axis = shift;
    my $y_axis = shift;
    my $col = shift;
    
    my $x_data = $data_set->{'x'};
    my $y_data = $data_set->{'y'};
    
    # Get plot info
    my $y_base = $self->{'plot'}->{'y'};
    my $x_base = $self->{'plot'}->{'x'};

    my $y_scale = $y_axis->{'scaling'};
    my $x_scale = $x_axis->{'scaling'};
        
    my $x_min = $x_axis->{'min'};
    my $y_min = $y_axis->{'min'};
    my $y_max = $y_axis->{'max'};
        
    # Define increment ratio
    my $x_inc = $x_axis->{'inc'};
    my $y_inc = $y_axis->{'inc'};
    
    my $len = @{$y_data} - 1;
    my $pos = 0;
    my $x;
    my $y;
    my $val;
    my $xval;
    
    while ($pos < $len) {
	if (! defined $y_data->[$pos]) {
	    $pos++;
	    next;
	    }
	    
	# Sec val
 	$val = $y_scale->($y_data->[$pos]);

	$val = $y_max if ($val > $y_max);
	$val = $y_min if ($val < $y_min);

	$xval = $x_scale->($x_data->[$pos]);

	$x = $x_base + (($xval - $x_min) * $x_inc);
	$y = $y_base - (($val - $y_min) * $y_inc);
    	
	# If dot block style
	if ($data_set->{'dot_style'} == SQUARE_DOT) {
	    $self->{'gd'}->filledRectangle(&{$self->{'transform'}}($x-2,$y-2),
					   &{$self->{'transform'}}($x+2,$y+2),gdStyled); }
	# If dot X style
	elsif ($data_set->{'dot_style'} == CROSS_DOT) {
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x,$y),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y+1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y+1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y+2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y+2),gdStyled);
	    }
	# Dot O style
	elsif ($data_set->{'dot_style'} == CIRC_DOT) {
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y+1),gdStyled);
	    
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y+1),gdStyled);

	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y-2),gdStyled);
	    
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y+2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y+2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y+2),gdStyled);
    	    }
	# Dot + style
	elsif ($data_set->{'dot_style'} == PLUS_DOT) {
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y)  ,gdStyled);

	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-1,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+1,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y-1),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y+1),gdStyled);
	    
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x-2,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x+2,$y)  ,gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y-2),gdStyled);
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x  ,$y+2),gdStyled);
	    }
	# Dot . style
	elsif ($data_set->{'dot_style'} == POINT_DOT) {
	    $self->{'gd'}->setPixel(&{$self->{'transform'}}($x,$y),gdStyled);
	    }

	$pos++;
	}
    return 1;
}

# ============================================================
# Function: _plot_data
# Passed:   n/a
# Returns:  n/a
# On error: n/a
# Description: 
# Based on values, generates the curves on the gif.
# ============================================================

sub _plot_data {
    my $self = shift;
    my $dataref = shift;
    my $x_axis = shift;
    my $y_axis = shift;
    
    # Check basal line width value
    my $base_lwidth = $self->{'line_width'};
    $base_lwidth = 1 if $base_lwidth < 1;
    $base_lwidth = 1 if $base_lwidth > ($self->{'plot'}->{'h'} / 2);

    my $entry;
    my $col;
    my $data_pos = 0;
    my $lwidth;
    my $data_set;
    my $status;
    
    # Loop on data sets
    foreach $entry (@{$dataref->{'names'}}) {
	# Keep some references
	$data_set = $dataref->{'sets'}->{$entry};

	$col = $dataref->{'colours'}->[$data_pos];
	$col = $self->{'fg_colour'} if ! defined $col;
    	    	
	# If defined line style
	if (defined $data_set->{'line_style'}) {
	    # Get style
	    if ($data_set->{'line_style'} == PLAIN_LINE) {
		$self->{'gd'}->setStyle($col); }
	    elsif ($data_set->{'line_style'} == DOTTED_LINE) {
		$self->{'gd'}->setStyle($col,$col,$col,gdTransparent,gdTransparent); }
	    elsif ($data_set->{'line_style'} == DASHED_LINE) {
		$self->{'gd'}->setStyle($col,$col,$col,$col,$col,$col,$col,$col,
	    	    	    		$col,$col,gdTransparent,gdTransparent,gdTransparent); }

	    # Define line width, including user defined for this curve
	    $lwidth = $base_lwidth;
	    $lwidth = $data_set->{'line_width'} if defined $data_set->{'line_width'};
    	    
    	    $status = $self->_plot_lines ($data_set, $x_axis, $y_axis, $col, $lwidth, $data_set->{'ignore_gaps'});
    	    }
    	
	# If defined dot style
	if (defined $data_set->{'dot_style'}) {
	    $self->{'gd'}->setStyle($col);
	    
	    $status = $self->_plot_dots ($data_set, $x_axis, $y_axis, $col);
	    }
	
	if (! $status) {
	    print STDERR "Error occured while trying to plot dataset $entry.\n";
	    }
	    
	# Pass to next colour
	$data_pos++;
    	}
    return 1;
}

=head2 plot()

    $graph->plot ();

Curve_plot::plot must be called to generate the plot with the curves.
Does not return a value.

=cut

sub plot {
    my $self = shift;

    my $reset = 0;

    # Make a filled rec on the whole gif with background colour
    $self->{'gd'}->filledRectangle(&{$self->{'transform'}}($self->{'x'}, $self->{'y'}), 
    	    	    	    	   &{$self->{'transform'}}($self->{'x'} + $self->{'width'} - 1, 
							   $self->{'y'} + $self->{'height'} - 1),
				   $self->{'bg_colour'});
    
    # Define border correction;
    $self->{'top_corr'} = $self->{'border'} + $self->{'y_top'};
    $self->{'bottom_corr'} = $self->{'border'} + $self->{'y_off'};
    $self->{'left_corr'} = $self->{'border'} + $self->{'x_off'};
    $self->{'right_corr'} = $self->{'border'} + $self->{'x_top'};

    # Set graph limits first
    return 0 if (! $self->_set_graph_limits ());

    # If need to evenly space axis
    if ($self->{'axis'}->{'x'}->{'evenspace'}) {
	my @nxdata = $self->{'evenlyspace_method'}->($self->{'axis'}->{'x'}->{'labels'},
						     $self->{'plot'}->{'w'},
						     $self->{'axis'}->{'x'}->{'min'}, 
						     $self->{'axis'}->{'x'}->{'max'}, 
						     $self->{'axis'}->{'x'}->{'threshold'},
						     $self->{'font_width'},
						     );

	$self->{'axis'}->{'x'}->{'labels'} = \@nxdata;
	
	$reset = 1;
    	}

    # If need to evenly space axis
    if ($self->{'axis'}->{'x2'}->{'evenspace'}) {
    	# Adjust threshold if needed (get 'x' one)
	unless (defined ($self->{'axis'}->{'x2'}->{'threshold'})) {
	   $self->{'axis'}->{'x2'}->{'threshold'} = $self->{'axis'}->{'x'}->{'threshold'} if (defined $self->{'axis'}->{'x'}->{'threshold'});
	   }
	    
	my @nxdata = $self->{'evenlyspace_method'}->($self->{'axis'}->{'x2'}->{'labels'},
						     $self->{'plot'}->{'w'},
						     $self->{'axis'}->{'x2'}->{'min'}, 
						     $self->{'axis'}->{'x2'}->{'max'}, 
						     $self->{'axis'}->{'x2'}->{'threshold'},
						     $self->{'font_width'},
						     );

	$self->{'axis'}->{'x2'}->{'labels'} = \@nxdata;
	
	$reset = 1;
    	}

    # If need to evenly space Y axis
    if ($self->{'axis'}->{'y'}->{'evenspace'}) {
	my @nydata = $self->{'evenlyspace_method'}->($self->{'axis'}->{'y'}->{'labels'},
						     $self->{'plot'}->{'w'},
						     $self->{'axis'}->{'y'}->{'min'}, 
						     $self->{'axis'}->{'y'}->{'max'}, 
						     $self->{'axis'}->{'y'}->{'threshold'},
						     $self->{'font_width'},
						     );

	$self->{'axis'}->{'y'}->{'labels'} = \@nydata;
	
	$reset = 1;
    	}

    # If need to evenly space Y2 axis
    if ($self->{'axis'}->{'y2'}->{'evenspace'}) {
    	# Adjust threshold if needed (get 'y' one)
	unless (defined ($self->{'axis'}->{'y2'}->{'threshold'})) {
	   $self->{'axis'}->{'y2'}->{'threshold'} = $self->{'axis'}->{'y'}->{'threshold'} if (defined $self->{'axis'}->{'y'}->{'threshold'});
	   }

	my @nydata = $self->{'evenlyspace_method'}->($self->{'axis'}->{'y2'}->{'labels'},
						     $self->{'plot'}->{'w'},
						     $self->{'axis'}->{'y2'}->{'min'}, 
						     $self->{'axis'}->{'y2'}->{'max'}, 
						     $self->{'axis'}->{'y2'}->{'threshold'},
						     $self->{'font_width'},
						     );

	$self->{'axis'}->{'y2'}->{'labels'} = \@nydata;
	
	$reset = 1;
    	}

    # If axis modified reset limits
    $self->_set_graph_limits () if $reset;
    
    if ($self->{'auto_colour'} == 1 || $self->{'auto_color'} == 1) {
	$self->auto_colour (); }
    
    # Ok, now fill X values for all missing one
    foreach my $name (keys %{$self->{'gdata'}->{'sets'}}) {
    	if (! defined $self->{'gdata'}->{'sets'}->{$name}->{'x'}) {
	    my $len = scalar @{$self->{'gdata'}->{'sets'}->{$name}->{'y'}};
	    $self->{'gdata'}->{'sets'}->{$name}->{'x'} = [ 1 .. $len ];
	    }
	}
    
    if ($self->{'gdata2'}->{'nb_sets'} > 0) {
	foreach my $name (keys %{$self->{'gdata2'}->{'sets'}}) {
    	    if (! defined $self->{'gdata2'}->{'sets'}->{$name}->{'x'}) {
		my $len = scalar @{$self->{'gdata2'}->{'sets'}->{$name}->{'y'}};
		$self->{'gdata2'}->{'sets'}->{$name}->{'x'} = [ 1 .. $len ];
		}
	    }
    	}
    
    # Axis + border 1st
    $self->_draw_title;
    $self->_draw_x_titles;
    $self->_draw_y_titles;
    $self->_draw_x_labels;
    $self->_draw_y_labels;

    # Make the plot
    $self->_plot_data ($self->{'gdata'}, $self->{'axis'}->{'x'}, $self->{'axis'}->{'y'});
    $self->_plot_data ($self->{'gdata2'}, $self->{'axis'}->{'x2'}, $self->{'axis'}->{'y2'});

    # Decorate plot now
    $self->_draw_border;
    $self->_draw_legend ($self->{'legend'}, $self->{'legend_loc'}, $self->{'gdata'}) if (defined $self->{'legend_loc'});
    $self->_draw_legend ($self->{'legend2'}, $self->{'legend2_loc'}, $self->{'gdata2'}) if (defined $self->{'legend2_loc'});

    # Draw X axis
    $self->{'gd'}->line (&{$self->{'transform'}}($self->{'plot'}->{'x'}, $self->{'plot'}->{'y'}), 
    	    	    	 &{$self->{'transform'}}($self->{'plot'}->{'x'} + $self->{'plot'}->{'w'}, 
						 $self->{'plot'}->{'y'}), 
			 $self->{'fg_colour'}) if ($self->{'axis'}->{'x'}->{'plot_base'} == 1);

    if ($self->{'axis'}->{'x2'}->{'plot_labels'} || $self->{'axis'}->{'x2'}->{'plot_ticks'}) {
    	# Draw X2 axis
	$self->{'gd'}->line (&{$self->{'transform'}}($self->{'plot'}->{'x'}, 
						     $self->{'plot'}->{'y'} - $self->{'plot'}->{'h'}), 
    	    	    	     &{$self->{'transform'}}($self->{'plot'}->{'x'} + $self->{'plot'}->{'w'}, 
						     $self->{'plot'}->{'y'} - $self->{'plot'}->{'h'}),
			     $self->{'fg_colour'}) if ($self->{'axis'}->{'x2'}->{'plot_base'} == 1);
    	}
    # Draw Y axis
    $self->{'gd'}->line (&{$self->{'transform'}}($self->{'plot'}->{'x'}, 
						 $self->{'plot'}->{'y'} - $self->{'plot'}->{'h'}),
    	    	    	 &{$self->{'transform'}}($self->{'plot'}->{'x'}, $self->{'plot'}->{'y'}),
			 $self->{'fg_colour'}) if ($self->{'axis'}->{'y'}->{'plot_base'} == 1);
    if ($self->{'axis'}->{'y2'}->{'plot_labels'} == 1 || $self->{'axis'}->{'y2'}->{'plot_ticks'} == 1) {
    	# Draw Y axis
	$self->{'gd'}->line (&{$self->{'transform'}}($self->{'plot'}->{'x'} + $self->{'plot'}->{'w'},
						     $self->{'plot'}->{'y'} - $self->{'plot'}->{'h'}),
    	    	    	     &{$self->{'transform'}}($self->{'plot'}->{'x'} + $self->{'plot'}->{'w'},
						     $self->{'plot'}->{'y'}), 
			     $self->{'fg_colour'}) if ($self->{'axis'}->{'y2'}->{'plot_base'} == 1);
    	}

}

=head2 clear_plot()

    $graph->clear_plot ();

Curve_plot::clear_plot is used to clear the whole plot. This can be useful if you want to create several plots without redefining all the parameters. For example, you can create a first plot then use 'clear_plot' to clean the plot, next use the 'add_dataset' or 'remove_dataset' then 'plot' again to build a new one that will be stored in another GIF file using 'save_gif' (see below).
Does not return a value.

=cut

sub clear_plot {
    my $self = shift;

    # Make a filled rec on the whole gif with background colour
    $self->{'gd'}->filledRectangle(0, 0, 
				   &{$self->{'transform'}}($self->{'width'} - 1, 
							   $self->{'height'} - 1),
				   $self->{'bg_colour'});

#     # Draw plot border
#     $self->{'gd'}->rectangle(0, 0, 
#                              &{$self->{'transform'}}($self->{'width'} - 1, 
#     	    	    	         $self->{'height'} - 1),
#                                $self->{'fg_colour'});
}

=head2 save_gif()

    $graph->save_gif ("filename.gif");

Curve_plot::save_gif is called once the plot has been created. The argument must be a file name where the GIF will be stored.
It returns 1 on success and 0 on failure.

=cut

sub save_gif {
    my $self = shift;
    my $fname = shift;
    local (*FPT);
    
    return 0 if (! open (FPT, "+>$fname"));

    binmode FPT if $needs_binmode;
    
    print FPT $self->{'gd'}->gif;
    
    close FPT;

    return 1;
}

sub _axis_evenlyspace {
    my $array = shift;
    my $size = shift;
    my $ax_min = shift;
    my $ax_max = shift;
    my $uthresh = shift;
    my $fontwidth = shift ;

    my @ret;
    my @data;
    my $inc;
    my $len;
    my $pos;
    my $val;
    my $min;
    my $diff;
    my $nb_add;
    my $arr_pos;
    my $thresh;
    
    if (! defined $size) {
	confess ("The size of the axis passed to _axis_evenlyspace in undef!\n"); }

    # If no labels, use passed axis min and max values
    $array = [$ax_min, $ax_max] if (! defined $array);
    
    # Get limits
    $len = @{$array};
    $pos = 0;

    # Get first defined value
    for ($pos = 0; $pos < $len; $pos++) {
	last if (defined $array->[$pos] && $array->[$pos] != 0); }

    $min = $array->[$pos];
    
    # Calculate theorical increment
    # Define lowest unit in X axis

    @data = split (//, $min);

    # Remove first digit -> correspond to unit 1
    do {$val = shift (@data)} while ($val !~ /\d/);
    
    $inc = 1;
    
    foreach $val (@data) {
	$inc /= 10 if ($val =~ /\d/); }

    # Evaluate how many entries to add
    $diff = $array->[$len-1] - $array->[0];
    $nb_add = int ($diff / $inc) + 1;
    
    # Define threshold on the basis on the font width and axis size
    $thresh = ($nb_add * $fontwidth + 2) / $size;
    
    # Add user threshold
    $uthresh = 1 if (! defined ($uthresh) || $uthresh < 1);
    $uthresh = $nb_add / 2 if ($uthresh > ($nb_add / 2));
    
    # Update threshold based on used defined one
    $thresh *= $uthresh;

    # Redefine array len
#    $nb_add /= $thresh;
    $nb_add = int ($nb_add / $thresh);
    
    # Define corresponding increment value
#    $inc *= $thresh;
    $inc = $diff / $nb_add;
    
    # Create new array with spaced values
    $val = $array->[0];
    $arr_pos = 0;

    for ($pos = 0; $pos < $nb_add + 1; $pos++) {
	$ret[$arr_pos] = $val;
	
	$val += $inc;
	$arr_pos++;
    	}
	
    # Last value is the max value of X.
    # This operation may lead to a sharp end of the X axis depending on 
    # how bad are the original X axis values and what is the user defined 
    # threshold, but this method cannot provide all solutions...
    
#    $ret[$arr_pos] = $array->[$len-1];
    
    return @ret;
}

sub _add_legend {
    my $self = shift;
    my $dataref = shift;
    my $name = shift;
    my $legend = shift;
    my $legend_ref = shift;
    
    # Get entry reference
    my $ref = $self->_get_data_ref ($dataref, $name);
    
    return 0 if (! defined $ref);
        
    # Set the legend
    $legend = $dataref->{'names'}->[$ref] unless defined $legend;
    
    # Keep it (Keep space for the 1st entry 'Legend"' tag)
    $ref++;
    $self->{$legend_ref}->{'entries'}->[$ref] = $legend;
    
    return 1;
}

=head2 add_legend()

    $graph->add_legend ("Curve1", "Legend1");

Curve_plot::add_legend add a curve in the legend. By default, if a legend is displayed, nothing is included, each curve (data set) has to be added to the legend. The first argument is the curve name or reference number (0 = first). The second argument is optional. If provided, it will be used for the particular curve in the legend, otherwise the curve name.

=cut

sub add_legend {
    my $self = shift;
    my $ref = shift;
    my $legend = shift;
        
    return $self->_add_legend ($self->{'gdata'}, $ref, $legend, 'legend');
}

=head2 add_legend2()

    $graph->add_legend2 ("Curve1", "Legend1");

Curve_plot::add_legend2 does the same as 'add_legend' except that it applies for the second Y axis.

=cut

sub add_legend2 {
    my $self = shift;
    my $ref = shift;
    my $legend = shift;
    
    return $self->_add_legend ($self->{'gdata2'}, $ref, $legend, 'legend2');
}

sub auto_colour {
    my $self = shift;
        
    my $nb_colours = $self->{'gdata'}->{'nb_sets'} + $self->{'gdata2'}->{'nb_sets'};
    my @rgb;
	
    if ($nb_colours == 0) {
	cluck ("No data set (yet) specified for auto colour definitions.\n");
	return 0;
    	}
    
    # Necessary if only 1 set
    $nb_colours = 2 if ($nb_colours == 1);
    
    my $max_col = scalar @{$base_colours};
    my $shift = 0;
    my $pos = 0;
    my $count = 0;
    
    # Loop on first data set
    for (my $col_pos = 0; $col_pos < $self->{'gdata'}->{'nb_sets'}; $col_pos++) {
	next if defined ($self->{'gdata'}->{'colours'}->[$col_pos]);
    	
	if ($base_colours->[$count][0] >= $shift) {
	    $rgb[0] = $base_colours->[$count][0] - $shift; }
	else {
	    $rgb[0] = $shift; }
	    
	if ($base_colours->[$count][1] >= $shift) {
	    $rgb[1] = $base_colours->[$count][1] - $shift; }
	else {
	    $rgb[1] = $shift; }

	if ($base_colours->[$count][2] >= $shift) {
	    $rgb[2] = $base_colours->[$count][2] - $shift; }
	else {
	    $rgb[2] = $shift; }

#    	print STDERR "pos $col_pos, count = $count, rgb = $rgb[0] - $rgb[1] - $rgb[2]\n";
	
	$self->{'gdata'}->{'colours'}->[$col_pos] = $self->{'gd'}->colorAllocate($rgb[0],$rgb[1],$rgb[2]);
	
	$count++;
	
	if ($count == $max_col) {
	    $count = 0;
	    $shift += 10;
	    }
	if ($shift == 250) {
	    cluck ("Reached maximum number of colours, stopped defining a colour for the remaining curves.\n");
	    last;
	    }
	}

    # Loop on second data set
    for (my $col_pos = 0; $col_pos < $self->{'gdata2'}->{'nb_sets'}; $col_pos++) {
	next if defined ($self->{'gdata2'}->{'colours'}->[$col_pos]);
    	
	if ($base_colours->[$count][0] >= $shift) {
	    $rgb[0] = $base_colours->[$count][0] - $shift; }
	else {
	    $rgb[0] = $shift; }
	    
	if ($base_colours->[$count][1] >= $shift) {
	    $rgb[1] = $base_colours->[$count][1] - $shift; }
	else {
	    $rgb[1] = $shift; }

	if ($base_colours->[$count][2] >= $shift) {
	    $rgb[2] = $base_colours->[$count][2] - $shift; }
	else {
	    $rgb[2] = $shift; }

#    	print STDERR "pos $col_pos, count = $count, rgb = $rgb[0] - $rgb[1] - $rgb[2]\n";
	
	$self->{'gdata2'}->{'colours'}->[$col_pos] = $self->{'gd'}->colorAllocate($rgb[0],$rgb[1],$rgb[2]);
	
	$count++;
	
	if ($count == $max_col) {
	    $count = 0;
	    $shift += 10;
	    }
	if ($shift == 250) {
	    cluck ("Reached maximum number of colours, stopped defining a colour for the remaining curves.\n");
	    last;
	    }
	}
    return 1;
}

sub _get_col_hex {
    my $self = shift;
    my $dataref = shift;
    my $ref = shift;
    my @rgb;
    my $r;
    my $g;
    my $b;
    
    $ref = $self->_get_data_ref ($dataref, $ref);
    
    return "" if (! defined $ref);

    @rgb = $self->{'gd'}->rgb($dataref->{'colours'}->[$ref]);
		
    if ($rgb[0] == 0) {
    	$r = "00"; }
    else {
	$r = sprintf "%02x", $rgb[0]; }	
    if ($rgb[1] == 0) {
    	$g = "00"; }
    else {
	$g = sprintf "%02x", $rgb[1]; }
    if ($rgb[2] == 0) {
    	$b = "00"; }
    else {
	$b = sprintf "%02x", $rgb[2]; }	
    
    return ($r . $g . $b);
}

=head2 get_col_hex()

    $hex_colour = $graph->get_col_hex ($curve);

Curve_plot::get_col_hex returns a string with hexadecimal colour of a curve. Useful for web page content colouring according to the curve. The single argument must be the curve number (0 for first, 1 for second, ...) or curve name. 
Example: print HTML '<FONT COLOR="#', $graph->get_col_hex ($curve), '">';

Returns empty string on failure.

=cut

sub get_col_hex {
    my $self = shift;
    my $ref = shift;
    
    return $self->_get_col_hex($self->{'gdata'}, $ref);
}

=head2 get_col_hex2()

    $hex_colour = $graph->get_col_hex2 ($curve);

Curve_plot::get_col_hex2 does the same as 'get_col_hex' except that it applies to the set of data on the second Y axis.

Returns empty string on failure.

=cut

sub get_col_hex2 {
    my $self = shift;
    my $ref = shift;
    
    return $self->_get_col_hex($self->{'gdata2'}, $ref);
}

sub _image_map_coord {
    my $self = shift;
    my $dataref = shift;
    my $x_axis = shift;
    my $y_axis = shift;
    my $entry = shift;
    my $rad = shift;
    my $spec = shift;

    my $x_scale;
    my $y_scale;
    my $x_data;
    my $y_data;
    my $x_base;
    my $y_base;
    my $x_min;
    my $y_min;
    my $y_max;
    my $x_inc;
    my $y_inc;
    my $len;
    my $pos;
    my $x;
    my $y;
    my $val;
    my $cpos;
    my @up;
    my @down;
    my @ret;
    
    # Check passed reference
    $entry = $self->_get_data_ref($dataref, $entry);
    
    return undef if (! defined $entry);
    
    my $dataset = $dataref->{'sets'}->{$dataref->{'names'}->[$entry]};
    # Get array ref
    $x_data = $dataset->{'x'};
    $y_data = $dataset->{'y'};

    if (defined $spec) {
	return undef if (ref ($spec) ne "ARRAY");
	$len = @{$spec};
	return undef if (! $len);
    	}
    else {
	$len = @{$y_data};
	@{$spec} = (0 .. $len-1);
    	}
    
    return undef if (! defined $x_data || ! defined $y_data);

    $x_scale = $x_axis->{'scaling'};
    $y_scale = $y_axis->{'scaling'};
    
    $x_base = $self->{'plot'}->{'x'};
    $y_base = $self->{'plot'}->{'y'};

    $x_min = $x_axis->{'min'};
    $y_min = $x_axis->{'min'};
    $y_max = $x_axis->{'max'};
    
    # Define increment ratio
    $x_inc = $x_axis->{'inc'};
    $y_inc = $y_axis->{'inc'};
    
    # Check radius
    $rad = 0 if (! defined $rad || $rad < 0);
    
    # 1st point is x-1, y-1 of 1st cood
    $cpos = $spec->[0];

    $val = $x_scale->($x_data->[$cpos]);

    $x = int ($x_base + (($val - $x_min) * $x_inc)) - 1;
    
    $val = $y_scale->($y_data->[$cpos]);
    
    $val = $y_max if ($val > $y_max);
    $val = $y_min if ($val < $y_min);

    $y = int ($y_base - (($val - $y_min) * $y_inc)) + 1;
    
    push (@up, $x-$rad, $y-$rad);
    
    $pos = 0;

    while ($pos < $len) {
	$cpos = $spec->[$pos];
	
	if (! defined $y_data->[$cpos] || ! defined $x_data->[$cpos]) {
	    $pos++;
	    next;
	    }
	
	$val = $x_scale->($x_data->[$cpos]);

	$x = int ($x_base + (($val - $x_min) * $x_inc));

	$val = $y_scale->($y_data->[$cpos]);
	
	$val = $y_max if ($val > $y_max);
	$val = $y_min if ($val < $y_min);
		
	$y = int ($y_base - (($val - $y_min) * $y_inc));	
	
	push (@up, $x-$rad, $y-$rad);
	unshift (@down, $x+$rad, $y+$rad);
	$pos++;
    	}
    
    # If got nothing, return undef
    return undef if (! @down);
    
    # last point is x+1, y+1 of last cood
    
    $cpos = $spec->[$pos-1];

    $val = $x_scale->($x_data->[$cpos]);
    
    $x = int ($x_base + (($val - $x_min) * $x_inc)) - 1;

    $val = $y_scale->($y_data->[$cpos]);
    
    $val = $y_max if ($val > $y_max);
    $val = $y_min if ($val < $y_min);
    

    $y = int ($y_base - (($val - $y_min) * $y_inc)) + 1;    
    
    push (@down, $x+$rad, $y+$rad);

    @ret = join (',', @up, @down);

    return \@ret;
}

=head2 imagemap_coord()

    my $coord = $graph->imagemap_coord ($curve, 3, [5,6,7,8]);
    print HTML <<XXX;
    <AREA href="#subset$curve"
       onMouseOver="self.status='subset$curve';return true;"
       onMouseOut="self.status='';return true;"
       shape="polygon" coords="@{$coord}">
XXX

Curve_plot::imagemap_coord allows to retrieve an array reference with the imagemap polygon coordinates from a subset of the points in a dataset. The first argument is the curve name or reference number (0 = first), the second argument is the radius of the map and the third argument can be the reference to an array with the positions in the dataset (0 first, 1 second, ...). If the third argument is not specified, the coordinates will be defined for the whole dataset.
On success, it returns a reference to an array with the coordinates of the map for the subset. On failure it returns undef.

=cut

sub imagemap_coord {
    my $self = shift;
    my $entry = shift;
    my $rad = shift;
    my $spec = shift;
    
    return $self->_imagemap_coord ($self->{'gdata'}, $self->{'axis'}->{'x'},
    	    	    	    	$self->{'axis'}->{'y'}, $entry, $rad, $spec);
}

=head2 imagemap_coord2()

    my $coord = $graph->imagemap_coord2 ($curve, 3, [5,6,7,8]);
    print HTML <<XXX;
    <AREA href="#subset$curve"
       onMouseOver="self.status='subset$curve';return true;"
       onMouseOut="self.status='';return true;"
       shape="polygon" coords="@{$coord}">
XXX

Curve_plot::imagemap_coord2 does the same as 'imagemap_coord' except that it applies on the curves obtained on the second Y axis.

On failure it returns undef.

=cut

sub imagemap_coord2 {
    my $self = shift;
    my $entry = shift;
    my $rad = shift;
    my $spec = shift;
    
    return $self->_imagemap_coord ($self->{'gdata2'}, $self->{'axis'}->{'x2'},
    	    	    	    	$self->{'axis'}->{'y2'}, $entry, $rad, $spec);
}

=head2 start_imagemap()

    print HTML $graph->start_imagemap ($map_name);

Curve_plot::start_imagemap returns a string used to initialize an imagemap definition. The single argument is the name of the image map. Warning: the IMG tag must have the same imagemap name. In this case: <IMG SRC="curves.gif" USEMAP="$map_name">
The default name defined by the module is ... "imgmap"!
Returns an empty string on failure.

=cut

sub start_imagemap {
    my $self = shift;
    my $name = shift;
    
    my $ret = "";
    
    if ($self->{'end_imgmap'} != 1) {
	cluck ("The previous imagemap $self->{'imgmap_name'} was not ended!\n");
	return "";
    	}

    $name = "imgmap" if (! defined $name);
    $self->{'imgmap_name'} = $name;
    $self->{'end_imgmap'} = 0;

    $ret .= '<MAP NAME="' . $name . '">' . "\n";
    
    return $ret;
}

=head2 end_imagemap()

    print HTML $graph->end_imagemap();

Curve_plot::end_imagemap must be called to finish the imagemap definition started with start_imagemap. No argument are required.
Returns "</MAP>\n".

=cut

sub end_imagemap {
    my $self = shift;

    $self->{'end_imgmap'} = 1;
    return ("</MAP>\n");
}

sub _imagemap_area {
    my $self = shift;
    my $dataref = shift;
    my $x_axis = shift;
    my $y_axis = shift;
    my $entry = shift;
    my $map_name = shift;
    my $args = shift;
    
    my $coord;
    my $name;
    my $map;
    my $rad;
    my $ref;
    my $key;
    
    $entry = $self->_get_data_ref ($dataref, $entry);
    
    return "" if (! defined $entry);
        
    # Define data set name
    $name = $dataref->{'names'}->[$entry];

    print "<!-- In imagemap_area: $name -->\n";
    
    # Define radius
    if (defined $args->{'radius'}) {
	$rad = $args->{'radius'}; }
    else {
	$rad = 0; }
    
    $ref = $dataref->{'sets'}->{$dataref->{'names'}->[$entry]};
    
    # Update radius with line width
    if (defined $ref->{'line_width'}) {
	$rad += $ref->{'line_width'}; }
    else {
	$rad += $self->{'line_width'}; }

    # Define coords
    $coord = $self->_imagemap_coord ($dataref, $x_axis, $y_axis, $entry, $rad);

    $map = '<AREA ';
    $map .= 'name="' . $map_name . '"' if (defined $map_name);
    $map .= "\n\t";

    # Check args and fill map string
    foreach $key (keys %{$args}) {
	if ($key eq "href") {
	    $map .= 'href="' . $args->{$key} . '"' . "\n\t"; }
	elsif ($key eq "onMouseOver") {
	    $map .= 'onMouseOver="' . $args->{$key} . '"' . "\n\t"; }
	elsif ($key eq "onMouseOut") {
	    $map .= 'onMouseOut="' . $args->{$key} . '"' . "\n\t"; }
    	}

    # Complete string
    $map .= 'shape="polygon"' . "\n\t" . 'coords="' . "@{$coord}" . '"' . "\n\t>\n";

    return $map;
}

=head2 imagemap_area()

    print HTML $graph->imagemap_area ($curve, "map1", 'href' => "#$curve", 'radius' => 2, 'onMouseOver' => "self.status='$curve_name';return true;'");

Curve_plot::imagemap_area returns the map area definition corresponding to the passed curve reference number (0 = first) or curve name. The call must be placed between the start_imagemap and end_imagemap calls.
The first argument must be the curve number (0 for first, 1 for second, ...) or curve name.
The second argument is optional: it is the area name.
The third argument is optional, however if not set, the map is useless. 
It must be a hash table with the following parameters:
Returns empty string on failure.

=over 4

=item href:

URL to follow when clicked on the map.
Default: none. Example: 'href' => '/data/curve1_details.html'

=item radius:

Set the radius in pixel to be used to define the map around the curve.
Default: 1 pixel. Example: 'radius' => 2

=item onMouseOver:

Javascript call when mouse over map.
Default: none. Example: 'onMouseOver' => 'self.status=\'Curve name\''

=item onMouseOut:

Javascript call when mouse leave map.
Default: none. Example: 'onMouseOut' => 'self.status=\'\''

=back

=cut

sub imagemap_area {
    my $self = shift;
    my $entry = shift;
    my $map_name = shift;
    my %args = @_;

    return $self->_imagemap_area($self->{'gdata'}, $self->{'axis'}->{'x'}, 
    	    		$self->{'axis'}->{'y'}, $entry, $map_name, \%args);
}

=head2 imagemap_area2()

    print HTML $graph->imagemap_area ($curve, "map1", 'href' => "#$curve", 'radius' => 2, 'onMouseOver' => "self.status='$curve_name';return true;'");

Curve_plot::imagemap_area2 does the same as 'imagemap_area' except that it applies on the set of data related to the second Y axis.

=cut

sub imagemap_area2 {
    my $self = shift;
    my $entry = shift;
    my $map_name = shift;
    my %args = @_;

    return $self->_imagemap_area($self->{'gdata2'}, $self->{'axis'}->{'x2'}, 
    	    		$self->{'axis'}->{'y2'}, $entry, $map_name, \%args);
}

=head2 get()

    print "X axis x,y = ", $graph->get ('x', 'x_axis'), ",", $graph->get('y', 'x_axis'), 
    	  " width = ", $graph->get('w', 'x_axis'), "\n";

    print "Y labels x,y = ", $graph->get ('x', 'y_labels'), ",", $graph->get('y', 'y_labels'), 
    	  " width = ", $graph->get('w', 'y_labels'), "height = ", $graph->get('h', 'y_labels'), "\n";

Curve_plot::get is used to retrieve information about components of the graph.
The first argument is the type of information to extract and can be :

=over 4

=item 'x'

X coordinate in pixels.

=item 'y'

Y coordinate in pixels.

=item 'w'

Width of the object in pixels.

=item 'h'

Height of the object in pixels.

=back

The second argument is the graph object and can be 'plot_area', 'x_axis', 'y_axis', 'x_labels' or 'y_labels'.

More objects will be soon implemented like the legends and the second set of axis.

=cut

sub get {
    my $self = shift;
    my $type = shift;
    my $obj = shift;
    
    return undef unless (defined $type && defined $obj);
    my $ref;
    my $ret;
    
    # What object?
    if ($obj eq "plot_area" || $obj eq "x_axis" || $obj eq "y_axis") {
    	$ret = $self->{'plot'}->{$type};
	}
    elsif ($obj eq "x_labels") {
    	if ($type eq 'x') { 
	    $ret = $self->{'plot'}->{'x'} - int ($self->{'font_height'} / 2); }
	elsif ($type eq 'y') {
	    $ret = $self->{'plot'}->{'y'} + $self->{'axis'}->{'x'}->{'lab_size'} + 6; }
	elsif ($type eq 'w') {
	    $ret = $self->{'plot'}->{'w'} + $self->{'font_width'}; }
	elsif ($type eq 'h') {
	    $ret = $self->{'axis'}->{'x'}->{'lab_size'}; }
	}
    elsif ($obj eq "y_labels") {
    	if ($type eq 'x') { 
	    $ret = $self->{'plot'}->{'x'} - $self->{'axis'}->{'y'}->{'lab_size'} - 6; }
	elsif ($type eq 'y') {
	    $ret = $self->{'plot'}->{'y'} + int ($self->{'font_height'} / 2); }
	elsif ($type eq 'w') {
	    $ret = $self->{'axis'}->{'y'}->{'lab_size'}; }
	elsif ($type eq 'h') {
	    $ret = $self->{'plot'}->{'h'} + $self->{'font_height'}; }
	}
	
    # Check if available
    return undef unless defined $ret;
    
    return $ret;
}

sub _simple_trans {
    my $val = shift;

    return $val;
}

sub _simple_format {
    my $val = shift;

    return $val;
}

1;

__END__

=head1 AUTHOR

Copyright (c) 1999 by Raphael Leplae <lp1@sanger.ac.uk>. All rights reserved.
This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

GIFgraph(1), Chart(1), Chart::Plot(1)

=cut
