
#
# Perl Module for EMBL::AC
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::AC;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::_LineGroup;


#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();


BEGIN {
    @ISA = ( 'Exporter', 'EMBL::_LineGroup' );

    %fields = (
	       accessionList => undef    # array of accession numbers
    );
}

sub initialize {
    my $self = shift;
    $self->{'accessionList'} = [];
}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::_LineGroup;

    $self->rebless($class);
    return $self;
}

sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $lineGroup = new EMBL::AC;
    my $line;

    $line = $$lineRef;
    while(substr($line, 0, 2) eq 'AC') {
	$line = EMBL::TextFunc->cleanLine($line);
	push @{$lineGroup->accessionList()}, EMBL::TextFunc->listFromText($line, ';\s', ';');
	$line = <$fh>;
	if(not defined $line) { warn "AC: End of stream"; last; }
    }
    $$lineRef = $line;
    return $lineGroup;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
    my $line = shift;

    foreach $line (EMBL::TextFunc->textFromList($self->accessionList(), '; ', ';', 75)) {
	print $fh "AC   $line\n";
    }

    return $fh;
}



1;  # says use was ok
__END__

=head1 NAME

EMBL::AC

=head1 DESCRIPTION

B<EMBL::AC> represents the AC (accession) lines within an embl entry as specified in
http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/ac_line.html

B<EMBL::AC> inherits from B<EMBL::_LineGroup>

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

=over

=item C<accessionList>

This is an array containing a list of all the accession numbers associated
with this entry.  The first member will be the primary accession number, and
any following are the secondary or group accession numbers.

=back

=head1 Functions

=over

=item fromEMBLStream

This is a module function that accepts a \*filehandle and \$string as arguments.
See EMBL::_LineGroup for a description of what they are for.  It is designed to be
invoked as

    $lg = EMBL::AC->fromEMBLStream($fh, \$line);

within the parser.

=item toEMBLStream

Writes the accessions to a stream in embl formatting.  It takes a \*filehandle as
its only argument.

    $acLine->toStream(\*STDOUT);
