
#
# Perl Module for EMBL::DR
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::DR;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::_LineGroup;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();


BEGIN {
    # @ISA has our inheritance.
    @ISA = ( 'Exporter', 'EMBL::_LineGroup' );

    %fields = (
	       #Insert field names here as field => value,
	       'dbID' => undef,         # the database
	       'primaryID' => undef,    # primary identifier
	       'secondaryID' => undef   # secondary identifier
    );

    # check that you have not tried to define names used by base classes
    EMBL::DR->check4Clashes();
}

sub initialize {
    my $self = shift;
    # place code here to initialize member variables
    # use the syntax $self->{'memberName'} = $val;
    # a good place to set those arrays and hashes to [] & {}
    #

}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::_LineGroup;

    $self->rebless($class);
    return $self;
}

sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $lineGroup = new EMBL::DR;
    my $line;

    $line = $$lineRef;
    @_ = $line =~ m/^DR   ([^;]+);\s*([^;]+);\s*(.*)\.$/;
    $lineGroup->dbID(shift @_);
    $lineGroup->primaryID(shift @_);
    $lineGroup->secondaryID(shift @_);

    $$lineRef = <$fh>;
    return $lineGroup;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;

    print $fh "DR   ", $self->dbID(), "; ",
    $self->primaryID(), "; ",
    $self->secondaryID(), ".\n";

    return $fh;
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::DR

=head1 DESCRIPTION

B<EMBL::DR> represents the DR (database reference) lines within an embl entry as specified in
http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/dr_line.html

EMBL::DR inherits from L<EMBL::_LineGroup>

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

For a complete description of the meanings of these variables, consult
the online documentation at http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/dr_line.html

=over

=item dbID

The database ID.  This will be something like SWISS-PROT or EPD

=item primaryID

This will be the ID of the entry within the database.  E.g. an accession number

=item secondaryID

The secondary ID - the entry name, gene symbol, release number.

=back

=head1 Functions

=over

=item fromEMBLStream

This is a module function that accepts a \*filehandle and \$string as arguments.
See EMBL::_LineGroup for a description of what they are for.  It is designed to be
invoked as

    $lg = EMBL::DR->fromEMBLStream($fh, \$line);

within the parser.

=item toEMBLStream

Writes the database reference to a stream in embl formatting.  It takes a \*filehandle as
its only argument.

    $drLine->toStream(\*STDOUT);

