
#
# Perl Module for EMBL::Entry.pm
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
#Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::Entry;

use vars qw($AUTOLOAD @ISA @EXPORT_OK);
use Exporter;
use Carp;
use strict;

use SPEM::Entry;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

#
# @ISA has our inheritance.
#

@ISA = ( 'Exporter', 'SPEM::Entry' );

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new SPEM::Entry;

    bless $self, $class;

    return $self;
}

sub ID {
    my $self = shift;
    my $index = $self->_indexFromStart('EMBL::ID');   # look for this line type

    defined $index &&
	return $self->get($index)->entryname();       # got one, so return what the user thinks of as the ID
    return undef;                                     # no ID line - return undef
}

sub AC {
    my $self = shift;
    my $index = $self->_indexFromStart('EMBL::AC');   # look for this line type
    
    defined $index &&
	return $self->get($index)->accessionList->[0];# got one, so return the users idea of the accession
    return undef;                                     # no AC line - return undef
}

sub FT {
    my $self = shift;
    my $index = $self->_indexFromEnd('EMBL::FeatureTable'); # look for the feature table object

    defined $index && 
	return $self->get($index);                    # got one, so return it
    return undef;                                     # no feature table - return undef.
}

sub SQ {
    my $self = shift;
    my $index = $self->_indexFromEnd('Sequence');     # look for the sequence object

    defined $index &&
	return $self->get($index);                    # got one - return it
    return undef;                                     # no sequence object - return undef.
}

sub writeLine {
    my $self = shift;
    my $fh = shift;
    my $line = shift;

    $line->toEMBLStream($fh);
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::Entry.pm

=head1 DESCRIPTION

B<EMBL::Entry> represents a single embl entry and provides an api for its modification.

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Typical example of usage

=for html
<PRE>
use Embl;
open(EFETCH, "efetch X68838 |") || die "could not open pipe to efetch";
$set = read Embl(EFETCH);
close EFETCH;
foreach $entry ($set->entries) {  # loop over each entry in $set - in this case 1
    print $entry->AC, " has sequence\n", $entry->SQ->seq(), "\n";
}
</PRE>

=head1 Member variables

=over

=item lineEntries

The array of line groups that make up this entry. Edit this and the entry will change.

=back

=head1 Common data access functions

=over

=item ID

Returns the id of this entry - the first value on the ID line.
It is equivelent to calling entryname on the EMBL::ID line object. 
If it can not find an ID line, then it will print a warning message,
and return undef.

=item AC

Returns the primary accession number for this embl entry.
If it can not find an AC line, then it will print a warning message,
and return undef.

=item FT

Returns the EMBL::FeatureTable obect that represents the feature table for this entry.
Edits to this table will alter the entry.
If it can not find a feature table, then it will print a warning message,
and return undef.

=item SQ

Returns the Sequence obejct that represents the sequence associated with this entry.
Edits to the sequence will alter the entry.
If it can not find a sequence object, then it will print a warning message,
and return undef.

=back

=head1 Editing and Query Functions

=over

=item addLineGroup

=for html
Add a line group object to this embl group.

Optionaly specify where to add it.

    $entry->addLineGroup($what [, $where]);

If where is absent, then the line group will be added at the end of the entry.
If it is START or END, it will be added accordingly.  If it is a number, then
this is taken to be an index (starting at 0) for where to add it.  Should the
index be outside the range of the entry, it will be added at the end.

=item removeLineGroup

Removes the line group at the specified index. Returns the object removed,
or C<undef>.

    $entry->removeLineGroup($where);

=item get

Returns the line group at the specified position.

    $lg = $entry->get($where);

The C<where> index starts at zero.  If you try to index a group beyond the end
of this object then a clear message to that effect will be raised, and the
return value will be  C<undef>.

=item getLine

Returns a line group object of a given type. Use:

    $line = $entry->getLine('EMBL::CC');

to get the first comment line. Optionaly, a seccond parameter can be passed which
causes the search for the next line to begin at that index.

    $line = $entry->getLine('EMBL::CC', 8);

will return the first comment line on or after the line with index 8.

=item all

Returns an array containing all of the lines of a given type. Use:

    @lines = $entry->all('EMBL::Reference');

This would put into @lines an array containing all of the EMBL::Reference objects
within this entry. You could use this in a foreach loop very cleanly -

    foreach $ref ($entry->all('EMBL::Reference')) {
	# stuff with $ref
    }

=back

=head1 File Output

=over

=item toStream

Writes this entry to a stream.  Pass it a reference to a glob of a filehandle.
It will return the reference when done.  The entry will be writen to the stream
by calling toEMBLStream on each of the line groups contained in turn, and then
printing the // line.

=back

=head1 Functions for Developers

=over

=item _indexFromStart

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

This function takes a string that represents a class type, and returns an
index that can be used with C<get> which represents the first line of
that type in this object. Optionaly, you can parse it a starting index.  The
search will then start at that position (as the perl built in function C<index>
does).

=item _indexFromEnd

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

This function takes a string that represents a class type, and returns an
index that can be used with C<get> which represents the last line of
that type in this object. Optionaly, you can parse it a starting index.  The
search will then start at that position (as the perl built in function C<index>
does).

=back
