
#
# Perl Module for EMBL::Organism
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::Organism;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::_LineGroup;
use EMBL::TextFunc;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();


BEGIN {
    @ISA = ( 'Exporter', 'EMBL::_LineGroup' );

    %fields = (
	       #Insert field names here as field => undef,
	       species1 => undef,
	       classification1 => undef, # will be an array
	       species2 => undef,
	       classification2 => undef, # will be an array
	       organel => undef,
    );

  EMBL::Organism->check4Clashes();
}

sub initialize {
    my $self = shift;
    $self->{'classification1'} = [];
    $self->{'classification2'} = [];
}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = EMBL::_LineGroup::new $class;

    $self->rebless($class);
    return $self;
}

sub fromEMBLStream{
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $lineGroup = new EMBL::Organism;
    my $line;
    my $tag;
    my $organismNo = 1;

    $line = $$lineRef;
    while(1) {
	$tag = substr($line, 0, 2);
	$line = EMBL::TextFunc->cleanLine($line);
#	print "got line $tag, $line\n";
	if ($tag eq 'OS') {                              # species line
#	    print "OS line\n";
	    unless(defined $lineGroup->species1()) {     # first species line
		$lineGroup->species1($line);             # so set the first species name
	    } else {                                     # second species line
		$lineGroup->species2($line);             # so set the second species name
		$organismNo = 2;
	    }
	    next; 
	}
	if ($tag eq 'OC') {                              # classification line
#	    print "OC line\n";
	    if ($organismNo == 1) {                      # first classification line group
		push @{$lineGroup->classification1}, EMBL::TextFunc->listFromText($line, ';\s', '\.|;');
	    } else {                                     # second classification line group
		push @{$lineGroup->classification2}, EMBL::TextFunc->listFromText($line, ';\s', '\.');
	    }
	    next;
	}
	if ($tag eq 'OG') {
#	    print "OG Line\n";
	    $lineGroup->organel($line);
	    next;
	}
	if ($tag eq 'XX') {                              # just throw XX lines away
#	    print "XX line\n";
	    next;
	}
#	print "Matched nothing\n";
	last; # not a tag I can deal with
    } continue {
	$line = <$fh>;
	if(not defined $line) { warn "Organism: End of stream"; last; }
    }

    $$lineRef = $tag."   ".$line;
    return $lineGroup;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
    my $line;

    # first OS line
    print $fh "OS   ", $self->species1, "\n";
    # first OC group
    foreach $line (EMBL::TextFunc->textFromList($self->classification1(), '; ', '.', 75)) {
	print $fh "OC   $line\n";
    }

    # if there was a second species
    if($self->species2) {
	# print XX spacer
	print $fh "XX\n";
	# second OS line
	print $fh "OS   ", $self->species2, "\n";
	foreach $line (EMBL::TextFunc->textFromList($self->classification2(), ';', ';', 75)) {
	    print $fh "OC   $line\n";
	}
    }

    # if there is an OG specification
    if($self->organel) {
	# print an XX spacer
	print $fh "XX\n";
	# print the OG line
	print $fh "OG   ", $self->organel, "\n";
    }

    print $fh "XX\n";      # the fromEMBLStream eats the trailing XX, so we put it back in

    return $fh;
}


1;  # says use was ok
__END__

=head1 NAME

EMBL::Organism

=head1 DESCRIPTION

B<EMBL::Organism> represents the OS, OC and OG line group.

EMBL::Organism inherits from B<EMBL::_LineGroup>

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

=over

=item species1

The species of origin for this entry.

=item classification1

The classification of species1.  This is stoored as an array of 'nodes'
in the classification of the species, going from general to specific
as you move through the array.

=item species2

The name of the second species (only defined if the source organism
was a hybrid).

=item classification2

The classification of the second source species (again, only defined
if it is a hybrid).  This is in exactly the same format as classification1

=item organel

If not genomic, the organel or plasmid of origin.  Undefined otherwise.

=back

=head1 Functions

=over

=item fromEMBLStream

Reads in the entire set of organism lines, until a non XX, OS, OG or OC line is found.

=item toEMBLStream

Writes the organism lines to a stream, complete with trailing XX (as it eats the trailing
XX).
