
#
# Perl Module for EMBL::Set
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::Set;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use SPEM::Set;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();


@ISA = ( 'Exporter', 'SPEM::Set' );

my %fields = (
	   ACHash => undef,
	   );

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $parent = new SPEM::Set;
    my %permitted = (%{$parent->{'_permitted'}}, %fields);
    my $self = {
	%$parent,
	%fields,
    };
    
    $self->{'_permitted'} = \%permitted;
    $self->{'ACHash'} = {};

    bless $self, $class;

    return $self;
}

sub addEntry {
    my $self = shift;
    my $entry = shift;
    my $ac;

    if( (@_ != 0) || (not ref $entry) ) {
	carp 'Usage: $set->removeEntry($entry)';
	return undef;
    }

    $ac = $entry->AC();

    if(exists $self->ACHash->{$ac}) {
	carp "AC $ac is already being used.  This entry will not be added.";
	return undef;
    }

    $self->ACHash->{$ac} = $entry;

    SPEM::Set::addEntry($self, $entry); # must call super class to add entry there

    return 1;
}

sub removeEntry {
    my $self = shift;
    my $entry = shift;
    my $ac;

    if( (@_ != 0) || (not ref $entry) ) {
	carp 'Usage: $set->removeEntry($entry)';
	return undef;
    }

    $ac = $entry->AC();

    unless(exists $self->ACHash->{$ac}) {
	carp "AC $ac does not appear in this set.";
	return undef;
    }

    delete $self->ACHash->{$ac};

    SPEM::Set::removeEntry($self, $entry); # must also forward this to the superclass

    return 1;
}

sub forAC {
    my $self = shift;
    my $ac = shift;

    if( (@_ != 0) || (not defined $ac) ) {
	carp 'Use: $set->forAC($ac)';
	return undef;
    }

    if(exists $self->ACHash->{$ac}) {
	return $self->ACHash->{$ac};
    } else {
	carp "Could not find AC:$ac in this set.";
	return undef;
    }
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::Set

=head1 DESCRIPTION

B<EMBL::Set> represents a set of EMBL::Entry objects. It can be produced from a stream using EMBL::Parser, or you can make an empty one yourself.

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Functions

=head2 Adding, removing and retrieving entries

=over

=item addSet 

Adds all of the entries in one set to this set.  Will predominantly be
used in one of two ways:

    $set1->addSet($set2);
    $set->addSet( $parser->fromEMBLStream(\*FH) );

=item addEntry

Adds an SPEM::Entry object to this Set.
It returns undef if something went wrong, and true if it was OK
Warning messages will be printed if the entry to be added has the same ID or
accession number as an entry already present.  In this case, undef will be returned,
and the entry will I<not> be added.

    $set->addEntry($entry);

=item removeEntry

Removes an SPEM::Entry object from this set.
It return undef on failure, and true on success.
If the set does not contain the entry then it will print a warning message,
and return C<undef>.

If for any reason the entry's ID or AC has changed,
the entry may become unremovable and unaccessable - this is a known bug.

    $set->removeEntry($entry);

=item forID

Returns the entry in this set associated with the specified ID.  Will
print a warning message and return C<undef> should this set not find
the entry.

=item forAC

Returns the entry in this set associated with the specified AC. Will
print a warning message and return C<undef> should this set not find
the entry.

=back

=head2 Usefull funcitons

=over

=item entries

Returns an array of the entries contained within this set.  The array
(but not the entries contained within it) can be hacked around,
without the Set being modified. If you modify any of the entries within
the array, they will cange within the set as well.

    @entries = $set->entries();        # puts all of the entries into @entryies
    ($entry) = $set->entries();        # quite a usefull trick when you are fairly sure that the
                                       # set contains one entry such as parsing from efetch

=item empty

Returns true if this set is empty, and false otherwise.  Usefull when
looping over a set and adding/removing entries. This function is
prototyped, so can be called as follows

 while(not empty $entrySet && other_conditions) { ... }

although, the perl purists would probably argue that you should use

 while(not $entrySet->empty() && other_conditions) { ... }

=back

=head1 Member variables

=for html <img src = "http://www.sanger.ac.uk/~mrp/magic.gif">This
information is only aplicable to developers. There is no need for
anybody to access these variables outside of the class.<P>

=over

=item IDHash

This contains a hash of all the entries contained by their respective
ID values.

=item ACHash

This contains a hash of all the entries contained by their respective
AC values.

=back
