
### Hum::Pfetch

package Hum::Pfetch;

use strict;
use Carp;
use Hum::EMBL;
use Hum::Conf qw{ PFETCH_SERVER_LIST PFETCH_ARCHIVE_SERVER_LIST };
use IO::Socket;
use Exporter;

use vars qw{@ISA @EXPORT_OK};
@ISA = ('Exporter');
@EXPORT_OK = qw{
    get_Sequences
    get_descriptions
    get_lengths
    get_EMBL_entries
    };

sub get_server {
    return _connect_to_server(
        'pfetch',
        $PFETCH_SERVER_LIST);
}

sub get_archive_server {
    return _connect_to_server(
        'pfetch arvhive',
        $PFETCH_ARCHIVE_SERVER_LIST);
}

sub _connect_to_server {
    my( $type, $server_list ) = @_;
    local $^W = 0;
    
    foreach my $host_port (@$server_list) {
        my($host, $port) = @$host_port;
        #warn "Trying '$host:$port'\n";
        my $server = IO::Socket::INET->new(
            PeerAddr => $host,
            PeerPort => $port,
            Proto    => 'tcp',
            Type     => SOCK_STREAM,
            Timeout  => 10,
            );
        if ($server) {
            $server->autoflush(1);
            return $server;
        }
    }
    confess "No $type servers available";
}

sub get_Sequences {
    my( @id_list ) = @_;
    
    confess "No names provided" unless @id_list;
    
    my $server = get_server();
    print $server "-q @id_list\n";
    my( @seq_list, @missing_i );
    for (my $i = 0; $i < @id_list; $i++) {
        chomp( my $seq_string = <$server> );
        if ($seq_string eq 'no match') {
            # Add to list of indexes of missing sequences
            push(@missing_i, $i);
        } else {
            my $seq = Hum::Sequence->new;
            $seq->name($id_list[$i]);
            $seq->sequence_string($seq_string);
            $seq_list[$i] = $seq;
        }
    }
    
    # Try for any missing sequences from the archive server.
    if (@missing_i) {
        $server = get_archive_server();
        print $server "-q @id_list[@missing_i]\n";
        for (my $i = 0; $i < @missing_i; $i++) {
            my $id_index = $missing_i[$i];
            chomp( my $seq_string = <$server> );
            if ($seq_string ne 'no match') {
                my $seq = Hum::Sequence->new;
                $seq->name($id_list[$id_index]);
                $seq->sequence_string($seq_string);
                $seq_list[$id_index] = $seq;
            }
        }
    }
    
    return @seq_list;
}

sub get_descriptions {
    my( @id_list ) = @_;
    
    confess "No names provided" unless @id_list;
    
    my $server = get_server();
    print $server "-D @id_list\n";
    my( @desc_list );
    for (my $i = 0; $i < @id_list; $i++) {
        chomp( my $desc = <$server> );
        if ($desc eq 'no match') {
            $desc_list[$i] = undef;
        } else {
            $desc_list[$i] = $desc;
        }
    }
    return @desc_list;
}

sub get_lengths {
    my( @id_list ) = @_;
    
    confess "No names provided" unless @id_list;
    
    my $server = get_server();
    print $server "-l @id_list\n";
    my( @length_list );
    for (my $i = 0; $i < @id_list; $i++) {
        chomp( my $length = <$server> );
        if ($length eq 'no match') {
            $length_list[$i] = undef;
        } else {
            $length_list[$i] = $length;
        }
    }
    return @length_list;
}

sub get_EMBL_entries {
    my( @id_list ) = @_;
    
    confess "No names provided" unless @id_list;
    
    my $parser = Hum::EMBL->new;
    
    my $server = get_server();
    print $server "-F @id_list\n";
    my( @entries );
    for (my $i = 0; $i < @id_list; $i++) {
        my $entry = join '', <$server>;
        if ($entry eq "no match\n") {
            $entries[$i] = undef;
        } else {
            my $embl = $parser->parse(\$entry)  
                or confess "nothing returned from parsing '$entry'";
            $entries[$i] = $embl;
        }
    }
    return @entries;
}

1;

__END__

=head1 NAME - Hum::Pfetch

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

