
#
# Perl Module for SPEM::Set
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package SPEM::Set;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

# @ISA has our inheritance.
@ISA = ( 'Exporter' );

my %fields = (
	      #Insert field names here as field => undef,
	      IDHash => undef, # hash by ID of all entries
	      );

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = { '_permitted' => \%fields,
		 %fields };
    $self->{'IDHash'} = {};      # put a hash into IDHash


    bless $self, $class;

    return $self;
}

#
# Autoload up my field name set/get
#
sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) || carp "Ok, $self is not an object of mine!";
    my $name = $AUTOLOAD;

    # don't propagate DESTROY messages...

    $name =~ /::DESTROY/ && return;

    $name =~ s/.*://; #get only the bit we want
    unless (exists $self->{'_permitted'}->{$name} ) {
	carp "In type $type, can't access $name - probably passed a wrong variable into Sequence  ";
    }
    if (@_) {
	return $self->{$name} = shift;
    } else {
	return $self->{$name}
    }
}

sub addEntry {
my $usage = 'Usage: $set->addEntry($entry)';
    my $self = shift || carp $usage;
    my $entry = shift || carp $usage;
    my $id;

    if( @_ != 0 ) {
	carp $usage;
	return undef;
    }

    $id = $entry->ID();

    if(exists $self->IDHash->{$id}) {
	carp "ID $id is already being used.  This entry will not be added.";
	return undef;
    }

    $self->IDHash->{$id} = $entry;

    return 1;
}

sub removeEntry {
    my $usage = 'Usage: $set->removeEntry($entry)';
    my $self = shift || carp $usage;
    my $entry = shift || carp $usage;
    my $id;

    if( (@_ != 0) || (not ref $entry) ) {
	carp $usage;
	return undef;
    }

    $id = $entry->ID();

    unless(exists $self->IDHash->{$id}) {
	carp "ID $id does not appear in this set.";
	return undef;
    }

    delete $self->IDHash->{$id};

    return 1;
}

sub addSet {
    my $usage = 'Usage: $set->addSet($set)';
    my $self = shift || carp $usage;
    my $set = shift || carp $usage;
    my $entry;

    if( @_ != 0) {
	carp $usage;
	return undef;
    }

    for $entry ($set->entries) {
	$self->addEntry($entry);
    }

    return 1;
}

sub entries {
    my $self = shift;
    my @entries = values %{$self->IDHash()};

    return @entries;
}

sub forID {
    my $self = shift;
    my $id = shift;

    if( (@_ != 0) || (not defined $id) ) {
	carp 'Use: $set->forID($id)';
	return undef;
    }

    if(exists $self->IDHash->{$id}) {
	return $self->IDHash->{$id};
    } else {
	carp "Could not find $id in this set.";
	return undef;
    }
}

sub empty($) {
    my $self = shift;
    return $self->entries() == 0;
}

sub toStream {
    my $self = shift;
    my $fh = shift;
    my $entry = shift;

    foreach $entry ($self->entries()) {
	$entry->toStream($fh);
    }
}

1;  # says use was ok
__END__

=head1 NAME

SPEM::Set

=head1 DESCRIPTION

B<SPEM::Set> is part of the core modules that deal with parsing and
representing SWISSPROT and EMBL formatted information. See <SPEM::Spem>
for an overview. Throughout this documentation, the concrete example
of L<EMBL::Set> will be used for illustration.

B<SPEM::Set> is intended as the baseclass for objects that represent a
group of SPEM::Entry derived objects. It provides basic functionality
for adding and removing entries, and for accessing them. A derived
class may well provide a much richer set of access, sorting or filtering
functions. There are no virtual functions declared here.

You can largely 'inherit' this documentation into a derived class.

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Functions

=over

=item addEntry

Adds an SPEM::Entry object to this Set.
It returns undef if something went wrong, and true if it was OK
Warning messages will be printed if the entry to be added has the same ID or
accession number as an entry already present.  In this case, undef will be returned,
and the entry will I<not> be added.

=item removeEntry

Removes an SPEM::Entry object from this set.
It return undef on failure, and true on success.
If the set does not contain the entry then it will print a warning message,
and return C<undef>.

If for any reason the entry's ID or AC has changed,
the entry may become unremovable and unaccessable - this is a known bug.

=item entries

Returns an array of the entries contained within this set.  The array
(but not the entries contained within it) can be hacked around,
without the Set being modified.

=item empty

Returns true if this set is empty, and false otherwise.  Usefull when
looping over a set and adding/removing entries. This function is
prototyped, so can be called as follows

 while(not empty $entrySet && other_conditions) { ... }

although, the perl purists would probably argue that you should use

 while(not $entrySet->empty() && other_conditions) { ... }

=item forID

Returns the entry in this set associated with the specified ID.  Will
print a warning message and return C<undef> should this set not find
the entry.

=item addSet 

Adds all of the entries in one set to this set.  Will predominantly be
used in one of two ways:

    $set1->addSet($set2);
    $set->addSet( $parser->fromEMBLStream(\*FH) );

=back

=head1 Things to provide in a subclass

=over

=item other forms of access

This base class only provides the C<forID> access function. You may
wish to provide a forAC or forOrganism etc. function. These would
presumably return either single entries, or lists of entries.

If you are consistently writing methods like this, though, you should
consider writing a filter function (as for featureTable and feature)
to deal with varied queries.

=back

=head1 Member variables

=for html <img src = "http://www.sanger.ac.uk/~mrp/magic.gif">This
information is only aplicable to developers. There is no need for
anybody to access these variables outside of the class.<P>

=over

=item IDHash

This contains a hash of all the entries contained by their respective
ID values.

=back
