package Spangle::Ace;

=pod 

=head1 NAME

Spangle::Ace - Utility routines for humace

=head1 SYNOPSIS

  use Spangle::Ace;

Connect to acedb databases or servers :

  my $db = Spangle::Ace::connect('humsrv1',210000); 
  my $db = Spangle::Ace::connect('','',"/home/michele/acedb")

Delete subsequences in clones according to a string pattern
This deletes all subsequence tags in clone $clone and also the 
subsequence objects themselves.

  my $clone = new Spangle::Clone;
     $clone->id ("ff65b7");
     $clone->acedb($db);
     $clone->fetch_subsequences;

  my $str = Spangle::Ace::deleteSubsequences($clone,"SPAN");

Get the difference in two dates (YYYY-MM-DD). Returns the result in days.
If the inputs to the function are $date1 and $date2 then the result will
be positive if $date1 occurs earlier than $date2 and negative if $date1 is 
later than $date2.

  my $diff = Spangle::Ace::date_diff("1999-03-13","1998-07-30")

Return an array of clone id strings that have analysis_file tags after
a certain date.  Useful for updating analysis. Takes an ace database handle
and a date string (YYYY-MM-DD)

  my @newseqs = Spangle::Ace::updated_clones($db,"1999-07-03");

Returns the date in ace format (YYYY-MM-DD);

  my $date  = Spangle::Ace:get_date

=head1 DESCRIPTION

Humace database utilities

=head1 CONTACT

Describe contact details here

=head1 APPENDIX

The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use Ace;
use strict;
use Date::Manip;

=pod 

=head1 Methods in Spangle::Ace

=head2 connect

  Title   : connect
  Usage   : $db = Spangle::Ace::connect($host,$port)
  Function: Connects to an ace database or server
  Returns : Ace database object
  Args    : hostname,port number, path to database

=cut

sub connect {
  my ($host,$port,$path) = @_;

  $host = 'humsrv1' unless $host;
  $port = 310000    unless $port;

  my $db;

  if ($path) {
    my $program='giface';
    if($ENV{'LACE_LOCAL'}){
	$program='giface';
    }
    print(STDERR "Connecting to database in directory $path\n");
    $db = Ace->connect(-path          => $path,
                       -program       => $program,
                       #-program       => '/nfs/disk100/acedb/RELEASE.SUPPORTED/bin.ALPHA_4/giface',
                       #-program       => '/nfs/disk100/acedb/RELEASE.SUPPORTED/bin.ALPHA_4/tace',
                       #-program       => 'tace',
                       );
    $db || die "Can't connect to ace database $path";
  } else {
    print(STDERR "Connecting to port $port on $host\n");
    $db = Ace->connect(-host    => $host,
		       -port    => $port,
		       -query_timeout => 1200);

    $db || die "Can't connect to ace database on $host : $port";
  }


  $db->auto_save(0);
  return $db;
}

=pod

=head2 deleteSubsequences

  Title   : deleteSubsequences
  Usage   : my $acestr = Spangle::Ace::deleteSubsequences($clone,"SPAN");
  Function: Creates ace command to delete a subsequence from a clone object
  Returns : String
  Args    : string,string

=cut


sub deleteSubsequences {
    my ($clone,$patt) = @_;

    my $str1;
    my $str2;

    $str1 .= "\nSequence : " . $clone->id . "\n";
    

    foreach my $g ($clone->eachSubsequence) {
	if ($g->source_tag =~ /$patt/) {
	    $str1 .= "-D Subsequence " . $g->source_tag . "\n";
	    $str2 .= "\n-D Sequence "  . $g->source_tag . "\n";
	}
    }

    $str1 .= $str2;
    return $str1;
}

=pod

=head2 date_diff

  Title   : date_diff
  Usage   : my $diff = Spangle::Ace::date_diff($date1,$date2)
  Function: Calculates the difference in days between two dates (YYYY-MM-DD)
  Returns : int. -ve for $date1 > $date2, +ve for $date1 < $date2
  Args    : string,string (Dates are in format YYYY-MM-DD)

=cut

sub date_diff {
  my ($date1,$date2) = @_;

  # dates are in the format yyyy-mm-dd

  my $d1 = Date::Manip::ParseDate($date1);
  my $d2 = Date::Manip::ParseDate($date2);

  # This returns a number 0:0:Weeks:Days:Hours:Minutes:Seconds
  my $dcmp  = Date::Manip::DateCalc($d1,$d2);

  # return this value in days

  my @time = split(/:/,$dcmp);

  my $wks  = $time[2];
  my $dys  = $time[3];

  my $sign = 1;

  if ($dcmp =~ /^-/) {
    $sign = -1;
  }
  
  return ($sign * 7 * $wks + $dys);

}

=pod

=head2 updated_clones

  Title   : updated_clones
  Usage   : my @newseqs = Spangle::Ace::updated_clones($db,$since)
  Function: Finds all clones that have analysis newer than a date ($since)
  Returns : Array of strings
  Args    : Ace db handler, date string (yyyy-mm-dd)

=cut

sub updated_clones {
  my ($db,$since) = @_;

  # since is a date in yyyy-mm-dd format
  # AQL query for the clones plus analysis dates
  
  #  my $query = "select s,s->analysis_files[1],s->analysis_files[2] from s in class genome_sequence";

  my $query1 = "find genome_sequence";
  my $query2 = "show -a analysis_files";
 
             $db->raw_query($query1);
  my $data = $db->raw_query($query2);

  my @data  = split(/\n/,$data);
  my $sequence;
  my %newseqs;

  foreach my $d (@data) {
    if ($d =~ /^Sequence : +\"(.*)\"/) {
      $sequence = $1;
    } elsif ($d =~ /^Analysis_files/) {
      my @tmp = split(' ',$d);
      
      my $file = $tmp[1];
      my $date = $tmp[2];
      
      my $diff = Spangle::Ace::date_diff($since,$date);

      if ($diff > 0) {
	$newseqs{$sequence} = 1;
      }
    }  
  }
  return (keys %newseqs);
}

=pod

=head2 get_date

  Title   : get_date
  Usage   : my $date  = Spangle::Ace::get_date;
  Function: Returns current date in the format yyyy-mm-dd
  Returns : String
  Args    : none

=cut

sub get_date {
  my $date  = `date '+%Y-%m-%d'`;
  return $date;
}

=pod

=head2 get_clones

  Title   : get_clones
  Usage   : my @clone = Spangle::Ace::get_clones($db)
  Function: Returns all clone ids in current database
  Returns : Array of strings
  Args    : Ace db handler

=cut

sub get_clones {
  my ($db) = @_;

  my $query1 = "find genome_sequence";
  my $query2 = "list -a";

             $db->raw_query($query1);
  my $data = $db->raw_query($query2);
  

  my @data  = split(/\n/,$data);
  my @clones;

  foreach my $d (@data) {
    if ($d =~ /^Sequence : +\"(.*)\"/) {
      push(@clones,$1);
    }
  } 
  return @clones;
}

=head2 get_clone

  Title   : get_clone_object
  Usage   : my @out = Spangle::Ace::get_clone($db,$cloneid);
  Function: Gets the ace features for clone $cloneid, from ace database $db
  Returns : Array of string
  Args    : Ace db handle,string

=cut


sub get_clone {
  my ($db,$clone) = @_;

  my @query;

  push(@query,"find Genome_sequence $clone");
  push(@query,"show -a");
  push(@query,"follow subsequence");
  push(@query,"show -a");  

 
  my @data;

  foreach my $q (@query) {
    my $tmp = $db->raw_query($q);
    $tmp =~ s/\0//g;
    
    my @tmp = split(/\n/,$tmp);

    foreach my $line (@tmp) {
      push(@data,$line  . "\n");
    }
  }  

  @query = ();
  
  push(@query,"find Genome_sequence $clone");
  push(@query,"DNA");
  
  my $data = "";

  foreach my $q (@query) {
    my $tmp = $db->raw_query($q);
    my @tmp2 = split(/\n/,$tmp);
    
    foreach my $tmp3 (@tmp2) {
      if ($tmp3 !~ /\/\//  && 
          $tmp3 !~ /^\n$/  && 
          $tmp3 !~ /:/     &&
	  $tmp3 !~ /^>/) {
	$data .= $tmp3;
      }
    }
  }
  
  $data =~ s/\0//g;
  $data =~ s/\n//g;
  $data =~ s/ //g;
  $data =~ s/(.{72})/$1\n/g;
  
  push(@data,"\nDNA : $clone\n$data\n");
  push(@data,"\n\n");
  

  return @data;
}


=head2 get_clone_gff

  Title   : get_clone_gff
  Usage   : my @out = Spangle::Ace::get_clone_gff($db,$cloneid);
  Function: Gets the ace features for clone in gff format
  Returns : Array of string
  Args    : Ace db handle,string

=cut


sub get_clone_gff {
  my ($db,$clone) = @_;

  my @query;

  push(@query,"gif seqget $clone ; seqfeatures");

  my @data;

  foreach my $q (@query) {

    my @tmp = split(/\n/,$db->raw_query($q));
    
    foreach my $line (@tmp) {
      $line =~ s/\0//;
      push(@data,$line . "\n");
    }
  }

  return @data;
}

sub input_clones {
  my($infile,@args) = @_;

  my @clones;
  
  if ($infile) {
    open(IN,"<$infile") || die "Couldn't open input file $infile";
    
    while (<IN>) {
      chomp;
      if ($_ =~ / /) {
	warn ("Error in input file: clone name \"$_\" contains spaces\n");
      } elsif ($_ !~ /^$/) {
	push(@clones,$_);
      }
    }
  }
  push(@clones,@ARGV);
  return @clones;
}

sub open_filehandle {
  my ($acefile,$append) = @_;
  my $ace;

  if ($acefile) { 
    if ($append) {
      open(ACE,">>$acefile"); 
    } else {
      open(ACE,">$acefile"); 
    }
    $ace = \*ACE;
  } else {
    $ace = \*STDOUT;
  }
  
  return $ace;
}


sub fetch_link {
  my ($linkid,$db) = @_;
  
  my $linky = $db->fetch('Link',$linkid);
  
  die "Can't find link $linkid" unless defined($linky);
  
  my $linkobj = new Spangle::Clone;
  my $linkcoords;
  my $link;
  
  $linkobj->id   ($linkid);
  $linkobj->acedb($db);
  $linkobj->fetch_subsequences;
  
  return $linkobj;
}
  

1;
