package Spangle::Analysis::DotPlot;

use GD;

=pod

=head1 NAME
  
Spangle::Analysis::DotPlot - creates a coloured dotplot gif from a list
                               of matches

=head1 SYNOPSIS

  Creating DotPlot objects
  
  my $dp = new Spangle::Analysis::DotPlot(-length1 => $len1,   # length of sequence 1
					  -length2 => $len2,   # length of sequence 2
					  -size    => $size,   # size of gif in pixels
					  -start1  => $start1, # start coord in sequence 1 (for partial plots)
					  -start2  => $start2, # start coord in sequence 2 (for partial plots)
					  -end1    => $end1,   # end coord in sequence 1 (for partial plots)
					  -end2    => $end2,   # end coord in sequence 2 (for partial plots)
					  -col     => $col,    # colour scheme (1 = ? , 2 = ?);
					  );

  $dp->read_clones  ($clonefile);
  $dp->read_matches ($matchfile);     # or $dp->read_MSPcrunch($mspfile);
  $dp->dotplot      ($giffh);
  $dp->print_matches($out);

=head1 The Clone Coordinate file
 
The clone file is optional and is used to define the clone tiling path.
The format of the clone file is one line per clone with each line having
the following fields

  clone_name start_coord end_coord

e.g.

  AP000522 1     37693
  AP000523 37694 76727

=head1 The Match File

The match file lists all the matches between the 2 sequences
we want to dotplot. These can be generated using your
favourite sequence matching program.  Each line in the file
represents one match between the 2 sequences and the format
of the line is :

  start_coord_in_sequence1 start_coord_in_sequence2 length score

If the length is negative then this is a reverse strand match.

e.g.

  3754607 3754607 44 220
  3754819 3754819 59 295
  3755166 3755166 855 4275
  3756504 3756504 2177 10714
  3758851 3758851 229 1145
  3753718 14343454 -51 157
  3257935 3257935 931 4655
  3259255 3259255 70 350

  
=head1 DESCRIPTION

A module to create a coloured dotplot from
a series of ungapped match segments.

=head1 CONTACT

Michele Clamp (michele\@sanger.ac.uk)

=head1 APPENDIX

The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use strict;  
use vars   qw(@ISA);

use Bio::Root::RootI;

use Spangle::Analysis::Match;
use Bio::SeqFeature::Generic;

@ISA = qw(Bio::Root::RootI);

sub new {
  my($pkg,@args) = @_;

  my $self = bless {}, $pkg;

  my ($length1,$length2,$size,$start1,$start2,$end1,$end2,$col) 
      = $self->_rearrange([qw(LENGTH1
			      LENGTH2
			      SIZE
			      START1
			      START2
			      END1
			      END2
			      COL)],@args);



  $length1 || $self->throw("Length1 not input. Length1 = length of sequence 1");
  $length2 || $self->throw("Length2 not input. Length2 = length of sequence 2");

  $start1  = 1        unless defined($start1);
  $start2  = 1        unless defined($start2);
  $end1    = $length1 unless defined($end1);
  $end2    = $length2 unless defined($end2);
  $size    = 500      unless $size;
  
  $self->length1($length1);
  $self->length2($length2);
  $self->size   ($size);
  $self->start1 ($start1);
  $self->start2 ($start2);
  $self->end1   ($end1);
  $self->end2   ($end2);
  $self->col    ($col);

  print(STDERR "start1    = $start1\n");
  print(STDERR "start2    = $start2\n");
  print(STDERR "end1      = $end1\n");
  print(STDERR "end2      = $end2\n");
  print(STDERR "length1   = $length1\n");
  print(STDERR "length2   = $length2\n");
  print(STDERR "size      = $size\n");

  # Initialize the variables

  $self->{_matches}   = [];
  $self->{_clones}    = [];

  return $self;
}

    
sub length1 {
  my ($self,$length1) = @_;

  if (defined($length1)) {
    $self->{_length1} = $length1;
  }
  return $self->{_length1};
}

sub length2 {
  my ($self,$length2) = @_;

  if (defined($length2)) {
    $self->{_length2} = $length2;
  }
  return $self->{_length2};
}

sub start1 {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_start1} = $arg;
  }
  return $self->{_start1};
}
    
sub start2 {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_start2} = $arg;
  }
  return $self->{_start2};
}


sub end1 {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_end1} = $arg;
  }
  return $self->{_end1};
}
        
sub end2 {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_end2} = $arg;
  }
  return $self->{_end2};
}
        
sub size {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_size} = $arg;
  }
  return $self->{_size};
}
        
sub clones {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_clones} = $arg;
  }
  return @{$self->{_clones}};
}
        
sub col {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_col} = $arg;
  }
  return $self->{_col};
}


sub print_matches {
  my ($self,$fh) = @_;


  foreach my $match (@$self->get_all_matches) {
      print($fh $match->start1 . "\t" . $match->start2 . "\t" . $match->length . "\t" . $match->score . "\n");
  }

}
    
         
sub count {
  my ($self) = @_;

  my @matches = $self->get_all_matches;

  return scalar(@matches);

}

sub read_clone_coords {
    my ($self,$clonefile) = @_;

  return unless defined($clonefile);

  open(IN,"<$clonefile") || die "Couldn't open clone coord file $clonefile";

  while (<IN>) {
      chomp($_);

      my ($id,$start,$end)  = split(' ',$_);
      
      my $clone = new Bio::SeqFeature::Generic;

      defined($id)    || $self->throw("Clone id not defined - line input was [$_]");
      defined($start) || $self->throw("Clone start not defined - line input was [$_]");
      defined($end)   || $self->throw("Clone end not defined - line input was [$_]");

      $clone->source_tag($id);
      $clone->start($start);
      $clone->end  ($end);
      
      $self->add_clone($clone);

  }
  close(IN);

}

sub add_clone {
    my ($self,$arg) = @_;


    if (!(defined($self->{_clones}))) {
	$self->{_clones} = [];
    }

    if (defined($arg)) {
	if ($arg->isa("Bio::SeqFeatureI")) {
	    push(@{$self->{_clones}},$arg);
	} else {
	    $self->throw("[$arg] is not a Bio::SeqFeatureI");
	}
    } else {
	$self->throw("No input for DotPlot->add_clone");
    }
}

sub get_all_clones {
    my ($self) = @_;

    if (!(defined($self->{_clones}))) {
	$self->{_clones} = [];
    }

    return @{$self->{_clones}};
}

sub draw_box_and_scale {
    my ($self,$im,$width,$height) = @_;
    
    my $xoffset   = $width * 0.2;
    my $yoffset   = $width * 0.2;

    
    my $start  = int($self->start1);  
    my $end    = int($self->end1);
    
    my $start1 = $self->start1;
    my $start2 = $self->start2;
    my $end1   = $self->end1;
    my $end2   = $self->end2;
    
    my $newim = new GD::Image($width + 2 * $xoffset, $height + 2 * $yoffset);
    
    my $white  = $newim->colorAllocate(255,255,255);
    my $black  = $newim->colorAllocate(0,0,0);
    my $red    = $newim->colorAllocate(255,0,0);
    
    $newim->rectangle($xoffset,$yoffset,$width+$xoffset+1,$height+$yoffset+1,$black);
    
    my $tmp    = int(log($end1-$start1)/log(10));
    my $tick   = 10**($tmp-1);

    my $ticklen   = 10;
    my $tickcount = 0;
    my $i         = $start;

    while ($i <= $end) {

	$newim->line($xoffset + ($i-$start1)*$width/($end1-$start1),$yoffset-$ticklen,
		     $xoffset + ($i-$start1)*$width/($end1-$start1),$yoffset,$black);
    
	my $val = int($i/1000);

	if ($tickcount % 5 == 0) {
	    $newim->stringUp(gdSmallFont,$xoffset +  ($i-$start1)*$width/($end1-$start1) - 5,
			                 $yoffset - 4*$ticklen,
			                 $val,
			                 $black);

      $newim->line($xoffset + ($i-$start1)*$width/($end1-$start1),$yoffset-2* $ticklen,
		   $xoffset + ($i-$start1)*$width/($end1-$start1),$yoffset,$black);

      $tickcount = 0;
    }

    $tickcount++;
    $i += $tick;
  }

  $start = int($start2);
  $end =   int($end2);

  $i         = $start;
  $tickcount = 0;

    while ($i <= $end) {
	my $val = int($i/1000);
	$newim->line($xoffset-$ticklen, 
		     $yoffset + ($i-$start2)*$height/($end2-$start2),
		     $xoffset    ,
		     $yoffset + ($i-$start2)*$height/($end2-$start2),$black);
    
    

	if ($tickcount % 5 == 0) {
	    $newim->string(gdSmallFont,$xoffset - 5*$ticklen, $yoffset + ($i-$start2)*$height/($end2-$start2) - 5,$val,$black);
	    
	    $newim->line($xoffset-2*$ticklen, 
			 $yoffset + ($i-$start2)*$height/($end2-$start2),
			 $xoffset    ,
			 $yoffset + ($i-$start2)*$height/($end2-$start2),$black);

	    $tickcount = 0;
	}
	$tickcount++;
	
	$i += $tick;
    }
    
    my @xclones;
    my @yclones;
    
    my @clones = $self->clones;
    
    if (defined($self->clones)) {
	foreach my $cl (@clones) {
	    if (!($cl->end < $start1 || $cl->start > $end1)) {
		push(@xclones,$cl);
	    }
	    if (!($cl->end < $start2 || $cl->start > $end2)) {
		push(@yclones,$cl);
	    }
	}
	
	my $y   = 5;
	my $dir = 1;
	
	if ($#xclones <= 25) {
	    foreach my $cl (@xclones) {
		my $clstart = $cl->start;
		my $clend   = $cl->end;
	
		if ($clstart < $start1) {$clstart = $start1;}
		if ($clend   > $end1) {$clend   = $end1;}
	
		my $xstart = $xoffset + ($clstart-$start1)*$width/($end1-$start1);
		my $xend   = $xoffset + ($clend  - $start1)*$width/($end1-$start1);
	
		$newim->rectangle($xstart,$y,
				  $xend  ,$y+12,$black);
	
		my $id = $cl->source_tag;
		$newim->string(gdSmallFont, ($xstart + 3) , $y,$id,$black);
	
		$y   = $y + $dir * 12;
		$dir = $dir * -1;
	    }
	}  
	my $x   = 15;
	$dir = 1;
    

	if ($#yclones <= 25) {
	    foreach my $cl (@yclones) {
		my $clstart = $cl->start;
		my $clend   = $cl->end;
		
		if ($clstart < $start2) {$clstart = $start2;}
		if ($clend   > $end2) {$clend   = $end2;}
		
		my $ystart = $yoffset + ($clstart-$start2)*$height/($end2-$start2);
		my $yend   = $yoffset + ($clend  - $start2)*$height/($end2-$start2);
		
		$newim->rectangle($x,   $ystart,
				  $x+12,$yend,$black);
		
		my $id = $cl->source_tag;
		$newim->stringUp(gdSmallFont,  $x,$yend - 3,$id,$black);
		
		$x = $x + $dir * 12;
		$dir = $dir * -1;
	    }
	    
	}
    }
    $newim->copy($im,$xoffset+1,$yoffset+1,0,0,$width,$height);
  
  return $newim;
}

sub dotplot  {
  my ($self,$fh) = @_;

  $fh   = \*STDOUT unless $fh;

  my $size   = $self->size;

  my $start1 = $self->start1;
  my $start2 = $self->start2;
  
  my $end1   = $self->end1;
  my $end2   = $self->end2;

  my $len1   = ($end1 -  $start1 + 1 );
  my $len2   = ($end2 -  $start2 + 1 );
  
  if ($len2 > $len1) {
      $size = $size* $len1/$len2;
  }

  my $height = $size;
  my $width  = $size * $len2/$len1;

  print(STDERR "Gif width/height = $width/$height\n");

  my $im     = new GD::Image($width,$height);

  my $white  = $im->colorAllocate(255,255,255);
  my $black  = $im->colorAllocate(0,0,0);       
  my $red    = $im->colorAllocate(255,0,0);
  my $green  = $im->colorAllocate(0,255,0);
  my $blue   = $im->colorAllocate(0,0,255);
  my $orange = $im->colorAllocate(255,128,0);
  my $yellow = $im->colorAllocate(255,255,0);

  if (defined($self->col)) {
    $im->filledRectangle(0,0,$width,$height,$black);
  }

  my $count = 0;

 POINT: foreach my $match ($self->get_all_matches) {

     my $s1coord = $match->start1;
     my $s2coord = $match->start2;
      
     $count++;
     
     next POINT if ($s1coord < $start1 || $s2coord < $start2 || $s1coord > $end1 || $s2coord > $end2);

     my $s1     = ($s1coord - $start1)* $height/$len1;
     my $s2     = ($s2coord - $start2)* $width /$len2;
      
     my $len   = $match->length;
     my $score = $match->score;
     
     if ($self->col == 2) {
	 $score = $match->score/100;
     }
     
     my $col = $black;
     
     if ($self->col) {
	 if ($score > 4) {
	     $col = $red;
	 } elsif ($score > 3) {
	     $col = $orange;
	 } elsif ($score > 2) {
	     $col = $yellow;
	 } elsif ($score > 1) {
	     $col = $green;
	 } else {
	     $col = $blue;
	 }
     }
     if ($len > 0 ) {
	 
	 $im->line($s1,$s2,$s1+$len*$height/$len1,$s2+$len*$width/$len2,$col);
#	 $im->filledRectangle($s1-2,$s2-2,$s1+$len*$height/$len1+1,$s2+$len*$width/$len2+1,$col);
	 $im->line($s1+1,$s2,$s1+$len*$height/$len1 + 1,$s2+$len*$width/$len2,$col);
	 
     } else {
	 
	 #  $im->line($width - $s1,$s2,$width - $s1+$len*$height/$len1,$s2-$len*$width/$len2,$black);
	 
	 $im->line($s1,$s2,$s1+$len*$height/$len1,$s2-$len*$width/$len2,$col);
	 #$im->filledRectangle($s1-1,$s2-1,$s1+$len*$height/$len1+1,$s2-$len*$width/$len2+1,$col);
     }
     $count++;
 }
  
  my $newim = $self->draw_box_and_scale($im,$width,$height,$black);
  
  print $fh $newim->gif;
}

sub read_matches {
  my ($self,$matchfile) = @_;

  my $line;

  open(IN,"<$matchfile") || die "Couldn't open match file [$matchfile]";
									 
  while ($line = <IN>) {
      chomp($line);

      my ($start1,$start2,$length,$score) = split(' ',$line);

      my $match = new Spangle::Analysis::Match(-start1 => $start1,
					       -start2 => $start2,
					       -length => $length,
					       -score  => $score);

      $self->add_match($match);

  }
  close(IN);
}

sub read_MSPcrunch {
  my ($self,$mspfile) = @_;

  my $line;

  open(IN,"<$mspfile") || die "Couldn't open match file [$mspfile]";
									 
  while ($line = <IN>) {
      chomp($line);
      if ($line !~ /^\#/) {
	  my ($score,$pid,$qstart,$qend,$id1,$hstart,$hend,$id,$rest) = split(' ',$line);
	  
	  my $match;
	  
	  my $qlength = abs($qend-$qstart) + 1;
	  my $hlength = abs($hend-$hstart) + 1;
	  
	  if ($qlength != $hlength) {
	      $self->throw("Query and hit lengths don't match $qlength != $hlength");
	  }
	

	  # Reverse strand hit
	  if ($qstart > $qend) {
	      $match = new Spangle::Analysis::Match(-start1 => $qend,
						    -start2 => $hstart,
						    -length => -1 * $qlength,
						    -score  => $score);
	      
	  } else {
	      # Forward strand hit
	      $match = new Spangle::Analysis::Match(-start1 => $qstart,
						    -start2 => $hstart,
						    -length => $qlength,
						    -score  => $score);
	      
	  }
	  
	  $self->add_match($match);
      }
  }
  close(IN);
}


sub add_match {
  my ($self,$match) = @_;

  if (defined($match)) {
      if ($match->isa("Spangle::Analysis::Match")){ 
	  push(@{$self->{_matches}},$match);
      } else {
	  $self->throw("[$match] is not a Spangle::Analysis::Match");
      }
  } else {
      $self->throw("No input object for add_match");
  }
}


sub get_all_matches {
    my ($self) = @_;


    if (!(defined($self->{_matches}))) {
	$self->{_matches} = [];
    }

    return (@{$self->{_matches}});
}
1;





