package Spangle::Clone;

=pod

  =head1 NAME
  
  Spangle::Clone - ACEDB Clone object

  =head1 SYNOPSIS
  
  $cl = new Spangle::Clone
  $cl = new Spangle::Clone($id,$db);  # Make a new clone with id and acedb handle
  
=head1 DESCRIPTION

A module to store and manipulate genes and
database hits attached to an ACEDB clone.

=head1 CONTACT

Describe contact details here

=head1 APPENDIX

The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use strict;  
use vars   qw(@ISA);

use Bio::Tools::Blast;
use Bio::Seq;

use Spangle::Subseq;
use Spangle::Homol;
use Spangle::HomolSet;
use Spangle::Exon;

use Bio::Seq;

# Object preamble - inherits from Bio::SeqFeature::Generic

use Bio::SeqFeature::Generic;

@ISA = qw(Bio::SeqFeature::Generic);

sub new {
  my($pkg, @args) = @_;

  my $self = $pkg->SUPER::new(@args);

  # Initialize the variables
  $self->{_subseq}   = [];
  $self->{_new_span} = [];

  $self->primary_tag('clone');
  $self->source_tag ('Spangle');

  if (@args == 2) {
    $self->_make(@args);
  }

  return $self; # success - we hope!
}

sub _make {
  my ($self,$id,$db) = @_;

  $id || $self->throw("No clone id entered\n");
  
  $self->id   ($id);
  $self->acedb($db);
#  $self->fetchDNA();
}

sub id {
    my ($self,$arg) = @_;

    if (defined($arg)) {
	$self->{_id} = $arg;
    } else {
	return $self->{_id};
    }
}

sub acedb {
    my ($self,$arg) = @_;
    if (defined($arg)) {
	$self->{_acedb} = $arg;
    } else {
	return $self->{_acedb};
    }
}

sub fetch_tag {
  my ($self,$tag) = @_;

  unless (defined($self->acedb) && defined($self->id)) {
    print("Database handle or id (". $self->id . ") not defined\n");
    return;
  }

  if (!defined($self->{_seq})) {
    $self->{_seq} = $self->acedb->fetch('Sequence',$self->id);
    $self->throw("Can't find " . $self->id . " in ACE database") unless defined($self->{_seq});
  }

  my $outstr = $self->{_seq}->at($tag);

  undef($self->{_seq});
  return $outstr;
}

=head2 eachSubsequence

  Title   : eachSubsequence
  Usage   : @sub = $clone->eachSubsequence
  Function: Returns all the subsequences associated with this clone
  Returns : @Spangle::Subsequence
  Args    : none

=cut

sub eachSubsequence {
    my $self = shift;
    
    if (my $ss = $self->{_subseq}) {
        return @$ss;
    } else {
        return;
    }

    # Loop over the subseqfeatures and fetch out subsequences
    # Need to then convert coords to the current coordinate system

}
 
=head2 eachPredictedGene

  Title   : eachPredictedGene
  Usage   : @sub = $clone->eachPredictedGene
  Function: Returns all the predicted genes associated with this clone
  Returns : @Spangle::Subsequence
  Args    : none

=cut

sub eachPredictedGene {
  my $self = shift;

  my @arr;
  foreach my $subseq ($self->eachSubsequence) {
      if ($subseq->primary_tag eq "Predicted") {
	  push(@arr,$subseq);
      }
  }
  return @arr;
}



=head2 eachProposedGene

  Title   : eachProposedGene
  Usage   : @sub = $clone->eachProposedGene
  Function: Returns all the proposed genes associated with this clone
  Returns : @Spangle::Subsequence
  Args    : none

=cut

sub eachProposedGene {
  my $self = shift;

  my @arr;
  foreach my $subseq ($self->eachSubsequence) {
     if ($subseq->primary_tag eq "Proposed") {
       push(@arr,$subseq);
     }
  }
  return @arr;
}
=head2 eachProposedGene

  Title   : eachProposedGene
  Usage   : @sub = $clone->eachProposedGene
  Function: Returns all the proposed genes associated with this clone
  Returns : @Spangle::Subsequence
  Args    : none

=cut

sub eachGDGene {
  my $self = shift;

  my @arr;
  foreach my $subseq ($self->eachSubsequence) {
      if ($subseq->primary_tag eq "GD") {
       push(@arr,$subseq);
     }
  }
  return @arr;
}

sub stagger {
  my $self   = shift;
  my $offset = shift;

  $offset = 50000 unless $offset;

  my $start = 1;
  my $prevend = 0;
  
  while ($start < $self->end) {
    # Make a new feature
    my $end = $start + 2 * $offset;
    
    if ($end > $self->end) {$end = $self->end;}
    
    if ($end > $prevend) {
      my $f = new Spangle::Clone;
      
      $f->id        ($self->id);
      $f->source_tag("STAGGER");
      $f->start     ($start);
      $f->end       ($end);
      
      $self->add_sub_SeqFeature($f);
      

    }
    $start = $start + $offset;
    $prevend = $end;
  }
}

sub each_stagger {
  my $self = shift;

  my @staggers;

  foreach my $f ($self->sub_SeqFeature) {
    if ($f->source_tag eq "STAGGER") {
       push(@staggers,$f);
    }
  }

  return @staggers;
}


=head2 eachHomologySpan

  Title   : eachHomologySpan
  Usage   : @sub = $clone->eachHomologySpan
  Function: Returns all the homology spans associated with this clone
  Returns : @Spangle::Subsequence
  Args    : none

=cut

sub eachHomologySpan {
  my $self = shift;

  my @arr;
  foreach my $subseq ($self->eachSubsequence) {
     if ($subseq->source_tag =~  /SPAN/) {
       push(@arr,$subseq);
     }
  }
  return @arr;
}


=head2 eachOtherGene

  Title   : eachOtherGene
  Usage   : @sub = $clone->eachOtherGene
  Function: Returns all the unidentified genes 
  Returns : @Spangle::Subsequence
  Args    : none

=cut

sub eachOtherGene {
  my $self = shift;

  my @arr;
  foreach my $subseq ($self->eachSubsequence) {
     if ($subseq->primary_tag ne "SPAN"      &&
         $subseq->primary_tag ne "Proposed"  &&
         $subseq->primary_tag ne "GD"  &&
         $subseq->primary_tag ne "Predicted") {
       push(@arr,$subseq);
     }
  }
  return @arr;
}

=head2 homolset

  Title   : homolset
  Usage   : $clone->homolset
  Function: Gets the set of homologies associated with this clone
  Returns : nothing
  Args    : none

=cut

sub homolset {
    my ($self) = shift;

    if (!defined($self->{_homolset})) {

      $self->{_homolset} = Spangle::HomolSet->new;
      
      my @tmp = $self->_fetch_homols;

      foreach my $h (@tmp) {
	$self->{_homolset}->add_homol($h);
      }

    }

    return $self->{_homolset};
}

=head sequenced_by

  Title   : sequenced_by
  Usage   : my $lab = $clone->sequenced_by
  Function: Fetches the sequenced by tag
  Returns : string
  Args    : none

=cut

sub sequenced_by {
    my $self = shift;

    if (!defined($self->{_sequenced_by})) {
	    my $seq = $self->acedb->fetch('Sequence',$self->id);
        $self->throw("Can't find " . $self->id . " in ACE database") 
                                            unless defined($seq);
	    my $tag = $seq->at('Origin.Sequenced_by[1]');
	    $self->{_sequenced_by} = $tag;
    }

    return $self->{_sequenced_by};
}

=head2 _fetch_homols

  Title   : _fetch_homols
  Usage   : $clone->_fetch_homols
  Function: Fetches homols from ace database and creates a homolset
  Returns : nothing
  Args    : none

=cut

sub _fetch_homols {
    my $self = shift;

    my $db    = $self->acedb;
    my $clone = $self->id;
    
    defined($db)   || $self->throw("No acedb handle. Can't fetch homols\n");
    defined($clone)|| $self->throw("No clone id. Can't fetch homols\n");
    
    my @homolset;

    # This won't work for links - should only be used for individual clones

    my $data;

    $data = $db->raw_query("find sequence $clone");
    $data = $db->raw_query("show -a Homol");


    my @lines = split("\n",$data);
    my @homols;

    foreach my $h (@lines) {
      if ($h !~ /^\/\//     && 
          $h ne ""          &&
          $h !~ /^Sequence/ &&
	  $h !~ /^Motif_homol/) {
	  push(@homols,$h);
      }
    }

    my @newhomols;
    my @method = (
		  'BLASTX',
		  #'est_hmmfs',
		  'EST_blastn',
		  #	'sanger_sts_hmmfs',
		  'EST_eg',
		  #	'sanger_sts_eg',
		  'vertebrate_mRNA_eg',
		  'vertebrate_mRNA_blastn',
		 );

  HOMOL:  foreach my $homol (@homols) {

      my ($type,$id,$meth,$score,$qstart,$qend,$hstart,$hend)  = split(' ',$homol);
      next HOMOL unless defined($hend);

#      $id =~ s/^.*?:(.*)/$1/;
      
      $id =~ s/\"//g;
      $id =~ tr/a-z/A-Z/;
      $meth =~ s/\"//g;

      my @args;
      my $strand = 1;

      if ($hend < $hstart || $qend < $qstart) {
	my $tmp = $hend;
	$hend   = $hstart;
	$hstart = $tmp;
	$strand = -1;
      }	
	
	my $gf = Spangle::Homol->new(-start       => $qstart,
				     -end         => $qend,
				     -strand      => $strand,
				     -score       => $score,
				     '-id'        => $id,
				     -hstart      => $hstart,
				     -hend        => $hend);
	$gf->primary_tag($type);
	$gf->source_tag ($meth);

	if (@method) {
	    foreach my $m (@method) {
		if ($meth =~ $m) { 
		  push(@homolset,$gf);
		  last;
		}
	    }
	} else {
	  push(@homolset,$gf);
	}
#      print("Homol: " . $gf->id . " " . $gf->start . " " . $gf->end . "\n");
    }
    return @homolset;
}


=head2 _fetch_homols_gif

  Title   : _fetch_homols_gif
  Usage   : $clone->_fetch_homols_gif
  Function: Fetches homols from ace database and creates a homolset
  Returns : nothing
  Args    : none

=cut

sub _fetch_homols_gif {
    my $self = shift;

    my $db    = $self->acedb;
    my $clone = $self->id;
    
    defined($db)   || $self->throw("No acedb handle. Can't fetch homols\n");
    defined($clone)|| $self->throw("No clone id. Can't fetch homols\n");
    
    my @homolset;

    # I've changed this to use the gif server as it transforms
    # all the coordinates for you.

    # If the clone is in a link and is reversed then giface won't dump GFF. Argh!
    # We can dump from the sequence both ways round - one way will work. HACK HACK!

    my $data;


    my $start = $self->start;
    my $end   = $self->end;

    if (defined($start) && defined($end)) {
	print "Start/end $start $end\n";
      $data  =  $db->raw_query("gif seqget $clone -coords $start $end;seqfeatures ");
      $data .=  $db->raw_query("gif seqget $clone -coords $end $start;seqfeatures ");
    } elsif (defined($self->dna_seq)) {
      $start = 1;
      $end   = $self->dna_seq->length;
      print "dna Start/end $start $end\n";
      $data  =  $db->raw_query("gif seqget $clone -coords $start $end;   seqfeatures");
      $data .=  $db->raw_query("gif seqget $clone -coords $end   $start; seqfeatures");
    } else {
	print ("no coords\n");
      $data =  $db->raw_query("gif seqget $clone;   seqfeatures");
    }


    my @lines = split("\n",$data);
    my @homols;

    foreach my $h (@lines) {
      if ($h !~ /^\/\//     && 
          $h ne ""          && 
          $h !~ /^Sequence/ && 
          $h =~ /similarity/) {
	push(@homols,$h);
      }
    }

    my @newhomols;
    my @method = (
		  'BLASTX',
		  #'est_hmmfs',
		  'EST_blastn',
		  #	'sanger_sts_hmmfs',
		  'EST_eg',
		  #	'sanger_sts_eg',
		  'vertebrate_mRNA_eg',
		  'vertebrate_mRNA_blastn',
		 );

  HOMOL:  foreach my $homol (@homols) {

      my ($link,$meth,$type,$qstart,$qend,$score,$strand,$pog2,$id,$hstart,$hend)  = split(' ',$homol);
      next HOMOL unless defined($hend);

#      $id =~ s/^.*?:(.*)/$1/;
      
      $id =~ s/\"//g;
      $id =~ tr/a-z/A-Z/;
      $meth =~ s/\"//g;

      my @args;

      if ($strand eq ".") {
	$strand = 1;
      }

      if ($hend < $hstart) {
	my $tmp = $hend;
	$hend   = $hstart;
	$hstart = $tmp;
	$strand = -1;
      }	
	
	my $gf = Spangle::Homol->new(-start       => $qstart,
				     -end         => $qend,
				     -strand      => $strand,
				     -score       => $score,
				     '-id'        => $id,
				     -hstart      => $hstart,
				     -hend        => $hend);
	$gf->primary_tag($type);
	$gf->source_tag ($meth);

	if (@method) {
	    foreach my $m (@method) {
		if ($meth =~ $m) { 
		  push(@homolset,$gf);
		  last;
		}
	    }
	} else {
	  push(@homolset,$gf);
	}
#      print("Homol: " . $gf->id . " " . $gf->start . " " . $gf->end . "\n");
    }
    return @homolset;
}


=head2 fetch_subsequences

  Title   : fetch_subsequences
  Usage   : $clone->fetch_subsequences
  Function: Accesses the ace database and extracts subsequence information for this clone
  Returns : nothing
  Args    : none

=cut

sub fetch_subsequences {
  my ($self, $subseqlist) = @_;
  
  my $db    = $self->acedb;
  my $clone = $self->id;
  
  defined ($db)    || $self->throw("No acedb handle defined. Can't fetch subsequences");
  defined ($clone) || $self->throw("No clone id defined. Can't fetch subsequences");
  
  # See if we are in a link
  
  my $cl = $db->fetch('Sequence',$clone);
 # print("Fetched AcePerl clone object\n");
#  print("Checking whether clone is in link...");
  if (!defined($cl)) {$self->throw("No clone id $clone found in database\n");}

  my ($inlink) = $cl->at('Structure.From.Source');
  my ($islink) = $cl->at('Properties');
  
  $self->{_inlink} = $inlink;
  $self->{_islink} = $islink;
  
  #print("done\n");

 # print("Finding subsequences ");
  my ($subseq,$subsequence_gff) = $self->_find_subsequences($db,$clone,$subseqlist);
  print("done\n");

  #  print "Got subsequnces\n", map("$_\n", sort keys %$subseq);

  if ($islink eq "Link") {
    print("Found link object - following links\n");

    # Make clone objects from the link subsequences
    # and fetch_subsequences on those

    foreach my $sub (keys %$subseq) {
      eval {
	my $cl = Spangle::Clone->new;
	
	$cl->id      ($sub);
	$cl->acedb   ($db);
	#      $cl->fetchDNA();
	$cl->fetch_subsequences();
	
	$cl->start($subseq->{$sub}{start});
	$cl->end  ($subseq->{$sub}{end});
	
	$self->add_sub_SeqFeature($cl,'EXPAND');
	
	foreach my $s ($cl->eachSubsequence) {
	  my $start = $subseq->{$sub}{start};
	  my $end   = $subseq->{$sub}{end};
	  
	  $cl->_subcoords2real($s,$start,$end);
	  
	  $self->add_subsequence($s);
	}
      };

      warn("Problem extracting clone $sub. Skipping\n") if ($@);

    }
  } else {
    $self->_process_subsequences($db,$clone,$subseq,$subsequence_gff);
  }

}



sub _subcoords2real {
  my ($self,$subseq,$start,$end) = @_;
  my @gfs = $subseq->each_Exon;
  my @introns;
  
  #first calculate the global coordinate of the CDS_start and _end
  
  my $exonarraylength =  @gfs;
  

  if ($exonarraylength  > 1){   #make array of length of introns
    for (my $i=0;$i+1 < $exonarraylength ;$i++){
        $introns[$i] = $gfs[$i+1]->start - $gfs[$i]->end -1;
    }
    my $intronacc = 0 ;
    
    for (my $i=0;$i < $exonarraylength  ;$i++){

        my $exonstartlocal = $gfs[$i]->start - $intronacc; #convert exon coordinates to loacal
        my $exonendlocal   = $gfs[$i]->end   - $intronacc;
        if ($start < $end){
            if ($exonstartlocal <= $subseq->CDS_start && $subseq->CDS_start <= $exonendlocal){
                $subseq->CDS_start($subseq->CDS_start + $start + $intronacc -1)
            }
            if ($exonstartlocal <= $subseq->CDS_end && $subseq->CDS_end <= $exonendlocal) {
                $subseq->CDS_end($subseq->CDS_end + $start + $intronacc -1)
            }    
        }else{
            if ($exonstartlocal <= $subseq->CDS_start && $subseq->CDS_start <= $exonendlocal){
                 print "start $start, end $end";
                 $subseq->CDS_start($start - $subseq->CDS_start -  $intronacc + 1)
            }
            if ($exonstartlocal <= $subseq->CDS_end && $subseq->CDS_end <= $exonendlocal) {
                $subseq->CDS_end($start - $subseq->CDS_end -  $intronacc + 1)
            }
        }
     
        $intronacc = $intronacc + $introns[$i];
    }
  }elsif ($start < $end){
    $subseq->CDS_start($subseq->CDS_start + $start - 1)
  }else{
    $subseq->CDS_start($start - $subseq->CDS_start + 1)
  }
  
  
  #print("Sub seq is " . $subseq->primary_tag . "\t" . $subseq->source_tag . "\n");

  # This is wrong - should be strand dependent
  my $strand = $gfs[0]->strand;

  my $max = 0;
  my $min = 100000000;

  foreach my $ex (@gfs) {
    if ($end > $start) {
      # Forward gene
      $ex->start($ex->start + $start - 1);
      $ex->end  ($ex->end   + $start - 1);
      
    } elsif ($end < $start) {
      # reverse gene
      $ex->start($start - $ex->start + 1);
      $ex->end  ($start - $ex->end   + 1);
    }
    
    # Now change the coords around and set the strand
    
    if ($ex->start > $ex->end) {
      my $tmp = $ex->start;
      
      $ex->start ($ex->end);
      $ex->end   ($tmp);
      $ex->strand(-1);
      
    } else {
      $ex->strand(1);
    }

    if ($ex->start < $min) {$min = $ex->start;}
    if ($ex->end   > $max) {$max = $ex->end;}

  }
  $subseq->start($min);
  $subseq->end  ($max);

}
sub test_gene {
  my ($self,@ex) = @_;

  if ($#ex == 1) {
    my $valid = 0;
    
    my @ev0 = $ex[0]->each_supporting_feature;
    
    my $topscore0 = 0;
    my $coverage0 = $#ev0 + 1;
    
    foreach my $e (@ev0) {
      if ($e->source_tag eq "BLASTX") {
	if ($e->score_per_base*100/5 > $topscore0) {$topscore0 = $e->score_per_base*100/5;}
      } else {
	if ($e->score_per_base*100 > $topscore0) {$topscore0 = $e->score_per_base*100;}
      }
    }
    
    my @ev1 = $ex[1]->each_supporting_feature;
    
    my $topscore1 = 0;
    my $coverage1 = $#ev1 + 1;
    
    foreach my $e (@ev1) {
      if ($e->source_tag eq "BLASTX") {
	if ($e->score_per_base*100/5 > $topscore1) {$topscore1 = $e->score_per_base*100/5;}
      } else {
	
	if ($e->score_per_base*100 > $topscore1) {$topscore1 = $e->score_per_base*100;}
      }
    }
    print("Top scores are : " . $topscore0 . " " . $topscore1 . "\n");
    print("Coverage is : "    . $coverage0 . " " . $coverage1 . "\n");
    
    if ((($topscore0 > 97 && $coverage0 > 1) || ($topscore0 >= 99 && $coverage0 >0)) &&
	(($topscore1 > 97 && $coverage1 > 1) || ($topscore1 >= 99 && $coverage1 >0)))  {
      $valid = 1;
    }
    
    return $valid;
  } else {
    return 1;
  }
}

sub find_new_genes {
  my ($self,$fh,$comparetype) = @_;

  my @span = $self->each_new_span;
  $comparetype = "proposed" unless $comparetype;

  print("comparetype is $comparetype\n");

  if ($#span < 0) {
    @span = $self->eachHomologySpan;
  }

  my @new;
  my @new22;
  my @extended;
  my @merged;
  my @otherspan;
  
  return unless $#span >= 0;
  
  my %overlap;
  my %fstart;
  my %fend;

  my %GDoverlap;
  my %GDfstart;
  my %GDfend;

  my @realgenes;
  my @GDgenes;
  
  push(@GDgenes,  $self->eachGDGene);
  push(@realgenes,$self->eachProposedGene);
  push(@realgenes,$self->eachOtherGene);
  
SPAN:  foreach my $s (@span) {
    my @ex    = $s->each_Exon;
    my $valid = $self->test_gene(@ex);
      
    if ($valid != 1) {
      print("Invalid 2 exon gene\n");
      push(@otherspan,$s);
      $s->primary_tag('otherspan');
      next SPAN;
    } else {
      print("Valid  gene\n");    
      
      foreach my $p (@realgenes) {
	print "Comparing to " . $p->source_tag  . "\n";
	
	my @pex = $p->each_Exon;
	
	my $pstart = $pex[0]    ->start;
	my $pend   = $pex[$#pex]->end;
	
	
	foreach my $ex (@ex) {
	  foreach my $pex (@pex) {
	    my @overlap = Spangle::Homol::match($ex,$pex);
	    
	    if ($overlap[0]) {
	      $overlap{$s}{$p} = $overlap{$s}{$p} + 1;
	      if ($ex == $ex[0]) {
		$fstart{$s} = 1;
	      }
	      if ($ex == $ex[$#ex]) {
		$fend{$s} = 1;
	      }
	    }
	  }
	}
      }
      
      my $count = 0;
      my $noex = -1;
      
      foreach my $p (keys %{$overlap{$s}}) {
	$count++;
	if ($overlap{$s}{$p} ne "") {
	  if ($overlap{$s}{$p} > $noex) {
	    $noex = $overlap{$s}{$p};
	  }
	}
      }
      print ("Count is $count\n");

      foreach my $p (@GDgenes) {
	print "Comparing to " . $p->source_tag  . "\n";
	
	my @pex = $p->each_Exon;
	
	my $pstart = $pex[0]    ->start;
	my $pend   = $pex[$#pex]->end;
	
	
	foreach my $ex (@ex) {
	  foreach my $pex (@pex) {
	    my @overlap = Spangle::Homol::match($ex,$pex);
	    
	    if ($overlap[0]) {
	      $GDoverlap{$s}{$p} = $GDoverlap{$s}{$p} + 1;
	      if ($ex == $ex[0]) {
		$GDfstart{$s} = 1;
	      }
	      if ($ex == $ex[$#ex]) {
		$GDfend{$s} = 1;
	      }
	    }
	  }
	}
      }
      
      my $GDcount = 0;
      my $GDnoex = -1;
      
      foreach my $p (keys %{$GDoverlap{$s}}) {
	$GDcount++;
	if ($GDoverlap{$s}{$p} ne "") {
	  if ($GDoverlap{$s}{$p} > $GDnoex) {
	    $GDnoex = $GDoverlap{$s}{$p};
	  }
	}
      }
      
      print("GDcount is $GDcount\n");

      if ($GDcount == 0) {
	print("New C22 gene\n");
	push(@new22,$s);
      }
      
      if ($count == 0) {
	print("New gene\n");
	push(@new,$s);
	$s->primary_tag('new');
	
	
      } elsif ($count <= 2 && (!$fstart{$s} || !$fend{$s})) {
	print("Extended gene\n");
	push(@extended,$s);
	$s->primary_tag('extended');
	
      } elsif ($count > 2) {
	print("Possible merged genes\n");
	push(@merged,$s);
	$s->primary_tag('merged');
      }
      
      print("Number of genes overlapped $count\n");
      print("Max no of exons = $noex " . ($#ex+1) . "\n\n");
      
      
    }
  }
  
  my $str = "New      ";
  
  foreach my $n (@new) {
    $str .= " " . $n->source_tag;
  }
  
  $str .= "\n";
  
  $str .= "New 22 ";
  foreach my $n (@new22) {
    $str .= " " . $n->source_tag;
  }
  $str .= "\n";
  
  $str .= "Extended ";
  foreach my $n (@extended) {
    $str .= " " . $n->source_tag;
  }
  $str .= "\n";
  
  $str .= "Merged   ";
  foreach my $n (@merged) {
    $str .= " " . $n->source_tag;
  }
  $str .= "\n";
  
  $str .= "Otherspan";
  foreach my $n (@otherspan) {
    $str .= " " . $n->source_tag;
  }
  $str .= "\n";
  
  $str .= "Proposed ";
  foreach my $n ($self->eachProposedGene) {
    $str .= " " . $n->source_tag;
  }
  $str .= "\n";

  $str .= "Other ";
  foreach my $n ($self->eachOtherGene) {
    $str .= " " . $n->source_tag;
  }
  $str .= "\n";

  $str .= "Predicted ";
  foreach my $n ($self->eachPredictedGene) {
    $str .= " " . $n->source_tag;
  }
  $str .= "\n";

  
  print($fh $str);
  return $str;
}

sub _find_subsequences {
  my ($self,$db,$clone,$subseqlist) = @_;
  my  $subsequence_gff;
  
  my %wanted_subseq = map {$_, 1} @$subseqlist
    if $subseqlist;
  
  # Find the subsequence coords in the clone entry
  # These are used to change the internal subsequence exon
  # coordinates into clone coordinates
  
  my $cl = $db->fetch("Sequence", $clone);
  unless ($cl) {
    warn "Can't fetch Sequence '$clone'";
    return;
  }
  
  my $subseq = {};
  foreach my $seq ($cl->at('Structure.Subsequence[1]')) {
    my ($name, $start, $end) = map $_->name, $seq->row();
    
    if ($subseqlist) {
        next unless $wanted_subseq{$name};
    }
    #warn "Found '$name' $start -> $end\n";
    
    my $feature = Bio::SeqFeature::Generic->new();    
    $feature->primary_tag($name);
    $feature->start($start);
    $feature->end($end);
    
    $subsequence_gff->{$name} = $feature;
    
    $subseq->{$name}{start} = $start;
    $subseq->{$name}{end}   = $end;

    
    if ($start > $end) {
      $subseq->{$name}{strand} = -1;
    } else {
      $subseq->{$name}{strand} = 1;
    }
  }
  return ($subseq,$subsequence_gff);
}

sub _process_subsequences {
  my ($self,$db,$clone,$subseq,$subsequence_gff) = @_;

  my @names = keys(%$subseq);
  my $subseq2 = {};
  
  foreach my $s (@names) {
    $subseq2->{$s} = $db->fetch('Sequence',$s);
  }
#  print("Processing subsequences\n");

SUBSEQ:  foreach my $s (keys %$subseq2) {
#    print("Subsequence $s\n");

    my $s2          = $subseq2->{$s};
    
    my @start;
    my @end;

    if (!$s2) {
      $self->warn("No subsequence object defined for $s\n");
      next SUBSEQ;
    }

    
 #   print("Fetching exons\n");
    # Fetch the exons
    if ($s2->at('Structure.From.Source_exons')) {
      @start = $s2->at('Structure.From.Source_exons')->col;
      @end   = $s2->at('Structure.From.Source_exons')->col(2);
    } else {
      push(@start,1);
      push(@end,($subseq->{$s}{end}-$subseq->{$s}{start}+1));
    }   
    
  #  print("fetching other properties\n");

    # Fetch the other properties of the subsequence
    my $method          = $s2->fetch('Method');
    my $property        = $s2->at('Properties.Coding.CDS_predicted_by');
    my $locus           = $s2->at('Visible.Locus[1]');
    my $start_not_found = $s2->at('Properties.Start_not_found');
    my $end_not_found   = $s2->at('Properties.End_not_found');
    my $pseudogene =  $s2->at('Properties.Pseudogene');
    my $CDS_start   =  $s2->at('Properties.Coding.CDS[1]');
    my $CDS_end   =  $s2->at('Properties.Coding.CDS[2]');
#    my $CDS_start;
#    my $CDS_end;
#    my($cds);
#    if ($cds =  $s2->at('Properties.Coding.CDS')) {
#        ($CDS_start, $CDS_end) = map $_->name, $cds->row(1);
 #   }
    
    my @remark;

    if ($s2->at('Visible.Remark')) {
      @remark   = $s2->at('Visible.Remark')->col();
    }
    
    my $frame = 0;
    if ($s2->at('Properties.Start_not_found')) {
	$frame = $s2->at('Properties.Start_not_found')->right();
	if ($frame eq "") {
	    $frame = 0;
	} else {
	    $frame = $frame - 1;
	}
    }
    
    $frame = $frame % 3;

    my $type;
    
    if ($method eq "supported_CDS" || $method eq  "supported_mRNA") {
	$type = "Proposed";
    } elsif ($method eq "GD_CDS" || $method eq "GD_mRNA" || $method eq "GD_in_progress") {
	$type = "GD";
	print("Type is GD\n");
    } elsif ($property) {
      $type = "Predicted";
    } else {
      $type = "Other";
      $method = "SEQ";
    }

    my $tmpstart = $subsequence_gff->{$s}->start;
    my $tmpend   = $subsequence_gff->{$s}->end;

   # print("Creating subsequence object\n");

    # Create the main subsequence object
    my $subsequence = new Spangle::Subseq;
    
    $subsequence->strand      ($subseq->{$s}{strand});
    $subsequence->primary_tag ($type);
    $subsequence->source_tag  ($s);
    $subsequence->method      ($method);

    $subsequence->CDS_start   ($CDS_start);
    $subsequence->CDS_end   ($CDS_end);
    
    $subsequence->locus       ($locus);

    $subsequence->contig_dna  ($self->dna_seq);

    foreach my $r (@remark) {
      $subsequence->addRemark ($r);
    }

    $subsequence->frame($frame+1);
    $subsequence->start_not_found($start_not_found);
    $subsequence->end_not_found  ($end_not_found);
    $subsequence->pseudogene  ($pseudogene);
    
    if ($type eq "Predicted") {
	$subsequence->printFlag(0);
    }

    # Now create the exon objects and add to the subsequence object

    #print("Creating exon objects\n");

    for (my $i = 0; $i <= $#start; $i++) {
      
      my $ex = new Spangle::Exon;
      
      $ex->primary_tag($s);
      $ex->source_tag ($method);
      $ex->start      ($start[$i]);
      $ex->end        ($end[$i]);
      $ex->strand     ($subseq->{$s}{strand});
      $ex->phase      ($frame);
      
      $subsequence->add_Exon($ex);
      
    }
    
    # Now change all the internal frame coordinates in the subsequence objects into
    # clone coordinates
    
    $self->_subcoords2real($subsequence,$tmpstart,$tmpend);
    
    # Set the DNA sequence
    if (defined($self->dna_seq)) {
	#print(STDERR "Setting DNA\n");
	foreach my $ex ($subsequence->each_Exon) {
	    
	    if ($ex->end <= $self->dna_seq->length() && $ex->start > 0) {
		my $newseq = $self->dna_seq->trunc($ex->start,$ex->end);
		
		if ($ex->strand == -1) {
		    $newseq = $newseq->revcom;
                }
		} else {
		    print "Exon from subsequence " . $subsequence->source_tag . " longer than clone length or start < 0. DNA not added for exon\n";
		    next SUBSEQ;
		}
	}
    }
    #print(STDERR "Adding subsequence to clone\n");
    $self->add_subsequence($subsequence);
    #print("done\n");
  }
  
}

=head2 add_subsequence

  Title   : add_subsequence
  Usage   : $clone->add_subsequence($sub)
  Function: Adds a subsequence into the clone
  Returns : nothing
  Args    : none

=cut

sub add_subsequence {
    my ($self,$subseq) = @_;

    if (defined($subseq)) {
     # print("Adding subsequence [" . $subseq->source_tag . "  ". $subseq->primary_tag ."] " . $self->source_tag . "\n");

	push(@{$self->{_subseq}},$subseq);
    } 
}

=head2 remove_subsequence

  Title   : remove_subsequence
  Usage   : $clone->remove_subsequence($sub)
  Function: Removes a subsequence from clone object
  Returns : nothing
  Args    : none

=cut

sub remove_subsequence {
    my ($self,$subseq) = @_;

    my @new;

    if (defined($subseq)) {
       foreach my $sub ($self->eachSubsequence) {
         if ($sub != $subseq) {
            push(@new,$sub);
         }
       }
       $self->{_subseq} = [];
       push(@{$self->{_subseq}},@new);
    } 
}

=head2 add_new_span

  Title   : add_new_span
  Usage   : $clone->add_new_span($sub)
  Function: Adds a new span into the clone
  Returns : nothing
  Args    : Spangle::Subseq

=cut

sub add_new_span {
    my ($self,$subseq) = @_;

    push(@{$self->{_new_span}},$subseq);
}

=head2 each_new_span

  Title   : each_new_span
  Usage   : my @spans = $clone->each_new_span
  Function: returns an array of generated spans
  Returns : @Spangle::Subseq
  Args    : none

=cut

sub each_new_span {
    my ($self) = @_;

    return @{$self->{_new_span}};

}

=head2 reduceSubsequences

  Title   : reduceSubsequences
  Usage   : $clone->reduceSubsequences
  Function: Makes subsequences non redundant
  Returns : nothing
  Args    : none

=cut

=head2 chromosome

  Title   : chromosome
  Usage   : $clone->chromosome
  Function: Gets/ets chromosome
  Returns : string
  Args    : none

=cut


sub chromosome {
 my ($self,$arg) = @_;

 if (defined($arg)) {
   $self->{_chromosome} = $arg;
 } elsif (!defined($self->{_chromosome})) {
   $self->{_chromosome} = $self->fetch_tag('Origin.Chromosome[1]');
 }

 return $self->{_chromosome};

}

=head2 fetchDNA

  Title   : fetchDNA
  Usage   : $dnaseq  = Clone::fetchDNA
  Function: fetches dna from acedb
  Returns : Bio::Seq
  Args    : none

=cut


sub fetchDNA {
    my $self = shift;

    my $id  = $self->id
        or $self->throw("No clone id. Can't fetch DNA");
    my $dbh = $self->acedb
        or $self->throw("No acedb handle. Can't fetch DNA");

    my $seq = $dbh->fetch(Sequence => $id)
        or $self->throw("Can't find Sequence '$id'");
    my $dna = $seq->at('DNA')->fetch->at
        or $self->throw("Can't get DNA from Sequence '$id'");

    my $seq = new Bio::Seq(-SEQ => $dna);
    $self->dna_seq($seq);
}


=head2 dna_seq

  Title   : dna_seq
  Usage   : $dnaseq  = Clone::dna_seq
  Function: Returns the dna sequence of the clone
  Returns : Bio::Seq
  Args    : none

=cut

sub dna_seq {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_dnaseq} = $arg;
  }
  return $self->{_dnaseq};
}

sub lab {
  my ($self) = @_;

  if (!defined($self->{_lab})) {
      my $lab = $self->fetch_tag('Origin.Sequenced_by[1]');
      $self->{_lab} = $lab;
  }
  
  return $self->{_lab};
}

sub fetch_trims {
  my ($clone) = @_;

  my $trim_left  = $clone->fetch_tag('Structure.Trim_left[1]');
  my $trim_right = $clone->fetch_tag('Structure.Trim_right[1]');

  $clone->{_trim_left}  = $trim_left;
  $clone->{_trim_right} = $trim_right;

  return ($trim_left,$trim_right);
}

sub analysis_Directory {
    my ($clone) = @_;

    my $dir = $clone->fetch_tag('Analysis_details.Analysis_directory[1]');

    $dir =~ s|\~humpub|/nfs/disk100/humpub|;
    
    return $dir;
}

sub analysis_Files {
    my ($clone) = @_;

    if (!defined($clone->{_seq})) {
	$clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	$clone->throw("Can't find " . $clone->id . " in ACE database") unless defined($clone->{_seq});
    }

    my @files = $clone->{_seq}->at('Analysis_details.Analysis_files[1]');
    return @files;
}

sub fetch_DNA {
  my ($clone,$mask) = @_;
  my $dna;
  

  if (defined($clone->{_dnaseq})) {
      return $clone->{_dnaseq};
  }
  
  print("Fetching masked DNA for " . $clone->id . "\n") if $mask;

  my $dir = $clone->fetch_tag('Analysis_details.Analysis_directory[1]');
  
  $dir =~ s|\~humpub|/nfs/disk100/humpub|;
  
  print("Analysis directory is $dir\n");
  
  if ($dir eq "" ) {
    print("ERROR: No analysis directory defined\n");
    return;
  }

  if (! -e $dir) {
    $dir =~ s/\/.*$//;
    print("Trying in $dir\n");
    if (-e $dir) {
      opendir(DIR,$dir);
      my @files = readdir DIR;
      print("files in $dir are : @files\n");
      
      if ($#files == 0) {
	$dir = $dir . "/$files[0]";
      } else {
	print("ERROR: Can't find analysis directory\n");
	return;
      }
    } else {
      print("ERROR: Can't find analysis directory\n");
      return
    }

  }
      

  my $id = $clone->id;
  if ($id =~ /[Ee][Mm]/) {
    $id =~ tr/a-z/A-Z/;
  }
  $id =~ s/[Ee][Mm]://;
  

  my $dnafh;
  
  if ($mask) {
    
    my $seqfile = $dir . "/" . $id . ".seq.p";
    print("Trying to find masked file $seqfile\n");

    if (-e $seqfile) {
	print("Masked sequence file is $seqfile\n");
      open(DNA,"<$seqfile");
      $dnafh = \*DNA;
    }

  } else {
    
    my $seqfile1 = $dir . "/rawdata/" . $id . ".ace.gz";
    my $seqfile2 = $dir . "/rawdata/" . $id . ".ace";
    my $seqfile3 = $dir . "/" . $id . ".seq";
    
    if (-e $seqfile1) {
      open(DNA,"gunzip -c $seqfile1 |");
      print("Sequence file is $seqfile1\n");
      $dnafh = \*DNA;
      
    } elsif (-e $seqfile2) {
      open(DNA,"< $seqfile2");
      print("Sequence file is $seqfile2\n");

      $dnafh = \*DNA;
    } elsif (-e $seqfile3) {
      open(DNA,"<$seqfile3");
      print("Sequence file is $seqfile3\n");
      
    $dnafh = \*DNA;
    }
  }
  if (!defined($dnafh)) {
    print ("Couldn't fetch DNA for " . $clone->id . "\n");
    return;
  }

  my @data = <$dnafh>;
    
  my $line;

  while ($line = shift @data) {
    if ($line =~ /^DNA/ || $line =~ /^>/) {
      
      my $dnaline;
      
      while (($dnaline = shift @data) && ($dnaline !~ /^\n$/)) {

	$dnaline =~ s/\n//;
	$dna .= $dnaline;
      }
    }
  }

  close($dnafh);
  close(DNA);

  $clone->dna_seq(new Bio::Seq(-seq =>$dna));

  return $dna;

}

sub trim_left {
    my ($self,$arg) = @_;

    if (defined($arg)) {
	$self->{_trim_left} = $arg;
    }

    return $self->{_trim_left};
}

sub trim_right {
    my ($self,$arg) = @_;
    
    if (defined($arg)) {
	$self->{_trim_right} = $arg;
    }
    
    return $self->{_trim_right};
}

sub trimdna {
    my ($self,$arg) = @_;
    
    if (defined($arg)) {
	$self->{_trimdna} = $arg;
    }
    
    return $self->{_trimdna};
}

sub trimace {
    my ($self,$arg) = @_;
    
    if (defined($arg)) {
	$self->{_trimace} = $arg;
    }
    
    return $self->{_trimace};
}


sub trim {
    my ($self) = @_;

    my $lab = $self->lab;
    
    print(STDERR "Lab is $lab\n");
    
    my ($trimdna,$trimace)                  = $self->_get_trimmed_sequence;
    
    
    $self->trimdna($trimdna);
    $self->trimace($trimace);
}


sub _get_trimmed_sequence {
    my ($self) = @_;
    
    my $acestr;
    my $dnastr;
    
    my $dna = $self->fetch_DNA;
    
    my ($substart,$subend) = $self->find_subseq;
    
    print(STDERR "DNA start - end : $substart $subend\n");
    my $newdna = substr($dna,$substart-1,($subend - $substart + 1));
	
    # This is the 'proper' dna - revcomped if necessary
    $dnastr = $newdna;
    
    if ($self->strand == -1) {
	print(STDERR "reversing\n");
	$dnastr =~ tr/atcgATCG/tagcTAGC/;
	$dnastr = reverse($dnastr);
    }
    
    # This is for the ace file
    $newdna =~ s/(.{72})/$1\n/g;

    if ($self->lab ne "SC" && $self->lab ne "Sanger Centre") {
	my $trimname = $self->id . ".TRIMMED";
    
	$acestr .= "\n-D sequence \"$trimname\"\n";
	
	$acestr .= "\nSequence : \"" . $self->id . "\"\n";
	$acestr .= "-D subsequence $trimname\n";
	$acestr .= "-D DNA\n";
	
	$acestr .= "\nSequence : \"" . $self->id . "\"\n";
	$acestr .= "Subsequence "    . $trimname . " $substart $subend \n";
	
	$acestr .= "\nDNA : \""      . $trimname . "\"\n";
	$acestr .= "$newdna\n";
    }
    
    $acestr .= "\nSequence : \"" . $self->id . "\"\n";
    $acestr .= "Trim_left "      . $self->trim_left . "\n";
    $acestr .= "Trim_right "     . $self->trim_right . "\n";
    
    return ($dnastr,$acestr);

}

sub find_subseq {
    my ($clone) = @_;
    
    my $strand = $clone->strand;
    my $len    = abs($clone->end - $clone->start) + 1;
    
    my $substart;
    my $subend;

    my $new_trim_left  = $clone->trim_left;
    my $new_trim_right = $clone->trim_right;

    # do we trim from the left or the right?
    
    if ($strand == 1) {
	$substart = $new_trim_left;
	$subend   = $len;
	
    } else {
	
	my $tmp = $new_trim_left;
	$new_trim_left  = $len - $new_trim_right + 1;
	$new_trim_right = $len - $tmp + 1;   
	
	$substart = 1;
	$subend   = $new_trim_right;
    }

    return ($substart,$subend);
}
 
#########################################################################
# functions for Clone.pm
# Added by MAQ 4.4.00
# Used by Checker.pm
#########################################################################

#returns list of database entries
sub database {			
    my ($clone) = @_;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }
    # get list of database entries
    my @database  = $clone->{_seq}->at('DB_info.Database');
    return @database;
}    

#count the rows of under a database    
sub db_count_rows {
    my ($clone, $db_entry) =@_;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }
    #count the number of entries in the column ie. the number of rows
    my $rows  = $db_entry->col();
    return $rows;    
}   

#counts the number of entries under a database 
sub db_count_cols {
    my ($clone, $db_entry) =@_;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }
    #count the number of entries in the row ie. the number of columns
    my $columns  = $db_entry->row(1);
    return $columns;    
}   

#returns projectname
sub project_name {
    my ($clone) = @_;

    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }
    #get the project name
    my $project_name = $clone->{_seq}->at('Project.Project_name[1]');
    return $project_name;
}

#returns embl accession number    
sub embl_accession {
    my ($clone) = @_;
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }    
    #retrieve embl accession number (hop right two positions)
    my $embl = $clone->{_seq}->at('DB_info.Database.EMBL[2]');
    return $embl;
}       

#returns embl id number    
sub embl_id {
    my ($clone) = @_;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }    
    #retrieve embl accession number (hop right two positions)
    my $embl = $clone->{_seq}->at('DB_info.Database.EMBL[1]');
    return $embl;
}

sub suffix {
    my ($clone) = @_;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }    
    #projec suffix
    my $suffix = $clone->{_seq}->at('Project.Project_suffix[1]');
    return $suffix;
}

sub seqlength {
    my ($clone) = @_;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }    
    #retrieve length (hop right 2 from DNA tag)
    my $length = $clone->{_seq}->at('DNA[2]');
    return $length;
}

#NB to be replaced by fetch_subsequence call
sub subsequence_list {
    my ($clone) = @_;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }    
    #retrieve subsequences
    my @sub_list = $clone->{_seq}->at('Structure.Subsequence');
    return @sub_list;
}

#NB to be replaced by fetch_subsequence call
sub subseq_clone_coordinates {
    my ($clone, $subseq) = @_;
    my @coordinates = undef;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }
    if ($subseq)
    {
      #my $seq = $clone->{_seq}.'Structure.Subsequence'.$subseq;
      @coordinates = $subseq->row(1);   
    }
    return @coordinates; 
}

#NB to be replaced by fetch_subsequence call
sub subseq_internal_coordinates {
    my ($clone, $subseq) = @_;
    my @coordinates;
    my $tag;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }    
    if ($subseq)
    {
        my $offset; # used for row command
        my $seq = $clone->acedb->fetch('Sequence', $subseq);
        $clone->throw("Can't find " . $subseq . " in ACE database") 
                                        unless defined($seq);
        if ($seq->at('DNA')) #sequence is genomic or trimmed genomic
            {
                $tag = 'DNA[2]';
                $offset = 0;
                push (@coordinates, 1); #ensures correct length is calculated                            
            }   
        elsif ($seq->at('Properties.Coding.CDS') ||
                $seq->at('Properties.Transcript.mRNA'))    #sequence is link or mRNA
            {
                $tag = 'Structure.From.Source_exons[1]';
                $offset = 0;
            }
        elsif ($seq->at('Structure.From.Source_exons'))    #BUG fix: some CDSs don't return true to 'Properties.Coding.CDS'
            {
                $tag = 'Structure.From.Source_exons[1]';
                $offset = 0;
            }
            else                                         #sequence is embl genomic
            {
                $tag = 'Structure.Subsequence[1]';
                $offset = 1;
            }
        
        foreach my $exon ($seq->at($tag))
        {
            push (@coordinates, $exon->row($offset));
        }        
    }
    return @coordinates; 
}

sub link {
    my ($clone) = @_;
    
    if (!defined($clone->{_seq})) 
    {
	    $clone->{_seq} = $clone->acedb->fetch('Sequence',$clone->id);
	    $clone->throw("Can't find " . $clone->id . " in ACE database") 
                                        unless defined($clone->{_seq});
    }
    
    return "true" if ($clone->{_seq}->at('Properties.Link'));
    return undef;
}
