
### Bio::EnsEMBL::Ace::Filter::Das

### 
package Bio::EnsEMBL::Ace::Filter::Das;

use strict;
use Carp;
use Bio::SeqIO;
use Bio::EnsEMBL::Ace::Filter;
use Bio::EnsEMBL::ExternalData::DAS::DASAdaptor;
use Bio::EnsEMBL::ExternalData::DAS::DAS;

our @ISA = ('Bio::EnsEMBL::Ace::Filter');

my $verbose = 0;
my $DEBUG   = 0;

sub new {
    my ($class) = @_;

    my $self = bless {}, ref($class) || $class;

    return $self;
}
sub das_dsn{
    my ($self, $dsn) = @_;
    $self->{'_das_dsn'} = $dsn if $dsn;
    warn $self->{'_das_dsn'} if $verbose;
    return $self->{'_das_dsn'};
}
sub das_url{
    my ($self, $url) = @_;
    $self->{'_das_url'} = $url if $url;
    warn $self->{'_das_url'} if $verbose;
    return $self->{'_das_url'};    
}
sub das_proxy_url{
    my ($self, $proxy_url) = @_;
    $self->{'_das_proxy_url'} = $proxy_url if $proxy_url;
    warn $self->{'_das_proxy_url'} if $verbose;
    return $self->{'_das_proxy_url'};    
}
sub das_types{
    my ($self, $types) = @_;
    push(@{$self->{'_das_types'}}, split(",",$types)) if $types;
    warn $self->{'_das_types'} if $verbose;
    return $self->{'_das_types'};    
}
sub das_protocol{
    my ($self, $protocol) = @_;
    $self->{'_das_protocol'} = $protocol if $protocol;
    warn $self->{'_das_protocol'} if $verbose;
    return $self->{'_das_protocol'};
}

sub make_DASAdaptor{
    my ($self) = @_;
    unless($self->{'_das_adaptor'}){
        warn "Using DAS @ " . $self->das_url() . " " . $self->das_dsn() . "\n" if $DEBUG;
        # make External DAS Adaptor
        my $das_adapt = Bio::EnsEMBL::ExternalData::DAS::DASAdaptor->new(
                                                                         -url       => $self->das_url(),
                                                                         -protocol  => $self->das_protocol(),
                                                                         -dsn       => $self->das_dsn(),
                                                                         -types     => $self->das_types(),
                                                                         -proxy_url => $self->das_proxy_url(),
                                                                         );
        $das_adapt->_db_handle->debug($DEBUG);
        print STDERR "Have DAS Adaptor Obj $das_adapt \n" if $DEBUG;
        # make DAS obj
        my $ext_das = Bio::EnsEMBL::ExternalData::DAS::DAS->new($das_adapt);
        $self->{'_das_adaptor'} = $ext_das;
    }
    return $self->{'_das_adaptor'};
}

1;

__END__

=head1 NAME - Bio::EnsEMBL::Ace::Filter::Das

=head1 DESCRIPTION

Allows the creation of ACeDB .ace file contents 
from a EnsEMBL pipeline database and a DAS source.
 This is intended to be used as a base class,
please implement ace_data method in your subclass.

=head1 SETUP (.otter_config requirements)

Under the [B<species>.use_filters] stanza the 
name of the analysis is required.  For the 
DAS sources I imagined a dsn/das types string
would suffice.

under the [B<species>.filter.B<das_source>]
stanza there needs to be the options for the 
DataFactory to create the filter.  The first
required is B<module>.  This should be set to 
the extending module using it's full path as I 
saw Bio::EnsEMBL::Ace::Filter::Das::*.pm as 
being logical to do this. The formatting in 
ace_data is likely to be different for 
different das types for example.

A B<das_types>, B<das_dsn>, B<das_usrl>,
B<das_proxy_url> and B<das_protocol> are
all setable in this stanza.

B<create_without_ana> needs to be set to 1.
This is only needed because all the other filters 
need an analysis obj.  It was the least change 
of code, but other suggestions would probably
work just as well if not better.  This is kind 
of another thing to remember when setting up stuff.

-------- .otter_config START --------

 [B<species>.use_filters]
 das_source=1

 # the options for the Das source
 [B<species>.filter.B<das_source>]
 module=Bio::EnsEMBL::Ace::Filter::Das::DasExample
 # a dsn for the das source
 das_dsn=vega30
 # a url for the das source
 das_url=http://das.ensembl.org/das
 # a proxy, which caused me grief! 403 Forbidden was hidden pretty well!!!!!
 das_proxy_url=http://proxy.t.l.d:8080
 create_without_ana=1
 das_types=prediction
 
--------------END------------------

N.B.

I'm not sure if the Defaults module will allow lists :(
but we could try dealing with them in the Filter module
as above.
e.g. sub das_types{ ...; $das_types = [split(",",$_[1])] if $_[1]; ... }

=head1 SEE ALSO

B<Bio::Otter::Lace::Defaults>

=head1 AUTHOR

Roy Storey B<email> rds@sanger.ac.uk
