### Bio::EnsEMBL::Ace::Filter::Gene

package Bio::EnsEMBL::Ace::Filter::Gene;

use strict;
use Carp;
use Bio::SeqIO;
use Data::Dumper;
use Bio::EnsEMBL::Ace::Filter;
use vars '@ISA';
@ISA = ('Bio::EnsEMBL::Ace::Filter');

=pod

=head1 new

 makes a new object.

=cut

sub new {
    my ($class) = @_;

    my $self = bless {}, $class;

    return $self;
}

=head1 ace_data

 required by those above!

=cut

sub ace_data {    
    my ( $self, $slice) = @_;
    my $transcripts = $self->get_transcripts( $slice );   
    return $self->ace_data_from_transcripts( $slice, $transcripts );
}


=head1 Gene specific methods

 required by ace_data

=cut

sub get_transcripts {
    my ( $self, $slice ) = @_;

    if($slice){
        my $transcript_list;
        my $genes = $slice->get_all_Genes( ); 

        foreach my $gene (@$genes){
            my $gene_name = $gene->stable_id;
            my $tscripts  = $gene->get_all_Transcripts; 
            foreach my $t (@$tscripts ){
                $t->temporary_id($gene_name);
                push(@$transcript_list, $t);
            }
        }    
        return $transcript_list;
    }else{
        confess "No slice given";
    }
}

sub ace_data_from_transcripts {
    my ( $self, $slice, $transcripts ) = @_;

    my $contig_id   = $slice->name ;    
    my $genomic_ace = '';
    my $subseq_ace  = '';
    
    foreach my $transcript ( @$transcripts ) {
        my( $gen, $subseq );
        eval{
            ($gen, $subseq) = $self->_make_ace_data($slice, $transcript);
        };
        if ($@) {
            warn $@;
        } else {
            $genomic_ace .= $gen;
            $subseq_ace  .= $subseq;
        }
    }
    my $url_ace = '';
   
    if (my $url_str = $self->url_string) {    
        my $slice_name = $slice->name;
        my ($chr , $start_pos , $end_pos) = $slice_name =~ /(\d+)\.(\d+)\-(\d+)/ ;
    
        my %genes = map {$_,1} grep $_, map $_->temporary_id , @$transcripts;
        my @genes = %genes;
        
        foreach my $gene_name (keys %genes) {
            my $url   = $self->weblocation($url_str ,$gene_name,$chr,$start_pos , $end_pos );
            $url_ace .= qq`\nURL : "$gene_name"\nURL "$url"\n`;
        }
    }
    if ($genomic_ace) {
        return qq`\nSequence : "$contig_id"\n$genomic_ace\n$subseq_ace\n$url_ace\n`;
    }
    else {
        return '';   
    }
}

sub weblocation{
    my ( $self, $url_str, @list ) = @_; 
    return sprintf($url_str, @list);
}

sub url_string {
    my( $self, $url_string ) = @_;   
    if ($url_string){
        $self->{'_url_string'} = $url_string;
    }
    return $self->{'_url_string'} ;
}

sub next_gene_int {
    my( $self ) = @_;
    
    return ++$self->{'_gene_int'};
}

sub make_transcript_ace_name {
    my( $self, $transcript ) = @_;
    
    if (my $name = $transcript->stable_id) {
        if ($name =~ /^ENS/) {
            my($pre, $post) = $name =~ /(\D{1,4})0*(\d+)/;
            if ($pre and $post) {
                $name = $pre . $post;
            }
        }
        return $name;
    } else {
        $name = $self->analysis_object->logic_name . '.' . $self->next_gene_int;
        return $name;
    }
}

sub _make_ace_data {
    my ($self , $slice , $transcript ) = @_;
    # some variables
    my $transcript_id ;
    my $first_exon_start  ;
    my $last_exon_end ;
    my $slice_name;
    my $subseq_header; # to store initial details of  subsequence;
    my $subseq_data;  # to store the output to be printed for each subsequence 
    my ($cds_start , $cds_end); # to store start and end of coding region relative to the slice
    my $ace_phase; # stores the phase of the exon (###+++# , ##+++## , #+++###)
    my $meth_name; # stores the method used to find the gene (genscan , pfam, etc)
    
    
    $transcript_id = $self->make_transcript_ace_name($transcript);
    
    $slice_name = $slice->name;
    # get list of exons and check if start / end are found
    my ($start_not_found, $end_not_found, @exons_in_transcript) = $self->transcript_exons_in_slice($slice, $transcript); 
    unless (@exons_in_transcript) {
        die "Transcript $transcript_id has no exons in contig\n"
    }
    ## add number to end for overlapping genes?
    if ($start_not_found or $end_not_found) {
        my $i = $self->next_gene_int;
        $transcript_id .= '.' . ++$self->{'_partial_int'}{$transcript_id};
    }
    
    $subseq_data = qq{\nSequence : "$transcript_id"\nSource "$slice_name"\n}; 
    
     
    ($cds_start, $cds_end , $ace_phase) = $self->CDS_start_end( $transcript  ,@exons_in_transcript); 

    if (defined($cds_start) && defined($cds_end)){
        $subseq_data .= qq{CDS $cds_start $cds_end\n};        
        if ($start_not_found) {
            if (defined $ace_phase) {
                $subseq_data .= qq{Start_not_found $ace_phase\n};
            } else {
                $subseq_data .= qq{Start_not_found\n};
            }       
        }
    }
    else{
        if ($start_not_found) {
             $subseq_data .= qq{Start_not_found\n};
        }
    }
    
    

    if ($end_not_found) {
        $subseq_data .= qq{End_not_found\n};
    }
      
    # get name of method used to predict the gene and append to the output data variable
    $meth_name = $self->analysis_object->logic_name;
    $subseq_data .= qq{CDS_predicted_by "$meth_name"\nMethod "$meth_name"\n};
    
    if ($self->url_string) {
        my $gene_name = $transcript->temporary_id;
        $subseq_data .= qq{Web_location "$gene_name"\n};
    }
    if ($transcript->can('display_xref') and my $xr = $transcript->display_xref) {
        if (my $desc = $xr->description) {
	    $desc =~ s!\"!\\"!g; # remove the "
            $subseq_data .= qq{EMBL_dump_info DE_line "$desc"\n};
        }
    }

    # get start and end of coding region
    $first_exon_start = $self->first_exon_start( @exons_in_transcript );
    $last_exon_end    = $self->last_exon_end   ( @exons_in_transcript );
   
    # get list of exon co-ordinates in this subsequence and append to the output data
    # if gene is not on strand 1 then start and end co-ordinates should be swaped (as API always has start less than end)
    if ($exons_in_transcript[0]->strand == 1) {
        $subseq_data .= $self->fwd_strand_exon_list( $first_exon_start, @exons_in_transcript );
    } else {
        $subseq_data .= $self->rev_strand_exon_list( $last_exon_end, @exons_in_transcript );
    }

    # if gene is on stand -1 then start and end positions should be  swaped 
    if ($exons_in_transcript[0]->strand == -1) {
        ($first_exon_start, $last_exon_end) = ($last_exon_end, $first_exon_start );
    }
    
    return(qq{Subsequence "$transcript_id" $first_exon_start $last_exon_end\n}, $subseq_data );
}
{
    my %phase = (
       "-1"=> "1",
        "0" => "1",
        "1" => "3",
        "2" => "2",
    );

    sub ace_phase {
        my ( $self, $start_phase ) = @_;
        return $phase{$start_phase};
    }
}

sub first_exon_start {
    my ( $self, @exons_in_transcript ) = @_;

    my $first_exon_start;
    if ( $exons_in_transcript[0]->strand == 1 ) {
        $first_exon_start = $exons_in_transcript[0]->start;
    }
    else {
        $first_exon_start = $exons_in_transcript[ $#exons_in_transcript ]->start; 
    }
    
    return $first_exon_start;
}

sub last_exon_end {
    my ( $self, @exons_in_transcript ) = @_;

    my $last_exon_end;

    if ( $exons_in_transcript[0]->strand == 1 ) {
        $last_exon_end = $exons_in_transcript[ $#exons_in_transcript ]->end;
    }

    else {
        $last_exon_end = $exons_in_transcript[0]->end;
    }
    return $last_exon_end;
}

sub CDS_start_end {
    my ($self  ,$transcript , @exons_in_transcript )   = @_ ;

    my $translation = $transcript->translation ;
    my $start_exon  = $translation->start_Exon;
    my $end_exon    = $translation->end_Exon; 
    
    my ($cds_start , $cds_end ) ;
    my $cds_length = 0 ;
    my $splice_length = 0;
    
    my $coding_exon_present ;
    my $ace_phase;
    my $passed_end ;
       
    #go through exons, adding up the length. Set start / end to appropriate coords, when start / end  exon is encountered
    foreach my $exon  (@exons_in_transcript){
        if ($exon == $start_exon){
            # we have obviously been adding up non coding exons in this case - reset the value of $cds_length 
            $cds_length = 0;
            $cds_start =   $splice_length + $translation->start ;
         }
        if ($exon == $end_exon ){
            $cds_end = $splice_length + $translation->end; 
            $passed_end = 1;
        }
        unless ($passed_end){
            $cds_length += $exon->length;
        }
        $splice_length += $exon->length;
    
        # check to see if the exon is coding and get the phse of the first coding exon
        if ((! $coding_exon_present) && ($exon->phase != -1)){
            $coding_exon_present = 1;
            $ace_phase = $self->ace_phase($exon->phase) ;
        }
    }
    
    
    if ($coding_exon_present){    
        $cds_start ||= 1;
    }
    if ($cds_start){
        $cds_end   ||= $splice_length ;
    }
    
    
    
    return($cds_start, $cds_end , $ace_phase);
}

sub fwd_strand_exon_list {
    my( $self, $pos, @exons_in_transcript ) = @_;
    
    my $exon_list = '';
    foreach my $exon (@exons_in_transcript) {
        my $start = $exon->start - $pos + 1;
        my $end   = $exon->end   - $pos + 1;
        $exon_list .= qq{Source_Exons $start $end\n};
    }
    return $exon_list;

}

sub rev_strand_exon_list {
    my( $self, $pos, @exons_in_transcript ) = @_;
    
    my $exon_list = '';
    foreach my $exon (@exons_in_transcript) {
        my $start = $pos - $exon->end   + 1;
        my $end   = $pos - $exon->start + 1;
        $exon_list .= qq{Source_Exons $start $end\n};
    }
    return $exon_list;
}

sub transcript_exons_in_slice{
    my ($self, $slice, $transcript) = @_;
    
    unless ($slice and $transcript){
        confess "need both slice and transcript";
    }
    
    my $transcript_name = $transcript->stable_id;
   
    my $contig_id = $slice->name;
    my $exons = $transcript->get_all_Exons();
    
    my ($snf, $enf, @exons_in_transcript);
      
    foreach my $exon (@$exons){       
          # reject exons that have a start or end outside this slice
         unless  ($exon->start < 0 || $exon->end > $slice->length   ) {
             my $end = $exon->end;
             if ($exon->start == $end) {
                 warn "Resetting end of exon in '$transcript_name' where start = end = $end\n";
                 $exon->end($end + 3);
             }
             push (@exons_in_transcript, $exon);
         }
         else{
             if (@exons_in_transcript ) {
                 $enf = 1;
             } 
             else{
                 $snf = 1;
             }

        }
    }         
    return ($snf , $enf , @exons_in_transcript)
}


1;

__END__

=head1 NAME - Bio::EnsEMBL::Ace::Filter::Gene

=head1 AUTHOR

Colin Kingswood<email> ck2@sanger.ac.uk

