### Bio::EnsEMBL::Ace::Filter::Similarity

package Bio::EnsEMBL::Ace::Filter::Similarity;

use strict;
use Carp;
use Data::Dumper;
use strict;
use Bio::SeqIO;
use Bio::EnsEMBL::Ace::Filter;
use vars '@ISA';
@ISA = ('Bio::EnsEMBL::Ace::Filter');

sub new {
    my ($class) = @_;

    my $self = bless {}, $class;
    return $self;
}

sub prepare_query {
    confess "prepare_query method not implemented in ", ref(shift);
}

sub align_tag {
    confess "align_tag method not implemented in ", ref(shift);
}

sub is_protein {
    confess "is_protein method not implemented in ", ref(shift);
}

sub homol_tag {
    my ( $self, $homol_tag ) = @_;

    if ($homol_tag) {

        $self->{'_homol_tag'} = $homol_tag;
    }
    return $self->{'_homol_tag'};
}

sub method_tag {
    my ( $self, $method_tag ) = @_;

    if ($method_tag) {
        $self->{'_method_tag'} = $method_tag;
    }
    return $self->{'_method_tag'};

}

sub hseq_prefix {
    my ( $self, $hseq_prefix ) = @_;

    if (defined $hseq_prefix) {
        $self->{'_hseq_prefix'} = $hseq_prefix;
    }
    return $self->{'_hseq_prefix'};
}

sub file_handle {
    my( $self, $fh ) = @_;
    
    if ($fh) {
        $self->{'_fh'} = $fh;
    }
    return $self->{'_fh'};
}

sub drop_file_handle {
    my $self = shift;
    
    $self->{'_fh'} = undef;
}

my %db_prefix = (
    EMBL        => 'Em:',
    Swissprot   => 'Sw:',
    TrEMBL      => 'Tr:',
    );

sub get_db_prefix_and_description {
    my( $self, $hit_name, $dbh ) = @_;

    my $sth = $dbh->prepare_cached(q{
        SELECT hit_description
          , hit_db
        FROM hit_description
        WHERE hit_name = ?
        });
    $sth->execute($hit_name);
    my ($desc, $database) = $sth->fetchrow;
    $sth->finish;
    
    if ($desc) {
        $desc =~ s/"/\\"/g;
    } else {
        $desc = 'No description found';
    }

    return( $database ? $db_prefix{$database} : undef, $desc );
}

sub percent_identity_cutoff {
    my( $self, $percent_identity_cutoff ) = @_;
    
    if (defined $percent_identity_cutoff) {
        $self->{'_percent_identity_cutoff'} = $percent_identity_cutoff;
    }
    return $self->{'_percent_identity_cutoff'};
}

sub max_coverage {
    my( $self, $max_coverage ) = @_;
    
    if (defined $max_coverage) {
        $self->{'_max_coverage'} = $max_coverage;
    }
    $max_coverage = $self->{'_max_coverage'};
    return defined $max_coverage ? $max_coverage : 10;
}

sub filter_hits {
    my( $self, $contig_length, $sth ) = @_;
    
    my $percent_cutoff = $self->percent_identity_cutoff || 0;

    my $current_h_name = '';
    my $max_percent_id = 0;
    my $max_score      = 0;
    my( $current );
    my $percent_hits = {};
    my $count = 0;
    my( $percent_id, $h_name, $score );
    $sth->bind_col(1, \$h_name);
    $sth->bind_col(2, \$score);
    $sth->bind_col(8, \$percent_id);
    while (my $row = $sth->fetch) {
        next if $percent_id < $percent_cutoff;
        $count++;
        unless ($count % 1000) {
            print STDERR "  $count features\n";
        }

        if ($current_h_name ne $h_name) {
            if ($current) {
                unshift(@$current, $current_h_name);
                my $bin = $percent_hits->{$max_percent_id}{$max_score} ||= [];
                push(@$bin, $current);
            }

            $current = [];
            $current_h_name = $h_name;
            $max_percent_id = 0;
            $max_score      = 0;
        }

        $max_percent_id = $percent_id if $percent_id > $max_percent_id;
        $max_score      = $score      if $score      > $max_score;
        push(@$current, [@$row[2..$#$row]]);   
    }

    # Add the last match
    if ($current and @$current) {
        unshift(@$current, $current_h_name);
        my $bin = $percent_hits->{$max_percent_id}{$max_score} ||= [];
        push(@$bin, $current);
    }

    my( $sat_zones );
    if ($self->max_coverage) {
        $sat_zones = $self->_depth_filter_hits($percent_hits, $contig_length);
    } else {
        $sat_zones = [];
    }
    
    return($percent_hits, $sat_zones);
}

sub _depth_filter_hits {
    my( $self, $percent_score_hits, $contig_length ) = @_;

    die "contig length not given" unless $contig_length;
    
    my $max_coverage = $self->max_coverage;
    warn "max coverage=$max_coverage\n";
    
    my $coverage_map = [];

    my $total = 0;
    my $kept  = 0;
    foreach my $pc (sort {$b <=> $a} keys %$percent_score_hits) {
        my $score_hits = $percent_score_hits->{$pc};
        
        foreach my $sc (sort {$b <=> $a} keys %$score_hits) {
            my $hits = $score_hits->{$sc};
            
            for (my $n = 0; $n < @$hits;) {
                $total++;
                unless ($total % 1000) {
                    warn "Filtered: $total\n";
                }
                my $ht = $hits->[$n];
                my( $h_name, @matches ) = @$ht;
                my $keep_hit = 0;
                foreach my $mat (@matches) {
                    # Data at positions 0 and 1 are the start and end of the hit in the contig
		    foreach my $i ( $mat->[0] .. $mat->[1] ) {
		        my $depth = $coverage_map->[$i] ||= 0;
		        if ( $depth < $max_coverage ) {
			    $keep_hit = 1;
                        }
			# Increment depth at this position in the map
			$coverage_map->[$i]++;
		    }
                }
                if ($keep_hit) {
                    $kept++;
                    $n++;
                } else {
                    splice(@$hits, $n, 1);
                }
            }
            
            unless (@$hits) {
                delete($score_hits->{$sc});
            }
        }
    }
    warn "kept $kept out of $total hits\n";
    
    my $sat_zones = [];
    my $zone_start = undef;
    my $zone_score = 0;
    for (my $i = 1; $i <= $contig_length; $i++) {
        my $n = $coverage_map->[$i] || 0;
        if ($zone_start) {
            $zone_score += $n;
            if ($n < $max_coverage) {
                push(@$sat_zones, [$zone_start, $i - 1, $zone_score]);
                $zone_start = undef;
                $zone_score = 0;
            }
        }
        elsif ($n >= $max_coverage) {
            $zone_start = $i;
            $zone_score = $n;
        }
    }
    if ($zone_start) {
        # Are saturated up to end of contig:
        push(@$sat_zones, [$zone_start, $contig_length, $zone_score]);
    }
    printf STDERR "Found %d saturated zones\n", scalar @$sat_zones;
    return $sat_zones;
}

sub saturated_zones_ace_text {
    my( $self, $saturated_zones ) = @_;

    my $max = $self->max_coverage;
    my $homol_method = $self->method_tag;
    my $pat = qq{Feature "Saturated_$homol_method" \%d \%d 1 "avg. depth=\%.0f cutoff=$max"\n};
    my $ace_str = '';
    foreach my $zone (@$saturated_zones) {
        my ($start, $end, $score) = @$zone;
        $ace_str .= sprintf $pat, $start, $end, $score / ($end - $start + 1);
    }
    return $ace_str;
}

## this should not be run to get Data from a Similarity.pm object. 
## instead it has been written so that DnaSimilarity , ProteinSimilarity and RefSeq objects can inherit and use it   
sub ace_data{
    my ($self, $slice) = @_ ;
    
    my $fh           = $self->file_handle     or confess "file_handle not set";   
    my $homol_method = $self->method_tag;
    my $homol_tag    = $self->homol_tag;
    my $align_tag    = $self->align_tag       or confess "align_tag not set";
    my $ana_obj      = $self->analysis_object or confess "analysis_object not set";
    my $is_protein   = $self->is_protein;
    
    my $analysis_id = $ana_obj->dbID;
    my $logic_name  = $ana_obj->logic_name;
    

    my $dbh = $slice->adaptor->db->db_handle;    
    my $tp = $slice->get_tiling_path();
    my $hit_type = $is_protein ? 'Protein' : 'Sequence';
     
    foreach my $tile (@$tp) {
        my $contig_id     = $tile->component_Seq->dbID;
        my $contig_name   = $tile->component_Seq->name;
        
        print STDERR "Fetching '$logic_name' for '$contig_name'\n";
        
        my $gen_ace = qq{\nSequence : "$contig_name"\n};  # Stores data in genomic sequence
        my $hit_ace = ''; # Stores database hits
        my $url_string = $self->url_string;

        my $sth = $self->prepare_query($dbh) ; # varies depending on the inheriting class
    
        $sth->execute($contig_id, $analysis_id);

        my ($filtered_hits, $saturated_zones) = $self->filter_hits($tile->component_Seq->length, $sth);

        while (my ($percent, $score_bin) = each %$filtered_hits) {
            while (my ($score, $bin) = each %$score_bin) {
                foreach my $hit (@$bin) {
                    my $h_name = shift @$hit;

                    # Start ace text for new hit
                    my ($prefix, $desc) = $self->get_db_prefix_and_description($h_name, $dbh);       
                    unless ($prefix){
                        $prefix = $self->hseq_prefix || '';
                        $desc = 'No description found';
                        #warn "Missing description for '$h_name' \n";
                    }
                    my $current_full_name = $prefix . $h_name;   
                    
                    $hit_ace .= qq{\n$hit_type : "$current_full_name"\nTitle "$desc"\n};
                    
                    foreach my $match (@$hit) {
                      
                        my ($start, $end, $h_start, $h_end, $strand, $percent_id, $cigar, $hit_strand) = @$match;
                                              
                        $hit_strand ||= 1;  

                        # est2genome has strand info from splice sites which we are losing
                        # contig_strand is always 1 at the moment
                        # align_coords will break when we fix this
                        $strand = $strand * $hit_strand;

                        #produces an number of align coordintes here
                        my ($seq_coord , $target_coord , $other) = Bio::EnsEMBL::Ace::Filter::Cigar_ace_parser::align_coords(
                            $cigar,
                            $start,
                            $end,
                            $h_start,
                            $strand,
                            $is_protein,
                            );                    
                    
                        if ($strand == -1){
                            ($start, $end) = ($end, $start);
                        }
                        
                        $hit_ace .= qq{Homol DNA_homol "$contig_name" "$homol_method" $percent_id $h_start $h_end $start $end \n};               
                      
                        my $query_line = qq{Homol $homol_tag "$current_full_name" "$homol_method" $percent_id $start $end $h_start $h_end};       
                        #adds the align coords here
                        if (@$seq_coord > 1){                        
                            for (my $i = 0; $i < @$seq_coord; $i++){
                                $gen_ace .=  $query_line . " $align_tag $seq_coord->[$i] $target_coord->[$i] $other->[$i]\n";
                            }
                        } else {
                            $gen_ace .= $query_line . "\n";
                        }
                    }
                        
                    if ($url_string) {
                        my $url_obj_name = "${homol_method}_$h_name";
                        $hit_ace .= qq{Web_location "$url_obj_name"\n}

                            . qq{\nURL "$url_obj_name"\n}
                            . qq{URL "} . sprintf($url_string, $h_name) . qq{"\n};
                    }
                }                    
            }                    
        }
        
        $gen_ace .= $self->saturated_zones_ace_text($saturated_zones);

        # Print to file after each clone to avoid out of memory problems
        if ($hit_ace) {
            print $fh $gen_ace, $hit_ace;
        }
    }
}

sub url_string{
    my ($self, $url) = @_;
    $self->{'_url_string'} = $url if $url;
    return $self->{'_url_string'};
}


1;

__END__

=head1 NAME - Bio::EnsEMBL::Ace::Filter::Similarity

=head1 AUTHOR

Stephen Keenan B<email> keenan@sanger.ac.uk

