
### Bio::Otter::Lace::AceDatabase

package Bio::Otter::Lace::AceDatabase;

use strict;
use Carp;
use File::Path 'rmtree';
use Symbol 'gensym';
use Fcntl qw{ O_WRONLY O_CREAT };
use Ace;
use Bio::Otter::Lace::Defaults;
use Bio::Otter::Lace::PipelineDB;
use Bio::Otter::Lace::SatelliteDB;
use Bio::Otter::Converter;
use Bio::Otter::Lace::PersistentFile;

use Bio::EnsEMBL::Ace::DataFactory;
use Bio::EnsEMBL::Ace::Filter::Gene;
use Bio::EnsEMBL::Ace::Filter::DNA;

my $DATASET_HASH_FILE    = '.slice_dataset';
my $LOCK_REGION_XML_FILE = '.lock_region.xml';

sub new {
    my( $pkg ) = @_;
    
    return bless {}, $pkg;
}
sub Client {
    my( $self, $client ) = @_;
    
    if ($client) {
        $self->{'_Client'} = $client;
    }
    return $self->{'_Client'};
}

sub home {
    my( $self, $home ) = @_;
    
    if ($home) {
        $self->{'_home'} = $home;
    }
    elsif (! $self->{'_home'}) {
	my $readonly_tag = $self->Client->write_access ? '' : $self->readonly_tag();
	# warn "readonly_tag '$readonly_tag'\n";
        $self->{'_home'} = "/var/tmp/lace.${$}${readonly_tag}";
    }
    return $self->{'_home'};
}
sub readonly_tag{
    my ($self) = @_;
    return '.ro';
}

sub title {
    my( $self, $title ) = @_;
    
    if ($title) {
        $self->{'_title'} = $title;
    }
    elsif (! $self->{'_title'}) {
        $self->{'_title'} = "lace.$$";
    }
    return $self->{'_title'};
}

sub tar_file {
    my( $self, $tar_file ) = @_;
    
    if ($tar_file) {
        $self->{'_tar_file'} = $tar_file;
    }
    elsif (! $self->{'_tar_file'}) {
        foreach my $root ($ENV{'OTTER_HOME'}, $ENV{'LACE_LOCAL'}, '/nfs/humace2/hum/data') {
            next unless $root;
            my $file = "$root/lace_acedb.tar";
            if (-e $file) {
                warn "FOUND '$file'\n";
                $self->{'_tar_file'} = $file;
                last;
            }
        }
    }
    return $self->{'_tar_file'};
}

sub tace {
    my( $self, $tace ) = @_;
    
    if ($tace) {
        $self->{'_tace'} = $tace;
    }
    return $self->{'_tace'} || 'tace';
}

sub error_flag {
    my( $self, $error_flag ) = @_;
    
    if (defined $error_flag) {
        $self->{'_error_flag'} = $error_flag;
    }
    return ($self->{'_error_flag'} ? 1 : 0);
}
sub add_acefile {
    my( $self, $ace ) = @_;
    
    my $af = $self->{'_acefile_list'} ||= [];
    push(@$af, $ace);
}

sub list_all_acefiles {
    my( $self ) = @_;
    
    if (my $af = $self->{'_acefile_list'}) {
        return @$af;
    } else {
        return;
    }
}

sub write_otter_acefile {
    my( $self, $ss ) = @_;

    my $dir = $self->home;
    my $otter_ace = "$dir/rawdata/otter.ace";
    my $fh = gensym();
    open $fh, "> $otter_ace" or die "Can't write to '$otter_ace'";
    if ($ss) {
        print $fh $self->fetch_otter_ace_for_SequenceSet($ss);
    } else {
        print $fh $self->fetch_otter_ace;
    }
    close $fh or confess "Error writing to '$otter_ace' : $!";
    $self->add_acefile($otter_ace);
    $self->save_slice_dataset_hash;
}

sub fetch_otter_ace {
    my( $self ) = @_;

    my $client = $self->Client or confess "No otter Client attached";
    
    my $ace = '';
    my $selected_count = 0;
    foreach my $dsObj ($client->get_all_DataSets) {
        my $ss_list = $dsObj->get_all_SequenceSets;
        foreach my $ss (@$ss_list) {
            if (my $ctg_list = $ss->selected_CloneSequences_as_contig_list) {
                $dsObj->selected_SequenceSet($ss);
                $ace .= $self->ace_from_contig_list($ctg_list, $dsObj);
                foreach my $ctg (@$ctg_list) {
                    $selected_count += @$ctg;
                }
            }
        }
    }
    
    if ($selected_count) {
        return $ace;
    } else {
        return;
    }
}

sub fetch_otter_ace_for_SequenceSet {
    my( $self, $ss ) = @_;
    
    my $client = $self->Client
        or confess "No otter client attached";
    my $dsObj = $client->get_DataSet_by_name($ss->dataset_name);
    confess "Can't find DataSet that SequenceSet belongs to"
        unless $dsObj;

    $dsObj->selected_SequenceSet($ss);
    my $ctg_list = $ss->selected_CloneSequences_as_contig_list
        or confess "No CloneSequences selected";
    return $self->ace_from_contig_list($ctg_list, $dsObj);
}

sub ace_from_contig_list {
    my( $self, $ctg_list, $dsObj ) = @_;

    my $client = $self->Client or confess "No otter Client attached";
    my $ace = '';
    my $write_access = $client->write_access();

    foreach my $ctg (@$ctg_list) {
        if($write_access){
            # only lock the region if we have write access.
            my $lock_xml = $client->lock_region_for_contig_from_Dataset($ctg, $dsObj);
            $self->write_lock_xml($lock_xml, $dsObj->name);
        }
        my $xml        = Bio::Otter::Lace::TempFile->new;
        $xml->name('lace.xml');
        my $write      = $xml->write_file_handle;
        my $xml_string = $client->get_xml_for_contig_from_Dataset($ctg, $dsObj);
        print $write $xml_string ;
        # If we're here we now have all the locks!!!

        ### Nasty that genes and slice arguments are in
        ### different order in these two subroutines
        my ($genes, $slice, $sequence, $tiles, $feature_set) =
            Bio::Otter::Converter::XML_to_otter($xml->read_file_handle);
        $ace .= Bio::Otter::Converter::otter_to_ace($slice, $genes, $tiles, $sequence, $feature_set);
        # We need to record which dataset each slice came
        # from so that we know where to save it back to.
        # this gets done in the write_lock_xml so only need to do it here
        # if we haven't got write access.
        $self->save_slice_dataset($slice->display_id, $dsObj->name) unless $write_access;
    }
    return $ace;
}

sub write_lock_xml{
    my ($self, $xml, $ds_name) = @_;

    if($xml && $ds_name){
        my $lock_xml = Bio::Otter::Lace::PersistentFile->new();
        $lock_xml->root($self->home);
        $lock_xml->name($LOCK_REGION_XML_FILE);
        my $write = $lock_xml->write_file_handle();

        print $write $xml;

        my $read = $lock_xml->read_file_handle();
        my ($genes,$slice,$seqstr,$tiles) = Bio::Otter::Converter::XML_to_otter($read);
        my $slice_name = $slice->display_id();
        $self->save_slice_dataset($slice_name, $ds_name);
        $lock_xml->mv(".${slice_name}${ds_name}${LOCK_REGION_XML_FILE}");
    }
}

sub save_slice_dataset {
    my( $self, $slice_name, $ds_name ) = @_;

    if($slice_name && $ds_name){
        print STDERR "Saving $slice_name under $ds_name \n";
        $self->{'_slice_name_dataset'}->{$ds_name} ||= [];
        push(@{$self->{'_slice_name_dataset'}->{$ds_name}}, $slice_name);
    }
}

sub slice_dataset_hash {
    my $self = shift;
    confess "slice_dataset_hash method is read-only" if @_;

    my $h = $self->{'_slice_name_dataset'};
    unless ($h) {
        #warn "Creating empty hash";
        $h = $self->{'_slice_name_dataset'} = {};
    }
    return $h;
}

# Makes hash persistent for "lace -recover"
# (Could store in Dataset_name tag in database?)
sub save_slice_dataset_hash {
    my( $self ) = @_;

    my $h    = $self->slice_dataset_hash;
    
    my $hash_file = Bio::Otter::Lace::PersistentFile->new;
    $hash_file->root($self->home);
    $hash_file->name($DATASET_HASH_FILE);
    my $write = $hash_file->write_file_handle;

    while (my ($ds_name, $slices) = each %$h) {
        $ds_name =~ s/\t/\\t/g;      # Escape tab characterts in dataset name (likely ?)
        map { s/\t/\\t/g } @$slices; # Escape tab characterts in slice   name (v. unlikely)
        print $write "$ds_name\t@$slices\n";
    }
}

sub recover_slice_dataset_hash {
    my( $self ) = @_;

    my $cl   = $self->Client or confess "No Otter Client attached";
    my $h    = $self->slice_dataset_hash;
    
    my $hash_file = Bio::Otter::Lace::PersistentFile->new;
    $hash_file->root($self->home);
    $hash_file->name($DATASET_HASH_FILE);
    my $read = $hash_file->read_file_handle;
    
    while (<$read>) {
        chomp;
        my ($ds_name, @slices) = split(/\t/, $_);
        $ds_name =~ s/\\t/\t/g;     # Unscape tab characterts in dataset name (v. unlikely)
        map { s/\\t/\t/g } @slices; # Unscape tab characterts in slice   name (v. unlikely)
        $h->{$ds_name} = \@slices;
    }
}


sub save_all_slices {
    my( $self ) = @_;
    
    #warn "SAVING ALL SLICES";
    $self->error_flag(1);

    # Make sure we don't have a stale database handle
    $self->drop_aceperl_db_handle;

    my $sd_h = $self->slice_dataset_hash;
    #warn "HASH = '$sd_h' has ", scalar(keys %$sd_h), " elements";
    ### This call to each was failing to return anything
    ### the second time it was called, proabably because
    ### we were exiting each the first with an exception
    ### so the iterator didn't get reset.
    #while (my ($name, $ds) = each %$sd_h) {
    foreach my $ds_name (keys %$sd_h) {
        my $slices = $sd_h->{$ds_name};
        foreach my $slice(@$slices){
            warn "SAVING SLICE '$slice' WITH DATASET $ds_name\n";
            $self->save_otter_slice($slice, $ds_name);
        }
    }
    $self->error_flag(0);
}

sub save_otter_slice {
    my( $self, $name, $dataset_name ) = @_;
    $self->error_flag(1);
    confess "Missing slice name argument"   unless $name;
    confess "Missing DatsSet argument"      unless $dataset_name;

    my $ace    = $self->aceperl_db_handle;
    my $client = $self->Client or confess "No Client attached";
    
    # Get the Assembly object ...
    $ace->raw_query(qq{find Assembly "$name"});
    my $ace_txt = $ace->raw_query('show -a');

    # ... its SubSequences ...
    $ace->raw_query('query follow SubSequence where ! CDS_predicted_by');
    $ace_txt .= $ace->raw_query('show -a');

    # ... and all the Loci attached to the SubSequences.
    $ace->raw_query('Follow Locus');
    $ace_txt .= $ace->raw_query('show -a');
    
    # List of people for Authors
    $ace->raw_query(qq{find Person *});
    $ace_txt .= $ace->raw_query('show -a');

    # Then get the information for the TilePath
    $ace->raw_query(qq{find Assembly "$name"});
    $ace->raw_query('Follow AGP_Fragment');
    # Do show -a on a restricted list of tags
    foreach my $tag (qw{
        Otter
        DB_info
        Annotation
        })
    {
        $ace_txt .= $ace->raw_query("show -a $tag");
    }

    # Cleanup text
    $ace_txt =~ s/\0//g;            # Remove nulls
    $ace_txt =~ s{^\s*//.+}{\n}mg;  # Strip comments
    
    if($self->Client->debug){
        my $debug_file = Bio::Otter::Lace::PersistentFile->new();
        $debug_file->name("otter-debug.$$.save.ace");
        my $fh = $debug_file->write_file_handle();
        print $fh $ace_txt;
        close $fh;
    }else{
        warn "Debug switch is false\n";
    }

    my $ace_file = Bio::Otter::Lace::TempFile->new;
    $ace_file->name('lace_edited.ace');
    my $write = $ace_file->write_file_handle;
    print $write $ace_txt;
    my $xml = Bio::Otter::Converter::ace_to_XML($ace_file->read_file_handle);
    
    if($self->Client->debug){
        my $debug_file = Bio::Otter::Lace::PersistentFile->new();
        $debug_file->name("otter-debug.$$.save.xml");
        my $fh = $debug_file->write_file_handle();
        print $fh $xml;
        close $fh;
    }else{
        warn "Debug switch is false\n";
    }

    my $success = $client->save_otter_xml($xml, $dataset_name);
    $self->error_flag($success ? 0 : 1);
    return $success;
}

sub unlock_all_slices {
    my( $self ) = @_;
    
    my $sd_h = $self->slice_dataset_hash;

    # if the unlock otter slice goes wrong half way through
    # the recover will try to unlock the clones again.
    foreach my $ds_name (keys %$sd_h) {
        my $slices = $sd_h->{$ds_name};
        foreach my $slice(splice(@$slices)){
            $self->unlock_otter_slice($slice, $ds_name);
        }
    }
}

sub unlock_otter_slice{
    my( $self, $slice_name, $dataset_name ) = @_;

    confess "Missing slice name argument"   unless $slice_name;
    confess "Missing DatsSet name argument" unless $dataset_name;

    my $client   = $self->Client or confess "No Client attached";

    my $xml_file = Bio::Otter::Lace::PersistentFile->new;
    $xml_file->root($self->home);
    $xml_file->name(".${slice_name}${dataset_name}${LOCK_REGION_XML_FILE}");
    return unless -e $xml_file->full_name();
    my $xml = '';
    my $read = $xml_file->read_file_handle;
    while(<$read>){
        $xml .= $_;
    }
    return unless $xml;

    return $client->unlock_otter_xml($xml, $dataset_name);
    
}

sub aceperl_db_handle {
    my( $self ) = @_;

    my( $dbh );
    unless ($dbh = $self->{'_aceperl_db_handle'}) {
        my $home = $self->home;
        my $tace = $self->tace;
        
        # Check for ACEDB.wrm, or tace will hang waiting for
        # an answer to the "initialize database?" question.
        my $init_file = "$home/database/ACEDB.wrm";
        unless (-e $init_file) {
            confess "The file '$init_file' is missing - database has not been initialized";
        }
        
        $dbh = $self->{'_aceperl_db_handle'}
            = Ace->connect(-PATH => $home, -PROGRAM => $tace)
                or confess "Can't connect to database in '$home': ", Ace->error;
    }

    return $dbh;
}

sub drop_aceperl_db_handle {
    my( $self ) = @_;
    
    $self->{'_aceperl_db_handle'} = undef;
}

sub make_database_directory {
    my( $self ) = @_;
    
    my $home = $self->home;
    my $tar  = $self->tar_file or confess "tar_file not set";
    mkdir($home, 0777) or die "Can't mkdir('$home') : $!\n";
    
    my $tar_command = "cd $home ; tar xf $tar";
    if (system($tar_command) != 0) {
        $self->error_flag(1);
        confess "Error running '$tar_command' exit($?)";
    }
    
    # This acefile from the tar file needs to get parsed
    $self->add_acefile("$home/rawdata/methods.ace");
    
    $self->make_passwd_wrm;
    $self->edit_displays_wrm;
}

sub make_passwd_wrm {
    my( $self ) = @_;

    my $passWrm = $self->home . '/wspec/passwd.wrm';
    my ($prog) = $0 =~ m{([^/]+)$};
    my $real_name      = ( getpwuid($<) )[0];
    my $effective_name = ( getpwuid($>) )[0];

    my $fh = gensym();
    sysopen($fh, $passWrm, O_CREAT | O_WRONLY, 0644)
        or confess "Can't write to '$passWrm' : $!";
    print $fh "// PASSWD.wrm generated by $prog\n\n";

    # acedb looks at the real user ID, but some
    # versions of the code seem to behave differently
    if ( $real_name ne $effective_name ) {
        print $fh "root\n\n$real_name\n\n$effective_name\n\n";
    }
    else {
        print $fh "root\n\n$real_name\n\n";
    }

    close $fh;    # Must close to ensure buffer is flushed into file
}

sub edit_displays_wrm {
    my( $self ) = @_;
    
    my $home  = $self->home;
    my $title = $self->title;
    
    my $displays = "$home/wspec/displays.wrm";

    my $disp_in = gensym();
    open $disp_in, $displays or confess "Can't read '$displays' : $!";
    my @disp = <$disp_in>;
    close $disp_in;

    foreach (@disp) {
        next unless /^_DDtMain/;

        # Add our title onto the Main window
        s/\s-t\s*"[^"]+/ -t "$title/i;  # " sorry just to fix emacs syntax highlight
        last;
    }

    my $disp_out = gensym();
    open $disp_out, "> $displays" or confess "Can't write to '$displays' : $!";
    print $disp_out @disp;
    close $disp_out;
}


sub initialize_database {
    my( $self ) = @_;

    my $home = $self->home;
    my $tace = $self->tace;
    my @parse_commands = map "parse $_\n",
        $self->list_all_acefiles;

    my $parse_log = "$home/init_parse.log";
    my $pipe = "| $tace $home > $parse_log";
    
    my $pipe_fh = gensym();
    open $pipe_fh, $pipe
        or die "Can't open pipe '$pipe' : $!";
    # Say "yes" to "initalize database?" question.
    print $pipe_fh "y\n";
    foreach my $com (@parse_commands) {
        print $pipe_fh $com;
    }
    close $pipe_fh or die "Error initializing database exit($?)\n";

    my $fh = gensym();
    open $fh, $parse_log or die "Can't open '$parse_log' : $!";
    my $file_log = '';
    my $in_parse = 0;
    my $errors = 0;
    while (<$fh>) {
        if (/parsing/i) {
            $file_log = "  $_";
            $in_parse = 1;
        }
        
        if (/(\d+) (errors|parse failed)/i) {
            if ($1) {
                warn "\nParse error detected:\n$file_log  $_\n";
                $errors++;
            }
        }
        elsif (/Sorry/) {
            warn "Apology detected:\n$file_log  $_\n";
            $errors++;
        }
        elsif ($in_parse) {
            $file_log .= "  $_";
        }
    }
    close $fh;

    confess "Error initializing database\n" if $errors;
    return 1;
}

sub write_pipeline_data {
    my( $self, $ss, $ace_file ) = @_;

    my $dataset = $self->Client->get_DataSet_by_name($ss->dataset_name);
    $dataset->selected_SequenceSet($ss);    # Not necessary?
    my $ens_db = $dataset->get_cached_DBAdaptor();
    my $pipe_db;
    my $fetch_pipe = Bio::Otter::Lace::Defaults::fetch_pipeline_switch();
    if($fetch_pipe){
	$pipe_db = Bio::Otter::Lace::PipelineDB::get_DBAdaptor($dataset->get_cached_DBAdaptor);
	#$pipe_db->dnadb($ens_db->dnadb);
    }
    $ens_db = $pipe_db || $ens_db;
    
    $ens_db->assembly_type($ss->name);
    my $species = $dataset->species();
    warn "This is species <$species>\n";
    my $factory = $self->{'_pipeline_data_factory'} ||= $self->make_AceDataFactory($ens_db, $species);
    
    # create file for output and add it to the acedb object
    $ace_file ||= $self->home . "/rawdata/pipeline.ace";
    my $fh;
    if(ref($ace_file) eq 'GLOB'){
        $fh = $ace_file;
    }else{ 
        $fh = gensym();
        $self->add_acefile($ace_file);
        open $fh, "> $ace_file" or confess "Can't write to '$ace_file' : $!";
    }
    $factory->file_handle($fh);

    my $slice_adaptor = $ens_db->get_SliceAdaptor();
    
    # note: the next line returns a 2 dimensional array (not a one dimensional array)
    # each subarray contains a list of clones that are together on the golden path
    my $sel = $ss->selected_CloneSequences_as_contig_list ;
    foreach my $cs (@$sel) {

        my $first_ctg = $cs->[0];
        my $last_ctg = $cs->[$#$cs];

        my $chr = $first_ctg->chromosome->name;  
        my $chr_start = $first_ctg->chr_start;
        my $chr_end = $last_ctg->chr_end;

        my $slice = $slice_adaptor->fetch_by_chr_start_end($chr, $chr_start, $chr_end);
        
        ### Check we got a slice
        my $tp = $slice->get_tiling_path;
        my $type = $slice->assembly_type;
        #warn "assembly type = $type";
        if (@$tp) {
            foreach my $tile (@$tp) {
                print STDERR "contig: ", $tile->component_Seq->name, "\n";
            }
        } else {
            warn "No components in tiling path";
        }

        $factory->ace_data_from_slice($slice);
    }
    $factory->drop_file_handle;
    close $fh;
    
    Bio::Otter::Lace::SatelliteDB::disconnect_DBAdaptor($ens_db) if $fetch_pipe;
}

sub make_AceDataFactory {
    my( $self, $ens_db, $species ) = @_;

    # create new datafactory object - cotains all ace filters and produces the data from these
    my $factory = Bio::EnsEMBL::Ace::DataFactory->new;       
    # $factory->add_all_Filters($ensdb);   
    my $ana_adaptor = $ens_db->get_AnalysisAdaptor;

    ##----------code to add all of the ace filters to data factory-----------------------------------

    my $fetch_all_pipeline_data = Bio::Otter::Lace::Defaults::fetch_pipeline_switch();
    my $logic_to_load = {};
    my $module_options = {};
    if ($fetch_all_pipeline_data) {
	$logic_to_load = $self->Client->option_from_array([$species, 'use_filters']);
	$module_options = $self->Client->option_from_array([$species, 'filter']);

	# require them
	for my $s (keys %$logic_to_load){
	    next unless $logic_to_load->{$s};
	    # warn " AceBatabase.pm: $s\n";
	    my $file = $module_options->{$s}->{module}.".pm";
	    $file =~ s{::}{/}g;
	    # warn " AceDatabase.pm: requiring $file \n";
	    eval{  require "$file"  };
	    if($@){
		delete $logic_to_load->{$s};
		warn "AceDatabase couldn't find file $file\n";
                warn $@ if $self->Client->debug();
	    }
	}
    }

    # If we aren't fetching all the analysis, we only need the DNA
    my $submitcontig = $self->Client->option_from_array([qw!client dna!]) 
        || ( $fetch_all_pipeline_data ? 'SubmitContig' : 'otter' );
    $logic_to_load->{$submitcontig} = 1;
    $module_options->{$submitcontig}->{'module'} = 'Bio::EnsEMBL::Ace::Filter::DNA';

    foreach my $logic_name (keys %$logic_to_load){
	next unless $logic_to_load->{$logic_name};
	# class successfully required already.
	my $class = $module_options->{$logic_name}->{'module'};
        my $filt;
	# check there is an analysis
        if (my $ana = $ana_adaptor->fetch_by_logic_name($logic_name)) {
            $filt = $class->new;
            $filt->analysis_object($ana);
        } elsif($module_options->{$logic_name}->{'create_without_ana'}){
            $filt = $class->new();
        } else {
            warn "No analysis called '$logic_name'\n";
        }
        if($filt){
	    # find the options for the filter?
	    foreach my $option(keys (%{$module_options->{"$logic_name"}})){
		# warn "checking $filt for method $option\n";
		next unless $filt->can($option);
		my $value = $module_options->{"$logic_name"}->{$option};
		# warn "setting $option : $value \n";
		$filt->$option($value);
	    }

	    # add the filter to the factory
            $factory->add_AceFilter($filt);
        }
    }
        
    return $factory;
}


##  creates a data factory and adds all the appropriate filters to
##  it. It then produces a slice from the ensembl db (using the
##  $dataset coords) and produces output based on that slice in
##  ensembl.ace
sub write_ensembl_data {
    my( $self, $ss ) = @_;

    my $dataset = $self->Client->get_DataSet_by_name($ss->dataset_name);
    my $species = $dataset->species();
    my $ensembl_sources = $self->Client->option_from_array([$species, 'ensembl_sources']);
    ### Analysis logic name should not be hard coded, so now they're not.
    foreach my $key(keys(%$ensembl_sources)){
        my $logic = $ensembl_sources->{$key};
        warn "I'm gonna fetch using key '$key' in meta table and set logic to '$logic'\n";
        $self->write_ensembl_data_for_key($ss, $key, $logic);
    }
}

sub make_ensembl_gene_DataFactory {
    my( $self, $ens_db, $logic_name ) = @_;
    
    my $factory = Bio::EnsEMBL::Ace::DataFactory->new;
    my $ana_adaptor = $ens_db->get_AnalysisAdaptor;
    my $ensembl = Bio::EnsEMBL::Ace::Filter::Gene->new;
    $ensembl->url_string('http\:\/\/www.ensembl.org\/Homo_sapiens\/contigview?highlight=%s&chr=%s&vc_start=%s&vc_end=%s');    
    $ensembl->analysis_object( $ana_adaptor->fetch_by_logic_name($logic_name) );
    $factory->add_AceFilter($ensembl);
    return $factory;
}

sub write_ensembl_data_for_key {
    my( $self, $ss, $key, $logic_name ) = @_;

    my $debug_flag = 0;

    my $dataset = $self->Client->get_DataSet_by_name($ss->dataset_name);
    $dataset->selected_SequenceSet($ss);    # Not necessary?
    my $ens_db = Bio::Otter::Lace::SatelliteDB::get_DBAdaptor(
        $dataset->get_cached_DBAdaptor, $key
        ) or return;

    # create file for output and add it to the acedb object
    my $ace_file = $self->home . "/rawdata/$key.ace";
    my $fh = gensym();
    open $fh, "> $ace_file" or confess "Can't write to '$ace_file' : $!";
    $self->add_acefile($ace_file);

    my $type = $ens_db->assembly_type;
    # later on will have to get chromsome names...not proper way to do it
    my $ch = get_all_LaceChromosomes($ens_db);

    my $factory = $self->{'_ensembl_gene_data_factory'}{$logic_name}
        ||= $self->make_ensembl_gene_DataFactory($ens_db, $logic_name);

    my $slice_adaptor = $ens_db->get_SliceAdaptor();

    my $sel = $ss->selected_CloneSequences_as_contig_list;
    # unlike sanger (pipeline) databases, where data is clone based, 
    # in this case we need to deal with slice as a whole

    # Slightly smarter than rejecting entire slice if anything
    # different.  Is able to build a subslice if beginning or end
    # is incorrect, but can't build multiple subslices (all kinds
    # of duplicate partial gene problems could result in such
    # cases).

    # Since locally the agp could be correct, but globally wrong
    # has to deal with clone order walking in the wrong direction

    # Various patalogical cases are not dealt with optimally.  If
    # A matches; B doesn't but C, D, E and F match, will make a
    # subslice out of A.  Could be handelled, but would require a
    # double pass.

    foreach my $cs (@$sel) {

	my $otter_slice_name;
	{
	    # need to get name of slice in otter space (fetch from ensembl
	    # will be in a different coordinate space, but because of
	    # checks they are guarenteed to be equivalent)

	    my $first_ctg = $cs->[0];
	    my $last_ctg = $cs->[$#$cs];

	    my $chr = $first_ctg->chromosome->name;  
	    my $chr_start = $first_ctg->chr_start;
	    my $chr_end = $last_ctg->chr_end;
	    $otter_slice_name="$chr.$chr_start-$chr_end";
	}

	# check if agp of this DB is in sync for the selected clones
	# dump if in sync, else skip
	my $off=0;
	my $first=-1;
	my $first_dir;
	my $last;
	my $last_edge;
	my $slice_start;
	my $slice_end;
	my $fail;
	my $chr;
	for (my $i=0; $i < @$cs; $i++) {
	    my $ctg= $cs->[$i];

	    my $ens_ctg_set=get_LaceCloneSequence_by_sv($ens_db,$ch,
						      $ctg->accession,$ctg->sv,$type,
                                                      $debug_flag);
	    my $pass=0;
	    # should get only one match (present, but not unfinished)
	    if(scalar(@$ens_ctg_set)==1){
		my $ens_ctg=$ens_ctg_set->[0];
		# check if same part of contig is part of external agp
		if($ens_ctg->contig_start==$ctg->contig_start &&
		   $ens_ctg->contig_end==$ctg->contig_end
		   ){
		    print "DEBUG: same contig used\n" if $debug_flag;
		    # if first clone, save; else check order is still ok
		    if($first>-1){
			$fail=1;
			# check sequential
			if($i=$last+1){
			    # check consistent direction
			    my $this_dir=-1;
			    if($ens_ctg->contig_strand==$ctg->contig_strand){
				$this_dir=1;
			    }
			    if($first_dir==$this_dir){
				# check agp consecutive
				if($first_dir==1 && $ens_ctg->chr_start==$last_edge+1){
				    $last=$i;
				    $last_edge=$ens_ctg->chr_end;
				    $slice_end=$ens_ctg->chr_end;
				    $fail=0;
				}elsif($first_dir==-1 && $ens_ctg->chr_end==$last_edge-1){
				    # -ve direction not handled...so
				    confess "ERR: should never get here!!";
				}
			    }
			}
		    }else{
			print "DEBUG: saved first $i\n" if $debug_flag;
			$first=$i;
			$last=$i;
			$chr=$ens_ctg->chromosome->name;  
			if($ens_ctg->contig_strand==$ctg->contig_strand){
			    # same direction
			    $last_edge=$ens_ctg->chr_end;
			    $slice_start=$ens_ctg->chr_start;
			    $slice_end=$ens_ctg->chr_end;
			    $first_dir=1;
			}else{
			    $last_edge=$ens_ctg->chr_start;
			    $slice_start=$ens_ctg->chr_end;
			    $slice_end=$ens_ctg->chr_start;
			    $first_dir=-1;
			    # reverse direction

			    # FIXME temporary:
			    print "WARN: agp is in reverse direction";
			    print " - not currently handled\n";
			    $first=-1;

			}
		    }
		}
	    }
	    # right now, if $first not set for $i=0 can't continue
	    if($i==0 && $first==-1){$fail=1;}
	    # once started a slice with first, if fail then no point checking further
	    last if $fail;
	}
	# if something was saved
	if($first>-1){
	    print "DEBUG: Fetching slice $first:$slice_start-$last:$slice_end\n" if $debug_flag;
	    my $slice = $slice_adaptor->fetch_by_chr_start_end($chr, $slice_start, $slice_end);
	    $slice->name($otter_slice_name);
	    print $fh $factory->ace_data_from_slice($slice);
	}
    }
    close $fh;
    
    ### Disconnect Ensembl DBAdaptor
    Bio::Otter::Lace::SatelliteDB::disconnect_DBAdaptor($ens_db);
}


# look for contigs for this sv
sub get_LaceCloneSequence_by_sv{
    my($dba,$ch,$acc,$sv,$type, $debug_flag)=@_;
    print "DEBUG: checking $acc,$sv,$type\n" if $debug_flag;
    my %id_chr = map {$_->chromosome_id, $_} @$ch;
    my $sth = $dba->prepare(q{
        SELECT a.chromosome_id
          , a.chr_start
          , a.chr_end
          , a.contig_start
	  , a.contig_end
          , a.contig_ori
	FROM assembly a
	  , clone cl
	  , contig c
        WHERE cl.embl_acc= ?
          AND cl.embl_version= ?
	  AND cl.clone_id=c.clone_id
	  AND c.contig_id=a.contig_id
          AND a.type = ?
        });
    $sth->execute($acc,$sv,$type);
    my( $chr_id,
	$chr_start, $chr_end,
	$contig_start, $contig_end, $strand );
    $sth->bind_columns( \$chr_id,
			\$chr_start, \$chr_end,
			\$contig_start, \$contig_end, \$strand );
    my $cs = [];
    while ($sth->fetch) {
	my $cl=Bio::Otter::Lace::CloneSequence->new;
	#$cl->accession($acc);
	#$cl->sv($sv);
	#$cl->length($ctg_length);
	$cl->chromosome($id_chr{$chr_id});
	$cl->chr_start($chr_start);
	$cl->chr_end($chr_end);
	$cl->contig_start($contig_start);
	$cl->contig_end($contig_end);
	$cl->contig_strand($strand);
	#$cl->contig_name($ctg_name);
	push(@$cs, $cl);
	print "DEBUG: $chr_start-$chr_end; $contig_start-$contig_end\n" if $debug_flag;
    }
    return $cs;
}


sub get_all_LaceChromosomes {
    my($dba)=@_;
    my($ch);
    my $sth = $dba->prepare(q{
	SELECT chromosome_id
	    , name
	    , length
	FROM chromosome
	});
    $sth->execute;
    my( $chr_id, $name, $length );
    $sth->bind_columns(\$chr_id, \$name, \$length);
        
    while ($sth->fetch) {
	my $chr = Bio::Otter::Lace::Chromosome->new;
	$chr->chromosome_id($chr_id);
	$chr->name($name);
	$chr->length($length);
	push(@$ch, $chr);
    }
    return($ch);
}


sub DESTROY {
    my( $self ) = @_;
    my $home = $self->home;
    print STDERR "DESTROY has been called for AceDatabase.pm with home $home\n";
    if ($self->error_flag) {
        warn "Not cleaning up '$home' because error flag is set\n";
        return;
    }
    my $client = $self->Client;
    if($client){
        $self->unlock_all_slices();# if $client->write_access;
    }
    rmtree($home);
}

1;
